#include "apigw/src/node.hpp"

// AFAP includes.
#include "components/csv/src/CsvFile.hpp"
#include "components/numbers/src/Validator.hpp"
#include "components/tools/src/Tools.hpp"
#include "components/types/src/KeyValue.hpp"
#include "components/worker_node/src/config_parameters.hpp"
#include "TRecQ763Number.hpp"
#include "UvrTypes.hpp"

// Boost includes.
#include <boost/lexical_cast.hpp>

// STL includes.
#include <regex>
#include <filesystem>

namespace
{
    const std::string s_cmdIdUpdateRecords = "update-records";

    const std::string s_recordKeyIdSrc = "ID_SRC";
    const std::string s_recordKeyIdDst = "ID_DST";
    const std::string s_recordKeyReleaseCode = "RELEASE_CODE";
    const std::string s_recordKeyDuration = "DURATION";
    const std::string s_recordKeyCallId = "CALL_ID";
    const std::string s_recordKeyCallType = "CALL_TYPE";
    const std::string s_callTypeValueOutgoing = "OUT";
    const std::string s_callTypeValueIncoming = "IN";

    bool s_isAsyncMode = true;

    std::string parseRecord(const KeyValue::PairsList &keyValueList, Uvr::CallRecord &record);

    void verifyCallVerdictStub(const Uvr::VerificationResult &result)
    {
    }

    void callInformationRequestStub(
        const Uvr::UvrId_t uvrId,
        const Uvr::CallInformationRequest &request)
    {
    }

} // namespace

namespace apigw
{
    node::node(const TypeInst &ownInst,
               boost::asio::io_context &ioc,
               const configuration::manager &config)
        : pm::node(ownInst, ioc, config),
          m_verifyCallsTransactionManager(ownInst, verifyCallsManagerId, m_uvrId),
          m_digestsTransactionManager(ownInst, digestsManagerId, m_uvrId),
          m_fsLocal(ownInst)
    {
    }

    void node::start()
    {
        pm::node::start();
        registerCliCommands();
    }

    int api::init(boost::asio::io_context &ioc,
                  configuration::manager &config,
                  const bool asyncMode)
    {
        s_isAsyncMode = asyncMode;
        return base_api::init(ioc, config);
    }

    bool api::start(std::size_t instId)
    {
        std::list<std::string> requiredCallbacks = {callInformationRequestCallbackId};
        if (s_isAsyncMode)
        {
            requiredCallbacks.push_back(verifyCallVerdictCallbackId);
        }

        if (!getNode()->isRegistered(requiredCallbacks))
        {
            LOGERROR("Not all callbacks were registered.");
            return false;
        }
        else
        {
            return base_api::start(instId);
        }
    }

    void node::prepareCounters()
    {
        addCounter(pm::counter::calls::ATTMS, pm::counter::indexing::yes);
    }

    void node::handleSpecificPrimitive(const primitive::GenericPtr &prim)
    {
        switch (prim->m_header->id)
        {
            case ::primitive::VERIFY_CALL_VRD:
            {
                handleVerifyCallVerdict(prim);
                break;
            }
            case ::primitive::NOTIFY_CALL_INFO_REQ:
            {
                handleNotifyCallInfoReq(prim);
                break;
            }
            case ::primitive::QUERY_UVR_IDS_RSP:
            {
                handleQueryUvrIdsRsp(prim);
                break;
            }
            case ::primitive::QUERY_UVR_INFO_RSP:
            {
                handleQueryUvrInfoRsp(prim);
                break;
            }
            case ::primitive::QUERY_HUB_INFO_RSP:
            {
                handleQueryHubInfoRsp(prim);
                break;
            }
            case ::primitive::QUERY_OPERATOR_INFO_RSP:
            {
                handleQueryOperatorInfoRsp(prim);
                break;
            }
            case ::primitive::EXEC_CMD_REQ:
            {
                handleExecCmdReq(prim);
                break;
            }
            case ::primitive::STATUS_CHANGE_IND:
            {
                handleStatusChangeInd(prim);
                break;
            }
            default:
            {
                pm::node::handleSpecificPrimitive(prim);
            }
        }
    }

    void node::handleVerifyCallVerdict(const primitive::GenericPtr &prim)
    {
        const auto verifyCallVerdict =
            std::dynamic_pointer_cast<::primitive::VerifyCallVerdict>(prim);

        ::transaction::errors closeTransactionResult;
        const auto transactionData =
            m_verifyCallsTransactionManager.closeTransaction(verifyCallVerdict->m_header->ulpKey,
                                                             closeTransactionResult);

        Uvr::VerificationResult result{verifyCallVerdict->rlc,
                                       verifyCallVerdict->m_header->ulpKey,
                                       verifyCallVerdict->uvrT,
                                       transactionData.callId};

        Uvr::CallRecord record;
        record.numA = TRecQ763Number::toStdString(transactionData.callingNumber);
        record.numB = TRecQ763Number::toStdString(transactionData.calledNumber);
        record.numC = TRecQ763Number::toStdString(transactionData.originalCalledNumber);
        record.numD = TRecQ763Number::toStdString(transactionData.genericNumber);
        record.date = transactionData.get_start();
        record.dateAct = transactionData.get_stop();
        transactionData.idSrc == Uvr::undefinedOperatorId ? record.idSrc = std::nullopt
                                                          : record.idSrc = transactionData.idSrc;
        record.idDst = std::nullopt;
        record.sessionId = result.sessionId;
        record.idRel = Uvr::VerdictTrigger::antifraud;
        record.response = result.verificationVerdict;
        record.releaseCode = std::nullopt;
        record.callDuration = std::nullopt;
        record.idUvrT = result.idUvrT;
        record.callId = transactionData.callId;
        record.callType = Uvr::CallType::incoming;

        // If the user id is undefined, it means that the verification was
        // requested synchronously, so the result must be returned via a promise.
        const auto &userId = transactionData.userId;
        if (userId == Uvr::undefinedUserId)
        {
            m_verificationPromises.fulfill(verifyCallVerdict->m_header->ulpKey,
                                           result);

            if (!record.idUvrT.has_value()
                || Digest::shouldMakeIncident(record.idUvrT.value()))
            {
                std::vector<std::any> funcs;
                this->m_callbacks.getForAllUsers(storeRecordCallbackId, funcs);
                for (const auto &func : funcs)
                {
                    // TODO[multi-uvr]: Select the correct UVR ID.
                    std::any_cast<Uvr::StoreRecordCallback>(func)(m_uvrId, record);
                }
            }

            return;
        }

        // If the user ID is something specific, it means that the verification
        // was requested asynchronously, so the result must be returned via the
        // user-registered callback.
        const auto verifyCallVerdictCallback = m_callbacks.get(verifyCallVerdictCallbackId, userId);
        if (verifyCallVerdictCallback.has_value())
        {
            std::any_cast<Uvr::VerifyCallVerdictCallback>(verifyCallVerdictCallback)(result);
        }

        if (!record.idUvrT.has_value()
            || Digest::shouldMakeIncident(record.idUvrT.value()))
        {
            const auto storeRecordCallback = m_callbacks.get(storeRecordCallbackId, userId);
            if (storeRecordCallback.has_value())
            {
                // TODO[multi-uvr]: Select the correct UVR ID.
                std::any_cast<Uvr::StoreRecordCallback>(storeRecordCallback)(m_uvrId, record);
            }
        }
    }

    void node::handleNotifyCallInfoReq(const primitive::GenericPtr &prim)
    {
        auto req =
            std::dynamic_pointer_cast<::primitive::NotifyCallInfoRequest>(prim);
        if (req)
        {
            const auto uvrId = req->uvrId;

            LOGDEBUG(m_logTag,
                     "Received NOTIFY_CALL_INFO_REQ for UVR with ID=", uvrId,
                     " with ", req->fileNamesList.size(), " file(s).");

            for (auto const &reqFile : req->fileNamesList)
            {
                std::string unzipFilename;
                m_fsLocal.unzipFile(reqFile, unzipFilename);
                size_t numOfParsedRows = 0;
                auto parseLine = [this, uvrId, &numOfParsedRows](const std::string &line)
                {
                    Uvr::CallInformationRequest request(line.c_str());
                    std::vector<std::any> funcs;
                    this->m_callbacks.getForAllUsers(callInformationRequestCallbackId, funcs);
                    for (const auto &func : funcs)
                    {
                        std::any_cast<Uvr::CallInformationRequestCallback>(func)(uvrId, request);
                    }
                    ++numOfParsedRows;
                    return false; // no need to interrupt, processing can be continued.
                };
                if (!Csv::File::parse(unzipFilename, parseLine, false) || numOfParsedRows == 0)
                {
                    // Extract ID_REQ from Call Info Request file name.
                    Uvr::RequestId_t idReq = 0;
                    const std::filesystem::path filePath(unzipFilename);
                    const std::string fileStem = filePath.stem().string();
                    const std::regex idsRegex("REQ_(\\d+)_(\\d+)_.*");
                    std::smatch idsMatches;
                    if (std::regex_match(fileStem, idsMatches, idsRegex))
                    {
                        idReq = std::stoull(idsMatches[2].str());
                    }
                    else
                    {
                        LOGERROR(m_ownTypeInst,
                                 " Unable to extract ID_REQ from Call Info Request file name: ",
                                 unzipFilename);
                        return;
                    }

                    Uvr::CallInformationResponse response("", "", "", "", "", "",
                                                          idReq,
                                                          std::nullopt, std::nullopt,
                                                          Uvr::ResponseCode::badRequest);
                    processCallInformation({response});
                }
                m_fsLocal.deleteFile(unzipFilename);
            }
        }
    }

    void node::handleQueryUvrIdsRsp(const primitive::GenericPtr &prim)
    {
        const auto QueryUvrIdsRsp =
            std::dynamic_pointer_cast<::primitive::QueryUvrIdsRsp>(prim);

        ::transaction::errors closeTransactionResult;
        m_digestsTransactionManager.closeTransaction(QueryUvrIdsRsp->m_header->ulpKey,
                                                     closeTransactionResult);

        m_uvrIdsPromises.fulfill(QueryUvrIdsRsp->m_header->ulpKey,
                                 QueryUvrIdsRsp->uvrIds);
    }

    void node::handleQueryUvrInfoRsp(const primitive::GenericPtr &prim)
    {
        const auto rsp =
            std::dynamic_pointer_cast<::primitive::QueryUvrInfoRsp>(prim);

        ::transaction::errors closeTransactionResult;
        m_digestsTransactionManager.closeTransaction(rsp->m_header->ulpKey,
                                                     closeTransactionResult);

        m_uvrInfoPromises.fulfill(rsp->m_header->ulpKey, rsp->uvrInfo);
    }

    void node::handleQueryHubInfoRsp(const primitive::GenericPtr &prim)
    {
        const auto rsp =
            std::dynamic_pointer_cast<::primitive::QueryHubInfoRsp>(prim);

        ::transaction::errors closeTransactionResult;
        m_digestsTransactionManager.closeTransaction(rsp->m_header->ulpKey,
                                                     closeTransactionResult);

        m_hubInfoPromises.fulfill(rsp->m_header->ulpKey, rsp->hubInfo);
    }

    void node::handleQueryOperatorInfoRsp(const primitive::GenericPtr &prim)
    {
        const auto rsp =
            std::dynamic_pointer_cast<::primitive::QueryOperatorInfoRsp>(prim);

        ::transaction::errors closeTransactionResult;
        m_digestsTransactionManager.closeTransaction(rsp->m_header->ulpKey,
                                                     closeTransactionResult);

        m_operatorInfoPromises.fulfill(rsp->m_header->ulpKey, rsp->operatorInfo);
    }

    void node::handleExecCmdReq(const primitive::GenericPtr &prim)
    {
        auto req = std::dynamic_pointer_cast<::primitive::ExecCmdReq>(prim);

        const auto cmdId = req->cmd;
        if (cmdId == s_cmdIdUpdateRecords)
        {
            execCmdUpdateRecords(req);
        }
        else
        {
            const std::string response = m_logTag.value() + ": Unknown command \"" + cmdId + "\".";
            sendExecCmdRsp(req->m_header->from, {response});
        }
    }

    void node::execCmdUpdateRecords(const std::shared_ptr<primitive::ExecCmdReq> &req)
    {
        std::vector<std::any> funcs;
        this->m_callbacks.getForAllUsers(updateRecordsCallbackId, funcs);
        if (funcs.empty())
        {
            sendExecCmdRsp(req->m_header->from, {s_cmdIdUpdateRecords + ": Internal error. No callback registered."});
            return;
        }

        Uvr::Strings responses;
        Uvr::CallRecords records;

        // Parse args to records.
        for (size_t argIndex = 0; argIndex < req->cmdArgs.size(); ++argIndex)
        {
            std::string response = s_cmdIdUpdateRecords + " at ";
            response += "arg " + std::to_string(argIndex) + ", ";

            const auto keyValueList = KeyValue::parse(req->cmdArgs.at(argIndex));
            if (keyValueList.empty())
            {
                response += "error: Failed to parse arg.";
                responses.emplace_back(std::move(response));
                continue;
            }

            Uvr::CallRecord record;
            const auto error = parseRecord(keyValueList, record);
            if (!error.empty())
            {
                response += error;
                responses.emplace_back(std::move(response));
                continue;
            }

            records.emplace_back(std::move(record));
        }

        if (records.empty())
        {
            sendExecCmdRsp(req->m_header->from, {s_cmdIdUpdateRecords + ", error: No records were parsed from arguments."});
            return;
        }
        else if (records.size() < req->cmdArgs.size())
        {
            responses.emplace_back(s_cmdIdUpdateRecords + ", warning: Parsed " + std::to_string(records.size())
                                   + " records out of " + std::to_string(req->cmdArgs.size()) + " args.");
        }
        else
        {
            responses.emplace_back(s_cmdIdUpdateRecords + ": All args have been parsed.");
        }

        for (const auto &func : funcs)
        {
            // TODO[multi-uvr]: Select the correct UVR ID.
            const auto nbUpdatedRecords =
                std::any_cast<Uvr::UpdateRecordsCallback>(func)(m_uvrId, records);

            if (nbUpdatedRecords == 0)
            {
                responses.emplace_back(s_cmdIdUpdateRecords + ", error: No records were updated.");
            }
            else if (nbUpdatedRecords < records.size())
            {
                responses.emplace_back(s_cmdIdUpdateRecords + ", warning: Updated " + std::to_string(nbUpdatedRecords)
                                       + " out of " + std::to_string(records.size()) + " records.");
            }
            else
            {
                responses.emplace_back(s_cmdIdUpdateRecords + ": All records have been updated.");
            }
        }

        sendExecCmdRsp(req->m_header->from, std::move(responses));
    }

    void node::sendExecCmdRsp(const TypeInst &to, Uvr::Strings &&responses)
    {
        auto rsp = std::make_shared<primitive::ExecCmdRsp>(std::move(responses));
        rsp->m_header->to = to;
        workers::node::send(rsp);
    }

    void node::handleStatusChangeInd(const primitive::GenericPtr &prim)
    {
        const auto status =
            std::dynamic_pointer_cast<::primitive::StatusChangeInd>(prim);

        const auto setNodeStatusCallback = m_callbacks.get(setNodeStatusCallbackId, m_internalUserId);
        if (setNodeStatusCallback.has_value())
        {
            std::any_cast<SetNodeStatusCallback>(setNodeStatusCallback)(status->status);
        }
    }

    node::user_callbacks_map::~user_callbacks_map()
    {
        reset();
    }

    void node::user_callbacks_map::reset(void)
    {
        std::unique_lock<std::shared_mutex> lock(m_funcsGuard);

        if (!m_funcs.empty())
        {
            m_funcs.clear();
        }
    }

    Uvr::UserId_t node::user_callbacks_map::doRegister(const std::string &handlerName,
                                                       std::any func_ptr,
                                                       Uvr::UserId_t userId)
    {
        std::unique_lock<std::shared_mutex> lock(m_funcsGuard);

        if (userId == Uvr::undefinedUserId)
        {
            static Uvr::UserId_t userInCntr{1};
            userId = userInCntr++;

            if (m_funcs.find(userId) != m_funcs.end())
            {
                LOGERROR(" Failed to register callback=", handlerName, " for NEW userId=", userId);
                return Uvr::undefinedUserId;
            }

            m_funcs.emplace(userId, UserCallbacks());
        }

        auto it = m_funcs.find(userId);
        if (it == m_funcs.end())
        {
            LOGERROR(" Failed to register callback=", handlerName, " for EXISTING userId=", userId);
            return Uvr::undefinedUserId;
        }

        it->second.insert(std::pair<std::string, std::any>(handlerName, func_ptr));

        return userId;
    }

    bool node::user_callbacks_map::isRegistered(const std::list<std::string> &handlersNames) const
    {
        std::shared_lock<std::shared_mutex> lock(m_funcsGuard);

        if (m_funcs.empty() && !handlersNames.empty())
        {
            LOGWARN("The list of registered callbacks is empty.");
            return false;
        }

        bool isAllRegistered = true;
        for (const auto &handlerName : handlersNames)
        {
            for (const auto &userCallbacks : m_funcs)
            {
                const auto &callbacks = userCallbacks.second;
                if (callbacks.find(handlerName) == callbacks.end())
                {
                    LOGWARN("Callback \"", handlerName, "\"",
                            " is not registered for the API user with",
                            " id = \"", userCallbacks.first, "\"");
                    isAllRegistered = false;
                }
            }
        }

        return isAllRegistered;
    }

    std::any node::user_callbacks_map::get(const std::string &handlerName,
                                           Uvr::UserId_t userId) const
    {
        std::shared_lock<std::shared_mutex> lock(m_funcsGuard);

        std::any func;

        auto it = m_funcs.find(userId);
        if (it != m_funcs.end())
        {
            auto jt = it->second.find(handlerName);
            if (jt != it->second.end())
            {
                func = jt->second;
            }
        }
        return func;
    }

    void node::user_callbacks_map::getForAllUsers(const std::string &handlerName,
                                                  std::vector<std::any> &funcs) const
    {
        std::shared_lock<std::shared_mutex> lock(m_funcsGuard);

        for (const auto &userCallbacks : m_funcs)
        {
            const auto &callbacks = userCallbacks.second;
            auto funcIter = callbacks.find(handlerName);
            if (funcIter != callbacks.end())
            {
                LOGD("Callback \"", handlerName,
                     "\" is found for the user with id=\"",
                     userCallbacks.first, "\"");
                funcs.push_back(funcIter->second);
            }
        }
    }

    Uvr::UserId_t node::doRegister(const std::string &handlerName,
                                   std::any callback,
                                   Uvr::UserId_t userId)
    {
        return m_callbacks.doRegister(handlerName, callback, userId);
    }

    bool node::isRegistered(const std::list<std::string> &handlersNames) const
    {
        return m_callbacks.isRegistered(handlersNames);
    }

    void node::doRegisterSetNodeStatusCallback(std::any callback)
    {
        m_internalUserId = doRegister(apigw::setNodeStatusCallbackId,
                                      callback,
                                      Uvr::undefinedUserId);
        doRegister(apigw::verifyCallVerdictCallbackId,
                   Uvr::VerifyCallVerdictCallback(verifyCallVerdictStub),
                   m_internalUserId);
        doRegister(apigw::callInformationRequestCallbackId,
                   Uvr::CallInformationRequestCallback(callInformationRequestStub),
                   m_internalUserId);
    }

    Uvr::SessionId_t node::doVerifyCall(Uvr::UserId_t userId,
                                        Uvr::OperatorId_t idSrc,
                                        Uvr::OctetStringPtr callingNumber,
                                        Uvr::OctetStringPtr calledNumber,
                                        Uvr::OctetStringPtr originalCalledNumber,
                                        Uvr::OctetStringPtr genericNumber,
                                        const Uvr::CallId_t &callId,
                                        bool isTrusted)
    {
        // If the call is trusted, then nothing more needs to be done.
        if (isTrusted)
        {
            // Increase the calls counter.
            increaseCounter(pm::counter::calls::ATTMS, std::to_string(idSrc));
            return Uvr::undefinedSessionId;
        }

        // Open a transaction for verification.
        Uvr::SessionId_t transactionId = openVerifyCallTransaction(userId,
                                                                   idSrc,
                                                                   callingNumber,
                                                                   calledNumber,
                                                                   originalCalledNumber,
                                                                   genericNumber,
                                                                   callId);

        // Send a primitive for verification.
        sendVerifyCallQuery(idSrc,
                            callingNumber,
                            calledNumber,
                            originalCalledNumber,
                            genericNumber,
                            callId,
                            transactionId);

        return transactionId;
    }

    Uvr::VerificationResult
    node::doVerifyCallSynchronously(const Uvr::OperatorId_t idSrc,
                                    const Uvr::OctetStringPtr &callingNumber,
                                    const Uvr::OctetStringPtr &calledNumber,
                                    const Uvr::OctetStringPtr &originalCalledNumber,
                                    const Uvr::OctetStringPtr &genericNumber,
                                    const Uvr::CallId_t &callId,
                                    const bool isTrusted)
    {
        // If the call is trusted, then nothing more needs to be done.
        if (isTrusted)
        {
            // Increase the calls counter.
            increaseCounter(pm::counter::calls::ATTMS, std::to_string(idSrc));
            return Uvr::VerificationResult();
        }

        // Open a transaction for verification.
        const Uvr::SessionId_t transactionId = openVerifyCallTransaction(Uvr::undefinedUserId,
                                                                         idSrc,
                                                                         callingNumber,
                                                                         calledNumber,
                                                                         originalCalledNumber,
                                                                         genericNumber,
                                                                         callId);

        // Add a verification promise, and get a future for it.
        auto verificationFuture = m_verificationPromises.add(transactionId);

        // Send a primitive for verification.
        sendVerifyCallQuery(idSrc,
                            callingNumber,
                            calledNumber,
                            originalCalledNumber,
                            genericNumber,
                            callId,
                            transactionId);

        // Wait for the verification result.
        Uvr::VerificationResult result;
        try
        {
            // It is expected that the VFC will complete the task and in any case
            // return the verdict to us within the timeout. Therefore, here we are
            // waiting for a response without a timeout.
            result = verificationFuture.get();
        }
        catch (const std::exception &e)
        {
            LOGE(Log::Tag(m_ownTypeInst),
                 "Failed to verify the call (callingNumber=",
                 callingNumber,
                 ", calledNumber=",
                 calledNumber,
                 ").");
        }

        return result;
    }

    void node::doNotifyAboutOutgoingCall(Uvr::OctetStringPtr callingNumber,
                                         Uvr::OctetStringPtr calledNumber,
                                         const Uvr::CallId_t &callId,
                                         const Uvr::OptionalOperatorId &idDst)
    {
        if (!TRecQ763Number::Validator::verify(callingNumber))
        {
            LOGW(m_logTag, "Calling number is not valid.");
        }

        if (!TRecQ763Number::Validator::verify(calledNumber))
        {
            LOGW(m_logTag, "Called number is not valid.");
        }

        auto prim = std::make_shared<primitive::OutgoingCallIndication>(callingNumber,
                                                                        calledNumber);
        prim->m_header->to = TypeInst(workers::type::VFS, 0);
        workers::node::send(prim);

        std::vector<std::any> funcs;
        this->m_callbacks.getForAllUsers(storeRecordCallbackId, funcs);
        if (!funcs.empty())
        {
            Uvr::CallRecord record;
            record.numA = TRecQ763Number::toStdString(callingNumber);
            record.numB = TRecQ763Number::toStdString(calledNumber);
            record.numC = "";
            record.numD = "";
            record.date = timers::now();
            record.dateAct = timers::now();
            record.idSrc = std::nullopt;
            record.idDst = idDst;
            record.sessionId = Uvr::undefinedSessionId;
            record.idRel = Uvr::VerdictTrigger::antifraud;
            record.response = Uvr::VerificationResponse::notVerified;
            record.releaseCode = std::nullopt;
            record.callDuration = std::nullopt;
            record.idUvrT = std::nullopt;
            record.callId = callId;
            record.callType = Uvr::CallType::outgoing;

            for (const auto &func : funcs)
            {
                std::any_cast<Uvr::StoreRecordCallback>(func)(m_uvrId, record);
            }
        }
    }

    void node::processCallInformation(const Uvr::CallInformationResponses &responses)
    {
        LOGD(m_ownTypeInst, " processCallInformation with ",
             responses.size(), " responses");
        std::string report_name;
        std::string reportDirPath = m_config.getValue<std::string>(CONNECTIONS_PATH("ccngw", CALL_INFO_RESPONSES),
                                                                   CALL_INFO_RSP_DEFAULT_DIR);
        if (Csv::File::generate(reportDirPath + "/" + Csv::CallInfoResponse::prefix(m_uvrId, responses[0].idReq),
                                Csv::CallInfoResponse::header,
                                responses,
                                report_name))
        {
            auto prim = std::make_shared<primitive::NotifyCallInfoResponse>(report_name);
            prim->m_header->to = TypeInst(workers::type::CCNGW, 0);
            LOGDEBUG("Sending notify connection response to ", prim->m_header->to);
            workers::node::send(prim);
        }
    }

    Uvr::UvrIds node::getUvrIds(const Uvr::OctetStringPtr &number)
    {
        // Open a transaction for verification.
        const Uvr::SessionId_t transactionId = openDigestsTransaction();

        // Add a verification promise, and get a future for it.
        auto uvrIdsFuture = m_uvrIdsPromises.add(transactionId);

        const auto prim = std::make_shared<primitive::QueryUvrIdsReq>(number);
        prim->m_header->to = TypeInst(workers::type::VFC, 0);
        prim->m_header->ulpKey = transactionId;

        workers::node::send(prim);

        // Waiting for reply on the sent request.
        Uvr::UvrIds result = {0, 0};
        try
        {
            // It is expected that the VFC will complete the task and in any case
            // return a response to us within the timeout. Therefore, here we are
            // waiting for the response without a timeout.
            result = uvrIdsFuture.get();
        }
        catch (const std::exception &e)
        {
            LOGE(Log::Tag(m_ownTypeInst), "Failed to get UVR IDs for number = ", number);
        }

        return result;
    }

    Uvr::UvrId_t node::getOwnUvrId()
    {
        return ::workers::node::m_uvrId;
    }

    Uvr::UvrInfo node::getUvrInfo(const Uvr::UvrId_t &uvrId)
    {
        // Open a transaction for verification.
        const Uvr::SessionId_t transactionId = openDigestsTransaction();

        // Add a verification promise, and get a future for it.
        auto uvrFuture = m_uvrInfoPromises.add(transactionId);

        const auto prim = std::make_shared<primitive::QueryUvrInfoReq>(uvrId);
        prim->m_header->to = TypeInst(workers::type::VFGW, 0);
        prim->m_header->ulpKey = transactionId;

        workers::node::send(prim);

        // Waiting for reply on the sent request.
        Uvr::UvrInfo result;
        try
        {
            // It is expected that the VFGW will complete the task and in any case
            // return a response to us within the timeout. Therefore, here we are
            // waiting for the response without a timeout.
            result = uvrFuture.get();
        }
        catch (const std::exception &e)
        {
            LOGE(Log::Tag(m_ownTypeInst), "Failed to get information about UVR with ID = ", uvrId);
        }

        return result;
    }

    Uvr::HubInfo node::getHubInfo(const Uvr::Gt_t &hubId)
    {
        // Open a transaction for verification.
        const Uvr::SessionId_t transactionId = openDigestsTransaction();

        // Add a verification promise, and get a future for it.
        auto hubFuture = m_hubInfoPromises.add(transactionId);

        const auto prim = std::make_shared<primitive::QueryHubInfoReq>(hubId);
        prim->m_header->to = TypeInst(workers::type::VFGW, 0);
        prim->m_header->ulpKey = transactionId;

        workers::node::send(prim);

        // Waiting for reply on the sent request.
        Uvr::HubInfo result;
        try
        {
            // It is expected that the VFGW will complete the task and in any case
            // return a response to us within the timeout. Therefore, here we are
            // waiting for the response without a timeout.
            result = hubFuture.get();
        }
        catch (const std::exception &e)
        {
            LOGE(Log::Tag(m_ownTypeInst), "Failed to get information about HUB with ID = ", hubId);
        }

        return result;
    }

    Uvr::OperatorInfo node::getOperatorInfo(const Uvr::OperatorId_t &operatorId)
    {
        // Open a transaction for getting operator info.
        const Uvr::SessionId_t transactionId = openDigestsTransaction();

        // Add an operator info promise, and get a future for it.
        auto operatorInfoFuture = m_operatorInfoPromises.add(transactionId);

        const auto prim = std::make_shared<primitive::QueryOperatorInfoReq>(operatorId);
        prim->m_header->to = TypeInst(workers::type::VFGW, 0);
        prim->m_header->ulpKey = transactionId;

        workers::node::send(prim);

        // Waiting for reply on the sent request.
        Uvr::OperatorInfo result;
        try
        {
            // It is expected that the VFGW will complete the task and in any case return a
            // response to us, therefore, here we are waiting for the response without a timeout.
            result = operatorInfoFuture.get();
        }
        catch (const std::exception &e)
        {
            LOGE(Log::Tag(m_ownTypeInst), "Failed to get information about Operator with ID = ", operatorId);
        }

        return result;
    }

    Uvr::SessionId_t node::openVerifyCallTransaction(const Uvr::UserId_t userId,
                                                     const Uvr::OperatorId_t idSrc,
                                                     const Uvr::OctetStringPtr &callingNumber,
                                                     const Uvr::OctetStringPtr &calledNumber,
                                                     const Uvr::OctetStringPtr &originalCalledNumber,
                                                     const Uvr::OctetStringPtr &genericNumber,
                                                     const Uvr::CallId_t callId)
    {


        ::transaction::errors openTransactionResult = ::transaction::errors::FAIL;
        return m_verifyCallsTransactionManager.openTransaction(
            ::apigw::transaction::data(userId,
                                       idSrc,
                                       callingNumber,
                                       calledNumber,
                                       originalCalledNumber,
                                       genericNumber,
                                       callId),
            openTransactionResult);
    }

    Uvr::SessionId_t node::openDigestsTransaction()
    {
        ::transaction::errors openTransactionResult = ::transaction::errors::FAIL;
        return m_digestsTransactionManager.openTransaction(
            ::transaction::base_data(),
            openTransactionResult);
    }

    void node::sendVerifyCallQuery(const Uvr::OperatorId_t idSrc,
                                   const Uvr::OctetStringPtr &callingNumber,
                                   const Uvr::OctetStringPtr &calledNumber,
                                   const Uvr::OctetStringPtr &originalCalledNumber,
                                   const Uvr::OctetStringPtr &genericNumber,
                                   const Uvr::CallId_t callId,
                                   const Uvr::SessionId_t transactionId)
    {
        if (!TRecQ763Number::Validator::verify(calledNumber))
        {
            LOGW(m_logTag, "Called number is not valid.");
        }

        if (!TRecQ763Number::Validator::verify(originalCalledNumber, TRecQ763Number::Validator::emptyAllowed))
        {
            LOGW(m_logTag, "Original called number is not valid.");
        }

        if (!TRecQ763Number::Validator::verify(genericNumber, TRecQ763Number::Validator::emptyAllowed))
        {
            LOGW(m_logTag, "Generic number is not valid.");
        }

        const auto prim = std::make_shared<primitive::VerifyCallQuery>(idSrc,
                                                                       callingNumber,
                                                                       TRecQ763Number::toHashedOctetStringPtr(calledNumber),
                                                                       TRecQ763Number::toHashedOctetStringPtr(originalCalledNumber),
                                                                       genericNumber,
                                                                       callId);
        prim->m_header->to = TypeInst(workers::type::VFC, 0);
        prim->m_header->ulpKey = transactionId;

        workers::node::send(prim);
    }

    void node::registerCliCommands()
    {
        auto ind = std::make_shared<::primitive::RegisterCmdInd>(s_cmdIdUpdateRecords);
        ind->m_header->to = TypeInst(workers::type::CLIGW, 0);
        workers::node::send(ind);
    }

} // namespace apigw

namespace
{
    std::string parseRecord(const KeyValue::PairsList &keyValueList, Uvr::CallRecord &record)
    {
        for (size_t pairIndex = 0; pairIndex < keyValueList.size(); ++pairIndex)
        {
            std::string error = "pair " + std::to_string(pairIndex) + ", ";

            const auto &keyValue = keyValueList.at(pairIndex);

            auto key = keyValue.first;
            std::transform(key.begin(), key.end(), key.begin(), ::toupper);

            const auto &value = keyValue.second;

            if (key == "")
            {
                error += "error: Key is empty.";
                return error;
            }

            if (value == "")
            {
                error += "error: Value is empty.";
                return error;
            }

            try
            {
                if (key == s_recordKeyIdSrc)
                {
                    record.idSrc = boost::lexical_cast<Uvr::OperatorId_t>(value);
                }
                else if (key == s_recordKeyIdDst)
                {
                    record.idDst = boost::lexical_cast<Uvr::OperatorId_t>(value);
                }
                else if (key == s_recordKeyReleaseCode)
                {
                    record.releaseCode = boost::numeric_cast<Uvr::ReleaseCode_t>(boost::lexical_cast<int16_t>(value));
                }
                else if (key == s_recordKeyDuration)
                {
                    record.callDuration = boost::lexical_cast<Uvr::CallDuration_t>(value);
                }
                else if (key == s_recordKeyCallId)
                {
                    record.callId = value;
                }
                else if (key == s_recordKeyCallType)
                {
                    if (value == s_callTypeValueIncoming)
                    {
                        record.callType = Uvr::CallType::incoming;
                    }
                    else if (value == s_callTypeValueOutgoing)
                    {
                        record.callType = Uvr::CallType::outgoing;
                    }
                    else
                    {
                        error += "error: Unknonw value=\"" + value + "\" for key \"" + key + "\".";
                        return error;
                    }                    
                }
                else
                {
                    error += "error: Unknown key \"" + keyValue.first + "\".";
                    return error;
                }
            }
            catch (...)
            {
                error += "error: Invalid value \"" + value + "\" for key \"" + keyValue.first + "\".";
                return error;
            }
        }

        return "";
    }

} // namespace
