// Own include.
#include "Node.hpp"
#include "Config.hpp"

// AFAP includes.
#include "components/tools/src/Tools.hpp"
#include "components/config/src/config_manager.hpp"

namespace
{

    std::string s_cliPort;
    std::string s_cliInterface;

} // namespace

namespace CliGw
{
    Node::Node(const TypeInst &ownInst,
               boost::asio::io_context &ioContext,
               const configuration::manager &configManager)
        : workers::node(ownInst, ioContext, configManager)
    {
        m_connectionManager = std::make_shared<connections::manager>(ownInst, ioContext);
        m_connectionManager->::connections::interfaces::handler::client::bind(this);

        s_cliPort = configManager.getValue<std::string>(config::port, Log::Level::Info);
        s_cliInterface = configManager.getValue<std::string>(config::interface, Log::Level::Info);
    }

    void Node::start()
    {
        if (!s_cliPort.empty() && !s_cliInterface.empty())
        {
            LOGD(m_logTag, "CLI interface=\"", s_cliInterface, "\", port=\"", s_cliPort, "\".");
            m_connectionManager->start(s_cliInterface, s_cliPort);
        }

        workers::node::start();
    }

    void Node::stop()
    {
        m_connectionManager->stop();
        workers::node::stop();
    }

    void Node::notifyPrimitive(const primitive::GenericPtr &prim)
    {
        handleSpecificPrimitive(prim);
    }

    void Node::handleSpecificPrimitive(const primitive::GenericPtr &prim)
    {
        switch (prim->m_header->id)
        {
            case primitive::REGISTER_CMD_IND:
            {
                handleRegisterCmdInd(prim);
                break;
            }
            case primitive::EXEC_CMD_REQ:
            {
                handleExecCmdReq(prim);
                break;
            }
            case primitive::EXEC_CMD_RSP:
            {
                handleExecCmdRsp(prim);
                break;
            }
            default:
            {
                workers::node::handleSpecificPrimitive(prim);
            }
        }
    }

    void Node::handleRegisterCmdInd(const primitive::GenericPtr &prim)
    {
        const auto ind = std::dynamic_pointer_cast<primitive::RegisterCmdInd>(prim);

        std::unique_lock cmdsMapLock(m_cmdsMapMutex);
        m_cmdsMap.emplace(ind->cmd, ind->m_header->from);
    }

    void Node::handleExecCmdReq(const primitive::GenericPtr &prim)
    {
        const auto req = std::dynamic_pointer_cast<primitive::ExecCmdReq>(prim);

        {
            std::shared_lock cmdsMapLock(m_cmdsMapMutex);
            const auto cmdsIterator = m_cmdsMap.find(req->cmd);
            if (cmdsIterator != m_cmdsMap.end())
            {
                prim->m_header->to = cmdsIterator->second;
                workers::node::send(prim);
                return;
            }
        }

        Uvr::Strings responses;
        responses.emplace_back("Error: The command \"" + req->cmd + "\" does not exist.");
        // TODO[HUB-350]: Print how to check the help or all available commands.

        auto msg = std::make_shared<::primitive::ExecCmdRsp>(std::move(responses));
        msg->m_header->to = req->m_header->from;
        msg->m_header->ulpKey = req->m_header->ulpKey;
        m_connectionManager->send(msg);

        LOGE(m_logTag, "Command \"", req->cmd, "\" is not registered by any node.");
    }

    void Node::handleExecCmdRsp(const primitive::GenericPtr &prim)
    {
        prim->m_header->to = TypeInst(workers::type::CLI, 0);
        m_connectionManager->send(prim);
    }

} // namespace CliGw
