#pragma once

#include <boost/asio.hpp>
#include "worker_node/src/primitive.hpp"
#include "support/interfaces_template.hpp"

#ifndef MAX_CONN_BUFFER_SIZE
#define MAX_CONN_BUFFER_SIZE 2048
#endif

namespace connections
{
    using tcp = boost::asio::ip::tcp;
    using tcp_socket = tcp::socket;
    using tcp_socket_ptr = std::shared_ptr<tcp_socket>;
    class connection;
    using connection_ptr = std::shared_ptr<connection>;
    using Id = uint32_t;
    using buffer = std::array<char, MAX_CONN_BUFFER_SIZE>;

    enum type
    {
        INCOMING,
        OUTGOING_STATICIP,
        OUTGOING_DYNAMICIP
    };

    enum state
    {
        CONNECTION_INIT,
        CONNECTION_STARTED,
        CONNECTION_DOWN,
        CONNECTION_WAIT_WHOAMI,
        CONNECTION_SEND_WHOAMI,
        CONNECTION_ESTBLISHED
    };

    namespace interfaces
    {

        INTERFACE_S(controller,
                    INTERFACE_SERVER(void, handleMessage, connection_ptr, const primitive::raw_data_t &);
                    INTERFACE_SERVER(void, handle_brokenConnection, const TypeInst &);
                    INTERFACE_SERVER(void, handle_connectionClosed, const TypeInst &);)
        INTERFACE_C1(controller,
                     INTERFACE_CLIENT1(void, handleMessage, connection_ptr, const primitive::raw_data_t &);
                     INTERFACE_CLIENT1(void, handle_brokenConnection, const TypeInst &);
                     INTERFACE_CLIENT1(void, handle_connectionClosed, const TypeInst &);)
    } // namespace interfaces

    class connection : public std::enable_shared_from_this<connection>,
                       public interfaces::controller::client1
    {
    public:
        // Create incoming connection
        static connection_ptr createIncomingConnection(boost::asio::io_context &ioc,
                                                       const TypeInst &owner,
                                                       int id);
        // Create outgoing connection
        static connection_ptr createOutgoingConnection(boost::asio::io_context &ioc,
                                                       const TypeInst &owner,
                                                       const TypeInst &inst,
                                                       connections::type type,
                                                       const std::string &address,
                                                       const std::string &port,
                                                       int id);

        // Get the socket associated with the connection.
        tcp_socket_ptr socket();
        // Get reference to connection buffer
        connections::buffer &buffer();
        // Get connection type
        connections::type type() const;
        // Get connection state
        connections::state state() const;
        // Get remote instance-type
        const TypeInst &peerInstance() const;
        // Get remote IP address, empty string is returned in case OUTGOING connections
        const std::string &peerAddress() const;
        void peerAddress(const std::string &address);
        // Get remote port, empty string is returned in case OUTGOING connections
        const std::string &peerPort() const;
        void peerPort(const std::string &port);
        // Get remote port, empty string is returned in case OUTGOING connections
        int id() const;

        // Set peer instance when we get it
        void set_peerInstance(TypeInst inst);
        // Change connection state
        void change_state(connections::state state);

        void start_wait();
        void send(const primitive::raw_data_t &msg);
        void disconnect();
        void reset();

        friend std::ostream &operator<<(std::ostream &os, const connection &c);

    private:
        explicit connection(boost::asio::io_context &ioc,
                            connections::type type,
                            const TypeInst &owner,
                            int connId);
        explicit connection(boost::asio::io_context &ioc,
                            connections::type type,
                            const TypeInst &owner,
                            const TypeInst &inst,
                            std::string address, std::string port,
                            int connId);

        void readMessageSize(const boost::system::error_code &errorCode,
                             std::size_t bytesTransferred);

        void readMessage(const boost::system::error_code &errorCode,
                         std::size_t bytesTransferred,
                         uint16_t messageSize);

        void disconnect_int();

        // Connection type
        connections::type m_type;
        // Own instance-type
        TypeInst m_owner;
        // Remote instance-type
        TypeInst m_peerInst;
        // Connection state
        connections::state m_state;
        // connection ID
        int m_connId;
        // Buffer for incoming data.
        connections::buffer m_buffer;
        // Remote address (applicable for outgoing connections only)
        std::string m_peerAddress;
        // Remote port (applicable for outgoing connections only)
        std::string m_peerPort;
        // TCP socket for connection
        tcp_socket_ptr m_socket;

        boost::asio::strand<boost::asio::io_context::executor_type> m_strand;
    };

    std::string connType2str(connections::type type);
    std::string connState2str(connections::state state);
    std::ostream &operator<<(std::ostream &os, const connection &c);
}
