#include "connection_manager.hpp"
#include "components/tools/src/Tools.hpp"

#include <iostream>
#include <boost/bind/bind.hpp>

namespace connections
{
    manager::manager(const TypeInst &nodeInst,
                     boost::asio::io_context &ioc)
        : m_nodeInst(nodeInst),
          m_ioc(ioc),
          m_acceptor(m_ioc)
    {
    }

    void manager::start(const std::string &listener,
                        const std::string &port)
    {
        listenConnections(listener, port);
    }

    void manager::stop()
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);

        connections::interfaces::controller::server::stop_server();
        stop_accept();

        m_instToConnectionMap.clear();
        for (auto const &connIter : m_connections)
        {
            connIter.second->disconnect();
        }
        m_connections.clear();
    }

    connection_ptr manager::getConnection(TypeInst inst) const
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto connectionIter = m_instToConnectionMap.find(inst);
        if (connectionIter == m_instToConnectionMap.end())
        {
            return nullptr;
        }
        return connectionIter->second;
    }

    connection_ptr manager::getConnection(const Id &id) const
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto connectionIter = m_connections.find(id);
        if (connectionIter == m_connections.end())
        {
            return nullptr;
        }
        return connectionIter->second;
    }

    void manager::listenConnections(const std::string &listener,
                                    const std::string &port)
    {
        boost::system::error_code err;
        network::address ipAddress =
            network::make_address(listener, AF_INET, err);
        if (err)
        {
            return;
        }

        boost::asio::ip::tcp::endpoint endpoint(ipAddress,
                                                (boost::asio::ip::port_type)std::stoi(port));

        m_acceptor.open(endpoint.protocol(), err);
        if (err)
        {
            LOGERROR(m_nodeInst,
                     " Cannot open socket[", ipAddress, ":", port, "]: ",
                     err.message());
            return;
        }

        m_acceptor.set_option(boost::asio::ip::tcp::acceptor::reuse_address(true), err);
        if (err)
        {
            LOGERROR(m_nodeInst,
                     " Cannot set REUSE_ADDRESS option for socket[", ipAddress, ":", port, "]: ",
                     err.message());
            return;
        }

        m_acceptor.bind(endpoint, err);
        if (err)
        {
            LOGERROR(m_nodeInst,
                     " Cannot bind to socket[", ipAddress, ":", port, "]: ",
                     err.message());
            return;
        }

        m_acceptor.listen(boost::asio::socket_base::max_listen_connections, err);
        if (err)
        {
            LOGERROR(m_nodeInst,
                     " Cannot listen to socket[", ipAddress, ":", port, "]: ",
                     err.message());
            return;
        }

        m_stopped = false;
        start_accept();
    }

    void manager::send(primitive::GenericPtr prim) const
    {
        auto connectionPtr = getConnection(prim->m_header->to);
        if (connectionPtr == nullptr)
        {
            LOGERROR(m_nodeInst,
                     " Cannot find connection to send message with id: ", prim->m_header->id,
                     " to: ", prim->m_header->to);
            return;
        }

        prim->m_header->from = m_nodeInst;
        primitive::raw_data_t msg = primitive::Codec::encode(prim.get(), primitive::Codec::Type::Json, true);
        connectionPtr->send(msg);
    }

    void manager::send(const Id &id, primitive::GenericPtr prim) const
    {
        auto connectionPtr = getConnection(id);
        if (connectionPtr == nullptr)
        {
            LOGERROR(m_nodeInst,
                     " Cannot find connection to send message with id: ", prim->m_header->id,
                     " to: ", prim->m_header->to);
            return;
        }

        prim->m_header->from = m_nodeInst;
        primitive::raw_data_t msg = primitive::Codec::encode(prim.get(), primitive::Codec::Type::Json, true);
        connectionPtr->send(msg);
    }

    void manager::handle_brokenConnection(const TypeInst &inst)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);

        auto instToConnectionIter = m_instToConnectionMap.find(inst);
        if (instToConnectionIter == m_instToConnectionMap.end())
        {
            LOGINFO(m_nodeInst,
                    " Can't find broken connection to ", inst, " in TypeInst to Connection map.");
            return;
        }

        auto connectionPtr = instToConnectionIter->second;
        if (connectionPtr->type() != INCOMING)
        {
            return;
        }

        m_instToConnectionMap.erase(instToConnectionIter);

        auto idToConnectionIter = m_connections.find(connectionPtr->id());
        if (idToConnectionIter == m_connections.end())
        {
            LOGINFO(m_nodeInst,
                    " Can't find broken connection to ", inst, " in Id to Connection map.");
            return;
        }

        m_connections.erase(idToConnectionIter);
    }

    void manager::handle_connectionClosed(const TypeInst &inst)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);

        auto instToConnectionIter = m_instToConnectionMap.find(inst);
        if (instToConnectionIter == m_instToConnectionMap.end())
        {
            LOGINFO(m_nodeInst,
                    " Can't find closed connection to ", inst, " in TypeInst to Connection map.");
            return;
        }

        auto connectionPtr = instToConnectionIter->second;
        if (connectionPtr->type() != INCOMING)
        {
            return;
        }

        m_instToConnectionMap.erase(instToConnectionIter);

        auto idToConnectionIter = m_connections.find(connectionPtr->id());
        if (idToConnectionIter == m_connections.end())
        {
            LOGINFO(m_nodeInst,
                    " Can't find closed connection to ", inst, " in Id to Connection map.");
            return;
        }

        m_connections.erase(idToConnectionIter);
    }

    void manager::start_accept()
    {
        connection_ptr next_conn;
        next_conn = connection::createIncomingConnection(m_ioc, m_nodeInst, m_connIdCntr);
        m_connIdCntr++;
        next_conn->interfaces::controller::client1::bind(shared_from_this());
        next_conn->reset();
        next_conn->change_state(CONNECTION_STARTED);
        m_acceptor.async_accept(*(next_conn->socket()),
                                boost::bind(&manager::on_accept, this,
                                            next_conn,
                                            boost::asio::placeholders::error));
    }

    void manager::stop_accept()
    {
        boost::system::error_code ec;
        m_acceptor.close(ec);
        if (ec)
        {
            LOGERROR(m_nodeInst, " Failed to close acceptor: ", ec.message());
        }
        m_stopped = true;
    }

    void manager::on_accept(connection_ptr conn_ptr,
                            const boost::system::error_code &err)
    {
        // Stop listen to new incoming connections
        // if user wants to destroy all connections
        if (m_stopped)
            return;

        if (!err)
        {
            std::unique_lock<std::recursive_mutex> guard(m_guard);
            m_connections.emplace(conn_ptr->id(), conn_ptr);

            conn_ptr->change_state(CONNECTION_WAIT_WHOAMI);
            conn_ptr->start_wait();
        }
        else if (err != boost::asio::error::operation_aborted)
        {
            LOGERROR(m_nodeInst,
                     " Cannot accept connection: ", err.message());
        }
        else
        {
            return;
        }

        start_accept();
    }

    void manager::handle_whoami(connection_ptr conn_ptr, std::shared_ptr<primitive::generic> prim)
    {
        std::unique_lock<std::recursive_mutex> guard(m_guard);

        TypeInst inst = prim->m_header->from;
        if (conn_ptr->state() != CONNECTION_WAIT_WHOAMI)
        {
            LOGDEBUG(m_nodeInst,
                     " Received ", primitive::id2str(prim->m_header->id),
                     " in wrong state for ", *conn_ptr, " from ", inst,
                     ". Ignore it.");
            return;
        }

        conn_ptr->set_peerInstance(inst);

        auto instMapResult = m_instToConnectionMap.emplace(inst, conn_ptr);
        if (!instMapResult.second)
        {
            conn_ptr->disconnect();
            LOGERROR(m_nodeInst, " The connection towards ", inst, " already exists in TypeInst to Connection map.");
            return;
        }

        conn_ptr->change_state(CONNECTION_ESTBLISHED);

        LOGINFO(m_nodeInst, " Incoming connection up with ", conn_ptr->peerInstance());
    }

    void manager::handleMessage(connection_ptr conn_ptr, const primitive::raw_data_t &message)
    {
        if (message.empty())
        {
            return;
        }

        auto prim = primitive::Codec::decode(message);
        if (prim == nullptr)
        {
            return;
        }

        if (prim->m_header->id == primitive::WHOAMI)
        {
            handle_whoami(conn_ptr, prim);
        }
        else
        {
            if (conn_ptr->state() < CONNECTION_ESTBLISHED)
            {
                LOGERROR(m_nodeInst,
                         " Wrong connection state (expected state=ESTABLISHED): ", *conn_ptr);
                return;
            }
            interfaces::handler::client::notifyPrimitive(prim);
        }
    }
}
