#pragma once

// AFAP includes.
#include "components/log/src/Log.hpp"
#include "components/timers/src/timer_types.hpp"

// Boost includes.
#include <boost/json.hpp>

// STL includes.
#include <mutex>
#include <ostream>
#include <optional>

namespace timers
{
    Duration tag_invoke(boost::json::value_to_tag<Duration>,
                        boost::json::value const &jv);
} // namespace timers

namespace configuration
{
    using ConfigFileVersion_t = uint16_t;

    enum class errors
    {
        OK,
        FAIL
    };

    std::ostream &operator<<(std::ostream &os, const errors &err);

    ConfigFileVersion_t getConfigFileVersion();

    class manager
    {
    public:
        errors loadConfigFile(const std::string &filename);
        errors validateConfigParams();

        template <class T>
        T getValue(std::string const &path,
                   Log::Level logLevel = Log::Level::Error) const;

        template <class T>
        T getValue(std::string const &path,
                   const T &defaultValue,
                   Log::Level logLevel = Log::Level::Info) const;

        template <class T>
        std::vector<T> getList(std::string const &path) const
        {
            std::vector<T> result;
            std::lock_guard<std::recursive_mutex> guard(m_guard);
            boost::json::value val = getJsonValue(path);
            try
            {
                result = boost::json::value_to<std::vector<T>>(val);
            }
            catch (const std::exception &e)
            {
                LOGERROR("Cannot get list for path=\"", path, "\" due to exception: ", e.what());
            }
            return result;
        }

    protected:
        // Protector of shared data in multiple threads
        mutable std::recursive_mutex m_guard;
        // Parsed configuration into JSON object
        boost::json::value m_json_config;

    private:
        boost::json::value getJsonValue(const std::string &path) const;
    };

    using ManagerPtr = std::shared_ptr<manager>;

    namespace utils
    {
        timers::duration convertToDuration(std::optional<int> duration,
                                           std::optional<std::string> unit);
    } // namespace utils

} // namespace configuration

//------------------------------------------------------------------------------
// Definitions
//------------------------------------------------------------------------------

namespace configuration
{
    template <class T>
    T manager::getValue(const std::string &path,
                        Log::Level logLevel) const
    {
        T value = T();
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        boost::json::value jsonValue = getJsonValue(path);
        try
        {
            value = boost::json::value_to<T>(jsonValue);
        }
        catch (const std::exception &e)
        {
            LOG(logLevel, "Cannot get value for path=\"", path, "\" due to exception: ", e.what());
        }
        return value;
    }

    template <class T>
    T manager::getValue(std::string const &path,
                        const T &defaultValue,
                        Log::Level logLevel) const
    {
        T value = T();
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        boost::json::value jsonValue = getJsonValue(path);
        try
        {
            value = boost::json::value_to<T>(jsonValue);
        }
        catch (const std::exception &e)
        {
            LOG(logLevel, "Can't get value for path=\"", path, "\" due to exception: ", e.what(),
                ". Default value=\"", defaultValue, "\" will be used.");
            return defaultValue;
        }
        return value;
    }
} // namespace configuration
