#include "mimic.hpp"
#include "components/log/src/Log.hpp"
#include "components/csv/src/CsvFile.hpp"

namespace db
{
    namespace stub
    {
        mimic::mimic(Uvr::UvrId_t uvrId) : m_uvrId(uvrId)
        {
            this->db::interfaces::tables::incidents::server::tableName = db::incident::tableName;
            this->db::interfaces::tables::subscribers::server::tableName = db::subscriber::tableName;
            this->db::interfaces::tables::uvrs::server::tableName = db::uvr::tableName;
            this->db::interfaces::tables::hubs::server::tableName = db::hub::tableName;
            this->db::interfaces::tables::Operators::server::tableName = db::Operator::tableName;
        }

        bool mimic::connect(void)
        {
            return true;
        }

        void mimic::recreate(const ::db::table::name &tbName,
                             const ::db::table::fields &tbFields,
                             const ::db::table::settings &tbSettings)
        {
            if (tbName == this->db::interfaces::tables::uvrs::server::tableName)
            {
                m_vfnInfoPtr.reset();

                std::unique_lock uvrsByIdLock(m_uvrsByIdMutex);
                std::unique_lock uvrsByIpLock(m_uvrsByIpMutex);
                m_uvrsById.clear();
                m_uvrsByIp.clear();
            }
            else if (tbName == this->db::interfaces::tables::hubs::server::tableName)
            {
                std::unique_lock hubsLock(m_hubsMutex);
                m_hubs.clear();
            }
            else if (tbName == this->db::interfaces::tables::Operators::server::tableName)
            {
                std::unique_lock operatorsLock(m_operatorsMutex);
                m_operators.clear();
            }
            else
            {
                LOGERROR("The recreate is not implemented for \"" + tbName + "\" table");
            }
        }

        void mimic::fill_in(const ::db::table::name &tbName,
                            const std::string &filename)
        {
            Csv::File::parse(filename,
                             [this, &tbName](const std::string &line)
                             {
                                 if (tbName == this->db::interfaces::tables::uvrs::server::tableName)
                                 {
                                     uvr::RecordPtr recPtr = std::make_shared<::db::uvr::record>(line.c_str());
                                     if (recPtr->m_idUvr == this->m_uvrId)
                                     {
                                         this->m_vfnInfoPtr = recPtr;
                                     }
                                     std::unique_lock uvrsByIdLock(m_uvrsByIdMutex);
                                     std::unique_lock uvrsByIpLock(m_uvrsByIpMutex);
                                     this->m_uvrsById.emplace(recPtr->m_idUvr, recPtr);
                                     // Stores pointer to record mapped by primary and secondary UVR IPs
                                     this->m_uvrsByIp.emplace(recPtr->m_ipUvrP, recPtr);
                                     if (not recPtr->m_ipUvrS.empty())
                                         this->m_uvrsByIp.emplace(recPtr->m_ipUvrS, recPtr);
                                 }
                                 else if (tbName == this->db::interfaces::tables::hubs::server::tableName)
                                 {
                                     ::db::hub::record rec(line.c_str());
                                     std::unique_lock hubsLock(m_hubsMutex);
                                     this->m_hubs.emplace(rec.m_idHub, rec);
                                 }
                                 else if (tbName == this->db::interfaces::tables::Operators::server::tableName)
                                 {
                                     ::db::Operator::Record rec(line.c_str());
                                     std::unique_lock operatorsLock(m_operatorsMutex);
                                     this->m_operators.emplace(rec.idSrc, rec);
                                 }
                                 else
                                 {
                                     LOGERROR("The fill-in is not implemented for \"" + tbName + "\" table");
                                     return true;
                                 }
                                 return false;
                             });
        }

        ::db::limit_t mimic::size(const ::db::table::name &tbName)
        {
            if (tbName == this->db::interfaces::tables::incidents::server::tableName)
            {
                return m_incidents.size();
            }
            else if (tbName == this->db::interfaces::tables::subscribers::server::tableName)
            {
                return 0;
            }
            else if (tbName == this->db::interfaces::tables::uvrs::server::tableName)
            {
                std::unique_lock uvrsByIdLock(m_uvrsByIdMutex);
                return m_uvrsById.size();
            }
            else if (tbName == this->db::interfaces::tables::hubs::server::tableName)
            {
                std::unique_lock hubsLock(m_hubsMutex);
                return m_hubs.size();
            }
            else if (tbName == this->db::interfaces::tables::Operators::server::tableName)
            {
                std::unique_lock operatorsLock(m_operatorsMutex);
                return m_operators.size();
            }

            LOGERROR("Not implemented for table=", tbName);
            return 0;
        }

        bool mimic::is_exist(const ::db::table::name &tbName)
        {
            return true;
        }

        void mimic::get(::db::incident::rows &rows,
                        const timers::time_point &first,
                        const timers::time_point &last)
        {
            rows.clear();
            std::unique_lock<std::recursive_mutex> m_guard(m_incidentsGuard);
            for (auto it = m_incidents.begin(); it != m_incidents.end(); it++)
            {
                rows.push_back(*it);
            }
            m_incidents.clear();
        }

        void mimic::put(const ::db::incident::record &rec)
        {
            std::unique_lock<std::recursive_mutex> m_guard(m_incidentsGuard);
            m_incidents.push_back(rec);
        }

        void mimic::get(::db::incidentA::rows &rows,
                        const timers::time_point &first,
                        const timers::time_point &last)
        {
            rows.clear();
            std::unique_lock<std::recursive_mutex> m_guard(m_incidentsAGuard);
            for (auto it = m_incidentsA.begin(); it != m_incidentsA.end(); it++)
            {
                rows.push_back(*it);
            }
            m_incidentsA.clear();
        }

        void mimic::put(const ::db::incidentA::record &rec)
        {
            std::unique_lock<std::recursive_mutex> m_guard(m_incidentsAGuard);
            m_incidentsA.push_back(rec);
        }

        bool mimic::get(::db::uvr::record &rec,
                        const Uvr::UvrId_t &uvrId)
        {
            std::unique_lock uvrsByIdLock(m_uvrsByIdMutex);
            auto uvrIter = m_uvrsById.find(uvrId);
            if (uvrIter == m_uvrsById.end())
            {
                LOGERROR("Cannot find UVR by id=", uvrId);
                return false;
            }

            rec = *(uvrIter->second);
            return true;
        }

        void mimic::get(::db::uvr::rows &rows,
                        const ::Uvr::Ip_t &ipAddress)
        {
            std::unique_lock uvrsByIpLock(m_uvrsByIpMutex);
            rows.clear();
            auto recRange = m_uvrsByIp.equal_range(ipAddress);
            for (auto recIter = recRange.first; recIter != recRange.second; ++recIter)
            {
                rows.push_back(*(recIter->second));
            }
        }

        void mimic::get(::db::hub::rows &rows,
                        const ::db::gts_t &list)
        {
            std::unique_lock hubsLock(m_hubsMutex);
            rows.clear();
            for (auto const &hubId : list)
            {
                auto hubIter = m_hubs.find(hubId);
                if (hubIter == m_hubs.end())
                {
                    LOGERROR("Cannot find HUB by id=", hubId, " in parsed HUB file");
                    continue;
                }
                rows.push_back(hubIter->second);
            }
        }

        void mimic::get(::db::Operator::Rows &rows,
                        const ::db::OperatorsIds &list)
        {
            rows.clear();
            for (const auto &operatorId : list)
            {
                auto operatorIter = m_operators.find(operatorId);
                if (operatorIter == m_operators.end())
                {
                    LOGERROR("Cannot find Operator by id=", operatorId, " in parsed Operators file");
                    continue;
                }
                rows.push_back(operatorIter->second);
            }
        }

    } // namespace stub

} // namespace db
