#include "filesystem.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"

// STL includes.
#include <fstream>
#include <filesystem>

#define MAX_OUTPUT_LINE_SIZE 128

bool fs::command::list(TypeInst nodeId, const std::string &lsCmd, fs::files::list &output)
{
    LOGDEBUG("Execute command '", lsCmd, "'");
    std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(lsCmd.c_str(), "r"), &pclose);
    std::array<char, MAX_OUTPUT_LINE_SIZE> buffer;
    if (!pipe)
    {
        LOGERROR(nodeId,
                 " Cannot execute command=\"", lsCmd, "\": failed to open pipe");
        return false;
    }
    while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr)
    {
        std::string line(buffer.data());
        // Excluding echo of the command from output
        if (line.find("ls ") != std::string::npos)
            continue;
        std::size_t pos = line.find_last_of('/');
        if (pos == std::string::npos)
        {
            output.emplace_back(line.c_str(), line.length() - 1);
        }
        else // ignore path
        {
            output.emplace_back(line.c_str(), pos + 1, line.length() - pos - 2);
        }
    }
    std::sort(output.begin(), output.end(), std::less<std::string>());
    return true;
}

timers::time_point fs::utils::getTimestampFromFilename(const std::string &filename)
{
    if (filename.empty())
        return timers::time_point(); // return zero time point

    // The following format file is expected:
    // PREFIX_YYYY_MM_DD_HH_MM_SS[.suffix].extension
    std::size_t posBeg = filename.find_first_of("_");
    std::size_t posEnd = filename.find_first_of(".");
    if (posBeg == std::string::npos || posEnd == std::string::npos)
    {
        LOGWARN("Unexpected file format: filename=\"", filename, "\"");
        return timers::time_point();
    }
    std::string fileTs(filename.c_str(), posBeg + 1, posEnd - posBeg - 1);
    return timers::time_point(fileTs, "%Y_%m_%d_%H_%M_%S");
}

void fs::utils::filterFilesByTimestamp(const timers::time_point &ts,
                                       fs::files::list &inList,
                                       fs::files::list &outList)
{
    // reverse order: go from newest to oldest
    std::size_t pos = outList.size();
    for (auto fileIter = inList.rbegin(); fileIter != inList.rend(); fileIter++)
    {
        timers::time_point fileTp(getTimestampFromFilename(*fileIter));
        if (ts < fileTp)
            outList.insert(outList.begin() + pos, *fileIter); // to save right order: from oldest to newest
        else
            break; // find first file with older or equal timestamp, the files after it should be older
    }
}

bool fs::utils::findNewFiles(const fs::files::list &remList,
                             const fs::files::list &locList,
                             fs::files::list &diff)
{
    bool isNewFileFound = false;
    for (auto remIter = remList.begin(); remIter != remList.end(); remIter++)
    {
        auto locIter = locList.begin();
        for (; locIter != locList.end(); locIter++)
        {
            if (*remIter == *locIter)
                break;
        }
        if (locIter == locList.end())
        {
            diff.push_back(*remIter);
            isNewFileFound = true;
        }
    }
#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
    for (auto filename : diff)
    {
        LOGDEBUG("New file: ", filename);
    }
#endif
    return isNewFileFound;
}

fs::local::local(TypeInst nodeId) : m_nodeId(nodeId)
{
}

bool fs::local::readFile(const std::string &filename,
                         files::list &context)
{
    std::ifstream inFile(filename);
    if (!inFile.is_open())
    {
        LOGERROR("Can't open the file=\"", filename, "\"");
        return false;
    }

    std::string line;
    while (getline(inFile, line))
    {
        context.push_back(line);
    }
    inFile.close();
    return true;
}

bool fs::local::zipFile(const std::string &inFile, const std::string &outFile) const
{
    std::string cmd("zip -q -m -j " + outFile + " " + inFile);
    LOGD("zipFile: ", cmd);
    int result = system(cmd.c_str());
    if (result != 0)
    {
        LOGERROR(m_nodeId,
                 " Failed to ZIP file=\"", inFile, "\": result=", result, ", command=\"", cmd, "\"");
        return false;
    }
    return true;
}

bool fs::local::unzipFile(const std::string &inFile, std::string &outFile, const std::string &ext) const
{
    outFile.clear();
    std::string cmd;
    std::filesystem::path inFileFsPath(inFile);
    std::filesystem::path inFileParentPath = inFileFsPath.parent_path();

    if (inFileParentPath.empty())
    {
        cmd = std::string("unzip -q -jo ") + inFile;
    }
    else
    {
        cmd = std::string("unzip -q -jo -d ") + inFileParentPath.c_str() + " " + inFile;
    }

    int result = system(cmd.c_str());
    if (result != 0)
    {
        LOGERROR(m_nodeId,
                 " Failed to UNZIP file=\"", inFile, "\": result=", result, ", command=\"", cmd, "\"");
        return false;
    }

    // TODO Find out in the scope of HUB-414, if renaming of unzipped file is needed or not.
    outFile = inFileFsPath.replace_extension(std::filesystem::path(ext)).c_str();
    return true;
}

bool fs::local::copyFile(const std::string &srcFile,
                         const std::string &dstFile) const
{
    std::string cmd("cp -f " + srcFile + " " + dstFile);
    int result = system(cmd.c_str());
    if (result != 0)
    {
        LOGERROR(m_nodeId,
                 " Failed to copy file: result=", result, ", command=\"", cmd, "\"");
        return false;
    }
    return true;
}

bool fs::local::deleteFile(const std::string &filename) const
{
    std::string cmd("rm -f " + filename);
    int result = system(cmd.c_str());
    if (result != 0)
    {
        LOGERROR(m_nodeId,
                 " Failed to delete file: result=", result, ", command=\"", cmd, "\"");
        return false;
    }
    return true;
}

int fs::local::compareFiles(const std::string &filename1, const std::string &filename2) const
{
    std::string cmd("diff -q " + filename1 + " " + filename2 + " 2> /dev/null");
    int result = system(cmd.c_str());
    if (result != 0)
    {
        LOGDEBUG(m_nodeId,
                 " Files are not equal: result=", result, ", command=\"", cmd, "\"");
    }
    return result;
}

bool fs::local::listDir(const std::string &dirOrMask, files::list &files) const
{
    std::string cmd("ls -1 " + dirOrMask + " 2> /dev/null");
    return command::list(m_nodeId, cmd, files);
}

void fs::local::getLatestFile(const std::string &dirOrMask, std::string &filename) const
{
    filename.clear();
    files::list files;
    if (!listDir(dirOrMask, files))
        return;
    // Last file in alphabetic order should be with newest timestamp
    if (!files.empty())
        filename = files.back();
}

bool fs::local::createDir(const std::string &dir) const
{
    std::string cmd("mkdir -p " + dir);
    int result = system(cmd.c_str());
    if (result != 0)
    {
        LOGERROR(m_nodeId,
                 " Failed to create directory: result=", result, ", command=\"", cmd, "\"");
        return false;
    }
    return true;
}

bool fs::local::deleteDir(const std::string &dir) const
{
    std::string cmd("rm -rf " + dir);
    int result = system(cmd.c_str());
    if (result != 0)
    {
        LOGERROR(m_nodeId,
                 " Failed to delete directory: result=", result, ", command=\"", cmd, "\"");
        return false;
    }
    return true;
}
