#pragma once

// AFAP includes.
#include "components/timers/src/timer_types.hpp"
#include "components/worker_node/src/typeInst.hpp"

// STL includes.
#include <string>
#include <vector>

namespace fs
{
    namespace files
    {
        using list = std::vector<std::string>;

        enum
        {
            match = 0,
            not_match = 256,
            not_exist = 512
        };
    } // namespace files

    namespace command
    {
        bool list(TypeInst nodeId, const std::string &lsCmd, files::list &output);
    } // namespace command

    namespace utils
    {
        /// @brief Gets timestamp from filename
        /// @param[in] filename The filename with timestamp
        /// @return Timestamp as @ref timers::time_point
        timers::time_point getTimestampFromFilename(const std::string &filename);

        /// @brief Filters the files inside the passed @a inList, which contain
        ///        timestamp in its name higher of the specified @a ts timestamp,
        ///        and puts them to @a outList.
        /// @param[in]  ts      The specified timestamp.
        /// @param[in]  inList  The input files list to be filtered.
        ///                     The list is sorted A->Z after the function call.
        /// @param[out] outList The result list with filtered files.
        void filterFilesByTimestamp(const timers::time_point &ts,
                                    fs::files::list &inList,
                                    fs::files::list &outList);

        /// @brief Finds new files for local list regarding remote list.
        /// @param[in]  remList The list of remote files.
        /// @param[in]  locList The list of local files.
        /// @param[out] diff    The files, which are absent in local list.
        /// @returns true if has new files, otherwise false.
        bool findNewFiles(const fs::files::list &remList,
                          const fs::files::list &locList,
                          fs::files::list &diff);

    } // namespace utils

    class local
    {
        TypeInst m_nodeId;

    public:
        explicit local(TypeInst nodeId);

        bool readFile(const std::string &filename,
                      files::list &context);
        /// @brief Zip the specified file
        /// @param[in] inFile The full filename to zip
        /// @param[out] outFile The name of zipped file
        /// @return @c true if file was archived, otherwise @c false
        bool zipFile(const std::string &inFile, const std::string &outFile) const;

        /// @brief Unzip the specified file
        /// @param[in] inFile The full filename to unzip
        /// @param[out] outFile The name of unzipped file
        /// @param[in] ext The expected extansion of unzipped file
        /// @return @c true if file was unarchived, otherwise @c false
        bool unzipFile(const std::string &inFile, std::string &outFile, const std::string &ext = "csv") const;

        /// @brief Copies file
        /// @param srcFile The full filename to copy
        /// @param dstFile The full filename of destination
        /// @return @c true if file was copied, otherwise @c false
        bool copyFile(const std::string &srcFile, const std::string &dstFile) const;

        /// @brief Delete the specified file
        /// @param[in] filename The filename to delete
        /// @return @c true if file was deleted, otherwise @c false
        bool deleteFile(const std::string &filename) const;

        /// @brief Compares the content of two files
        /// @param filename1 The full name of first file
        /// @param filename2 The full name of second file
        /// @return result of execution of system() function, where:
        ///         @c match if files are equals,
        ///         @c not_match if files are different,
        ///         @c not_exist if one of file doesn't exist
        int compareFiles(const std::string &filename1,
                         const std::string &filename2) const;

        /// @brief Returns the list of files in alphabetic order
        ///        located into the specified local directory
        ///        and matched with the specified filename pattern
        /// @param[in] dirOrMask The specified local directory path with filename pattern
        /// @param[out] files The list of filenames located into the specified local directory
        ///                   and matched with the specified filename pattern
        /// @return @c true if the local directory was listed, otherwise @c false
        bool listDir(const std::string &dirOrMask, files::list &files) const;

        /// @brief Returns the latest file according to alphabetic order
        ///        among files located in the specified local directory
        ///        and matched with the specified filename pattern
        /// @param[in] dirOrMask The specified local directory path with filename pattern
        /// @param[out] filename The name of the latest file
        void getLatestFile(const std::string &dirOrMask, std::string &filename) const;

        /// @brief Create the specified local directory
        /// @param dir The full local directory path
        /// @return @c true if the local directory was created, otherwise @c false
        bool createDir(const std::string &dir) const;

        /// @brief Delete the specified local directory
        /// @param dir The full local directory path
        /// @return @c true if the local directory was deleted, otherwise @c false
        bool deleteDir(const std::string &dir) const;
    };
} // namespace fs
