#include "hasher.hpp"
#include "gosthash2012.h"

// STL includes.
#include <cstring>
#include <iomanip>

md::md(unsigned int len) : m_len(len)
{
    m_value = new unsigned char[m_len];
}
md::~md()
{
    delete[] m_value;
}

std::string md::do_base64()
{
    unsigned char md_value_base64[4 * (1 + (m_len / 3)) + 1];
    int md_value_base64_len = EVP_EncodeBlock(md_value_base64, m_value, m_len);
    return std::string(reinterpret_cast<const char *>(md_value_base64), md_value_base64_len);
}

std::string md::do_text() const
{
    return doText(m_value, m_len);
}

bool md::operator==(const md &other) const
{
    return (m_len == other.m_len &&
            (0 == std::memcmp(m_value, other.m_value, m_len)));
}

std::ostream &operator<<(std::ostream &os, const md *dgst)
{
    os << "Digest is: ";
    for (unsigned int i = 0; i < dgst->m_len; i++)
    {
        os << "|" << std::setw(2) << std::setfill('0') << std::hex << (int)dgst->m_value[i];
    }
    os << '\n';
    return os;
}

template <>
md_p hasher_t::do_hash<32>(const std::string &data)
{
    auto ret = std::make_shared<md>(32);

    gost2012_hash_ctx CTX;

    init_gost2012_hash_ctx(&CTX, (ret->m_len) << 3);
    gost2012_hash_block(&CTX, (const unsigned char *)data.c_str(), data.length());
    gost2012_finish_hash(&CTX, ret->m_value);

    return ret;
}

template <>
md_p hasher_t::do_hash<8>(const std::string &data)
{
    md_p md1 = do_hash<32>(data);

    const unsigned char *val = md1->m_value;

    uint64_t reduced_64 = (((*((const uint64_t *)&val[0]) ^
                             *((const uint64_t *)&val[8])) ^
                            *((const uint64_t *)&val[16])) ^
                           *((const uint64_t *)&val[24]));

    auto ret = std::make_shared<md>(8);
    memcpy(ret->m_value, &reduced_64, ret->m_len);

    return ret;
}

std::string hasher_t::do_hash_base64(const std::string &data)
{
    md_p md = do_hash<8>(data);

    unsigned char md_value_base64[4 * (1 + (md->m_len / 3)) + 1];

    int md_value_base64_len = EVP_EncodeBlock(md_value_base64, md->m_value, md->m_len);
    return std::string(reinterpret_cast<const char *>(md_value_base64), md_value_base64_len);
}

md_p undo_base64(const std::string &md_base64)
{
    auto ret = std::make_shared<md>(8);
    unsigned char md_value[md_base64.length()];
    EVP_DecodeBlock(md_value,
                    reinterpret_cast<const unsigned char *>(md_base64.c_str()),
                    static_cast<int>(md_base64.length()));
    memcpy(ret->m_value, md_value, ret->m_len);
    return ret;
}
