#pragma once

// STL includes.
#include <cstdint>
#include <string>

namespace Log
{
    enum Level : uint8_t
    {
        /// Disables all logs.
        Off,

        /// Used for logs that report an unwanted state that leads to the system
        /// termination. Immediate action is required to restart the system and
        /// to eliminate these errors.
        Fatal,

        /// Used for logs that report an unwanted state that prevents the system
        /// from continuing to function correctly. They require immediate actions
        /// to eliminate them.
        Error,

        /// Used for logs that report an unwanted state that does not prevent
        /// the system from normal working, but can lead to potential errors.
        /// Should be examined in case of errors.
        Warn,

        /// Used for logs that describe the normal behavior of the application.
        /// They can be used to monitor the process of work, but do not require
        /// a reaction or any action.
        Info,

        /// Used for logs that can help with troubleshooting, debugging,
        /// testing and optimization. Only useful for developers.
        Debug,

        /// Used for least important logs, only useful for developers. Use this
        /// type of logs to describe the step-by-step execution of algorithms or
        /// interaction with third-party libraries.
        Trace,

        /// Enables all logs.
        All
    };

    std::string toString(const Level level);
} // namespace Log

//------------------------------------------------------------------------------
// Definitions
//------------------------------------------------------------------------------

namespace Log
{
    inline std::string toString(const Level level)
    {
        switch (level)
        {
            case Level::Off:   return "off";
            case Level::Fatal: return "fatal";
            case Level::Error: return "error";
            case Level::Warn:  return "warn";
            case Level::Info:  return "info";
            case Level::Debug: return "debug";
            case Level::Trace: return "trace";
            case Level::All:   return "all";
        }

        return "<unk>";
    }
} // namespace Log
