#pragma once

// Log includes.
#include "GlobalLoggersGroup.hpp"
#include "SourceLocation.hpp"

namespace Log
{
    /// @name Global loggers group modification methods
    /// @{

    void addLogger(const LoggerPtr logger);
    void clearLoggers();

    /// @}

    /// @name Logging methods
    /// @{

    template <typename... Args>
    void f(const Args &...args);

    template <typename... Args>
    void fatal(const Args &...args);

    template <typename... Args>
    void e(const Args &...args);

    template <typename... Args>
    void error(const Args &...args);

    template <typename... Args>
    void w(const Args &...args);

    template <typename... Args>
    void warn(const Args &...args);

    template <typename... Args>
    void i(const Args &...args);

    template <typename... Args>
    void info(const Args &...args);

    template <typename... Args>
    void d(const Args &...args);

    template <typename... Args>
    void debug(const Args &...args);

    template <typename... Args>
    void t(const Args &...args);

    template <typename... Args>
    void trace(const Args &...args);

    template <typename... Args>
    void log(const Level level, const Args &...args);

    /// @}

} // namespace Log

//------------------------------------------------------------------------------
// Macros
//------------------------------------------------------------------------------

#define LOG_LEVEL_FATAL 1U
#define LOG_LEVEL_ERROR 2U
#define LOG_LEVEL_WARN  3U
#define LOG_LEVEL_INFO  4U
#define LOG_LEVEL_DEBUG 5U
#define LOG_LEVEL_TRACE 6U

#ifndef LOG_MACRO_LEVEL
#define LOG_MACRO_LEVEL LOG_LEVEL_INFO
#endif

#ifdef LOG_MACRO_ADD_SOURCE_LOCATION
#    define LOG(level, ...) Log::log(level, Log::SourceLocation(__FILE__, __LINE__), __VA_ARGS__)
#else
#    define LOG(level, ...) Log::log(level, __VA_ARGS__)
#endif

#if LOG_MACRO_LEVEL >= LOG_LEVEL_FATAL
#    define LOGFATAL(...) LOG(Log::Level::Fatal, __VA_ARGS__)
#    define LOGF(...)     LOGFATAL(__VA_ARGS__)
#else
#    define LOGFATAL(...)
#    define LOGF(...)
#endif

#if LOG_MACRO_LEVEL >= LOG_LEVEL_ERROR
#    define LOGERROR(...) LOG(Log::Level::Error, __VA_ARGS__)
#    define LOGE(...)     LOGERROR(__VA_ARGS__)
#else
#    define LOGERROR(...)
#    define LOGE(...)
#endif

#if LOG_MACRO_LEVEL >= LOG_LEVEL_WARN
#    define LOGWARN(...) LOG(Log::Level::Warn, __VA_ARGS__)
#    define LOGW(...)    LOGWARN(__VA_ARGS__)
#else
#    define LOGWARN(...)
#    define LOGW(...)
#endif

#if LOG_MACRO_LEVEL >= LOG_LEVEL_INFO
#    define LOGINFO(...) LOG(Log::Level::Info, __VA_ARGS__)
#    define LOGI(...)    LOGINFO(__VA_ARGS__)
#else
#    define LOGINFO(...)
#    define LOGI(...)
#endif

#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
#    define LOGDEBUG(...) LOG(Log::Level::Debug, __VA_ARGS__)
#    define LOGD(...)     LOGDEBUG(__VA_ARGS__)
#else
#    define LOGDEBUG(...)
#    define LOGD(...)
#endif

#if LOG_MACRO_LEVEL >= LOG_LEVEL_TRACE
#    define LOGTRACE(...) LOG(Log::Level::Trace, __VA_ARGS__)
#    define LOGT(...)     LOGTRACE(__VA_ARGS__)
#else
#    define LOGTRACE(...)
#    define LOGT(...)
#endif

//------------------------------------------------------------------------------
// Definitions
//------------------------------------------------------------------------------

namespace Log
{
    // Global loggers group modification methods:

    inline void addLogger(const LoggerPtr logger)
    {
        GlobalLoggersGroup::getInstance().add(logger);
    }

    inline void clearLoggers()
    {
        GlobalLoggersGroup::getInstance().clear();
    }

    // Logging methods:

    template <typename... Args>
    inline void f(const Args &...args)
    {
        fatal(args...);
    }

    template <typename... Args>
    inline void fatal(const Args &...args)
    {
        GlobalLoggersGroup::getInstance().fatal(args...);
    }

    template <typename... Args>
    inline void e(const Args &...args)
    {
        error(args...);
    }

    template <typename... Args>
    inline void error(const Args &...args)
    {
        GlobalLoggersGroup::getInstance().error(args...);
    }

    template <typename... Args>
    inline void w(const Args &...args)
    {
        warn(args...);
    }

    template <typename... Args>
    inline void warn(const Args &...args)
    {
        GlobalLoggersGroup::getInstance().warn(args...);
    }

    template <typename... Args>
    inline void i(const Args &...args)
    {
        info(args...);
    }

    template <typename... Args>
    inline void info(const Args &...args)
    {
        GlobalLoggersGroup::getInstance().info(args...);
    }

    template <typename... Args>
    inline void d(const Args &...args)
    {
        debug(args...);
    }

    template <typename... Args>
    inline void debug(const Args &...args)
    {
        GlobalLoggersGroup::getInstance().debug(args...);
    }

    template <typename... Args>
    inline void t(const Args &...args)
    {
        trace(args...);
    }

    template <typename... Args>
    inline void trace(const Args &...args)
    {
        GlobalLoggersGroup::getInstance().trace(args...);
    }

    template <typename... Args>
    void log(const Level level, const Args &...args)
    {
        GlobalLoggersGroup::getInstance().log(level, args...);
    }
} // namespace Log
