#pragma once

// Log includes.
#include "Logger.hpp"

// STL includes.
#include <unordered_set>

namespace Log
{
    class LoggersGroup
    {
    public:
        LoggersGroup() = default;
        virtual ~LoggersGroup() = default;

    public:
        /// @name Group modification methods
        /// @{

        void add(const LoggerPtr logger);

        void clear();

        /// @}

    public:
        /// @name Loggers configuration methods
        /// @{

        void setLevel(const Level level);

        /// @}

    public:
        /// @name Logging methods
        /// @{

        template <typename... Args>
        void f(const Args &...args);

        template <typename... Args>
        void fatal(const Args &...args);

        template <typename... Args>
        void e(const Args &...args);

        template <typename... Args>
        void error(const Args &...args);

        template <typename... Args>
        void w(const Args &...args);

        template <typename... Args>
        void warn(const Args &...args);

        template <typename... Args>
        void i(const Args &...args);

        template <typename... Args>
        void info(const Args &...args);

        template <typename... Args>
        void d(const Args &...args);

        template <typename... Args>
        void debug(const Args &...args);

        template <typename... Args>
        void t(const Args &...args);

        template <typename... Args>
        void trace(const Args &...args);

        template <typename... Args>
        void log(const Level level, const Args &...args);

        /// @}

    private:
        std::unordered_set<LoggerPtr> m_loggers;
    };
} // namespace Log

//------------------------------------------------------------------------------
// Definitions
//------------------------------------------------------------------------------

namespace Log
{
    // Group modification methods:

    inline void LoggersGroup::add(const LoggerPtr logger)
    {
        m_loggers.insert(logger);
    }

    inline void LoggersGroup::clear()
    {
        m_loggers.clear();
    }

    // Loggers configuration methods:

    inline void LoggersGroup::setLevel(const Level level)
    {
        for (const auto &logger : m_loggers)
        {
            logger->setLevel(level);
        }
    }

    // Logging methods:

    template <typename... Args>
    inline void LoggersGroup::f(const Args &...args)
    {
        fatal(args...);
    }

    template <typename... Args>
    inline void LoggersGroup::fatal(const Args &...args)
    {
        log(Level::Fatal, args...);
    }

    template <typename... Args>
    inline void LoggersGroup::e(const Args &...args)
    {
        error(args...);
    }

    template <typename... Args>
    inline void LoggersGroup::error(const Args &...args)
    {
        log(Level::Error, args...);
    }

    template <typename... Args>
    inline void LoggersGroup::w(const Args &...args)
    {
        warn(args...);
    }

    template <typename... Args>
    inline void LoggersGroup::warn(const Args &...args)
    {
        log(Level::Warn, args...);
    }

    template <typename... Args>
    inline void LoggersGroup::i(const Args &...args)
    {
        info(args...);
    }

    template <typename... Args>
    inline void LoggersGroup::info(const Args &...args)
    {
        log(Level::Info, args...);
    }

    template <typename... Args>
    void LoggersGroup::d(const Args &...args)
    {
        debug(args...);
    }

    template <typename... Args>
    inline void LoggersGroup::debug(const Args &...args)
    {
        log(Level::Debug, args...);
    }

    template <typename... Args>
    inline void LoggersGroup::t(const Args &...args)
    {
        trace(args...);
    }

    template <typename... Args>
    inline void LoggersGroup::trace(const Args &...args)
    {
        log(Level::Trace, args...);
    }

    template <typename... Args>
    inline void LoggersGroup::log(Level level, const Args &...args)
    {
        for (const auto &logger : m_loggers)
        {
            logger->log(level, args...);
        }
    }
} // namespace Log
