#pragma once

// Log includes.
#include "Logger.hpp"

// STL includes.
#include <condition_variable>
#include <mutex>
#include <queue>
#include <thread>

namespace Log
{
    class SyncLogger : public Logger
    {
    public:
        SyncLogger(LoggerPtr &&logger);
        ~SyncLogger() override;

    private:
        void processMessage(std::string &&message) override;

    private:
        void processQueue();

    private:
        LoggerPtr m_logger = nullptr;

        std::mutex m_queueMutex;
        std::queue<std::string> m_messagesQueue;
        bool m_shouldStop = false;
        std::thread m_processQueueThread;
        std::condition_variable m_processQueueCondition;
    };
} // namespace Log

//------------------------------------------------------------------------------
// Definitions
//------------------------------------------------------------------------------

namespace Log
{
    inline SyncLogger::SyncLogger(LoggerPtr &&logger)
        : m_logger(std::move(logger))
    {
        m_processQueueThread = std::thread(&SyncLogger::processQueue, this);
    }

    inline SyncLogger::~SyncLogger()
    {
        std::unique_lock<std::mutex> lock(m_queueMutex);
        m_shouldStop = true;
        lock.unlock();

        m_processQueueCondition.notify_one();
        m_processQueueThread.join();
    }

    inline void SyncLogger::processMessage(std::string &&message)
    {
        std::unique_lock<std::mutex> lock(m_queueMutex);
        m_messagesQueue.emplace(std::move(message));
        lock.unlock();

        m_processQueueCondition.notify_one();
    }

    inline void SyncLogger::processQueue()
    {
        while (true)
        {
            std::unique_lock<std::mutex> lock(m_queueMutex);
            m_processQueueCondition.wait(lock, [this] {
                return (!this->m_messagesQueue.empty()
                        || this->m_shouldStop);
            });

            if (m_shouldStop)
            {
                break;
            }

            auto message = std::move(m_messagesQueue.front());
            m_messagesQueue.pop();
            lock.unlock();

            m_logger->processMessage(std::move(message));
        }

        // Process the leftovers.
        while (!m_messagesQueue.empty())
        {
            auto message = std::move(m_messagesQueue.front());
            m_messagesQueue.pop();
            m_logger->processMessage(std::move(message));
        }
    }
} // namespace Log
