#include "numbers/src/TRecQ763Number.hpp"
#include "components/log/src/Log.hpp"

// AFAP includes.
#include "components/hasher/src/hasher.hpp"

// STL includes.
#include <cstring>

constexpr auto s_oddEvenIndicatorShift = 7;
constexpr auto s_oddDigitMask = 0x0F;
constexpr auto s_evenDigitMask = 0xF0;

constexpr auto s_natureOfAddressMask = 0x7F;
constexpr auto s_natureOfAddressHashedNumber = 0x7E;

namespace TRecQ763Number
{
    uint64_t toUInt64(const Uvr::OctetStringPtr &number)
    {
        const uint8_t *buf = number->data();

        // From the first byte we read the bit that is an Odd/Even indicator.
        const bool isEvenNbOfDigits = TRecQ763Number::isEvenNbOfDigits(number);

        // Then we skip the second byte, since it does not contain information useful to us.
        size_t nbBytesProcessed = 2;
        buf += nbBytesProcessed;

        // Now we read the digits for the number until the buffer runs out.
        uint64_t retNumber = 0;
        for (; nbBytesProcessed < number->size(); ++nbBytesProcessed)
        {
            // Read odd digit.
            retNumber *= 10;
            retNumber += (*buf & s_oddDigitMask);

            // Read even digit.
            if (nbBytesProcessed < (number->size() - 1) || isEvenNbOfDigits)
            {
                retNumber *= 10;
                retNumber += ((*buf & s_evenDigitMask) >> 4);
            }

            ++buf;
        }

        return retNumber;
    }

    std::string toStdString(const Uvr::OctetStringPtr &number)
    {
        if (number == nullptr || number->empty())
        {
            return "";
        }

        const bool isNumHashed = isNumberHashed(number);
        if (isNumHashed)
        {
            const uint64_t *numberHash = getNumberHashPtr(number);
            return doText(numberHash, 8);
        }

        return std::to_string(toUInt64(number));
    }

    std::string toStdString(const Uvr::OctetStringPtr &number, const size_t nbBytes)
    {
        if (number == nullptr || number->empty())
        {
            return "";
        }

        const size_t newSize = std::min(nbBytes + 2, number->size());
        if (newSize < 3)
        {
            return "";
        }

        Uvr::OctetStringPtr partOfNumber = std::make_shared<Uvr::OctetString>(number->begin(), number->begin() + newSize);

        const bool isNumHashed = isNumberHashed(partOfNumber);
        if (isNumHashed)
        {
            const uint64_t *numberHash = getNumberHashPtr(partOfNumber);
            return doText(numberHash, newSize - 2);
        }

        return std::to_string(toUInt64(partOfNumber));
    }

    Uvr::OctetStringPtr toOctetStringPtr(const Uvr::Number &number)
    {
        if (number.empty())
        {
            return Uvr::OctetStringPtr();
        }

        auto retNumber = std::make_shared<Uvr::OctetString>(2 + ((number.size() + 1) / 2), 0);
        uint8_t *buf = retNumber->data();

        // Write to the first byte the bit that is an Odd/Even indicator.
        bool isEvenNbOfDigits = (number.size() % 2) == 0;
        if (!isEvenNbOfDigits)
        {
            *buf |= (1 << s_oddEvenIndicatorShift);
        }

        // TODO: write the rest of the required information to the first byte.
        // TODO: write the required information to the second byte.
        buf += 2;

        // Write the digits for the number to the rest bytes.
        for (size_t digitIndex = 0; digitIndex < number.size(); ++digitIndex)
        {
            // Write odd digit.
            *buf |= ((number[digitIndex] - '0') & s_oddDigitMask);

            // Write even digit.
            if (++digitIndex < number.size())
            {
                *buf |= (((number[digitIndex] - '0') << 4) & s_evenDigitMask);
            }

            ++buf;
        }

        return retNumber;
    }

    Uvr::OctetStringPtr toHashedOctetStringPtr(const Uvr::OctetStringPtr &number)
    {
        if (number == nullptr || isNumberHashed(number))
        {
            return number;
        }
        else
        {
            return toHashedOctetStringPtr(toStdString(number));
        }
    }

    Uvr::OctetStringPtr toHashedOctetStringPtr(const Uvr::Number &number)
    {
        if (number.empty())
        {
            return Uvr::OctetStringPtr();
        }

        auto retNumber = std::make_shared<Uvr::OctetString>(hashedNumberSize, 0);
        uint8_t *buf = retNumber->data();

        // Write to the first byte the Nature Of Address Indicator with hashed value.
        *buf |= s_natureOfAddressHashedNumber;

        // Move to starting byte for number hash.
        buf += 2;

        // Make number hash.
        hasher_t hasher;
        md_p md = hasher.do_hash<8>(number);

        // Copy hash.
        std::memcpy(buf, md->m_value, 8);

        return retNumber;
    }

    bool isEvenNbOfDigits(const Uvr::OctetStringPtr &number)
    {
        return !(static_cast<bool>(*number->data() >> s_oddEvenIndicatorShift));
    }

    bool isNumberHashed(const Uvr::OctetStringPtr &number)
    {
        return ((*number->data() & s_natureOfAddressMask) == s_natureOfAddressHashedNumber);
    }

    const uint64_t *getNumberHashPtr(const Uvr::OctetStringPtr &number)
    {
        if (number->size() != hashedNumberSize)
        {
            LOGWARN(" hash has invalid length of ", number->size());
        }

        return reinterpret_cast<const uint64_t *>(number->data() + 2);
    }

    bool areEqual(const Uvr::OctetStringPtr &firstNumber, const Uvr::OctetStringPtr &secondNumber)
    {
        const bool isFirstNumberHashed = isNumberHashed(firstNumber);
        const bool isSecondNumberHashed = isNumberHashed(secondNumber);

        // If both numbers are hashed, then we compare the hashes byte by byte
        // without taking into account the additional information in the first two bytes.
        if (isFirstNumberHashed && isSecondNumberHashed)
        {
            if (firstNumber->size() == secondNumber->size())
            {
                const auto *firstNumberHashPtr = getNumberHashPtr(firstNumber);
                const auto *secondNumberHashPtr = getNumberHashPtr(secondNumber);

                return std::memcmp(firstNumberHashPtr, secondNumberHashPtr, firstNumber->size() - 2) == 0;
            }
            else
            {
                return false;
            }
        }

        // If both numbers are not hashed, then we compare the numbers byte by byte,
        // taking into account only one bit (Odd/Even Indicator) from the first two bytes.
        if (!isFirstNumberHashed && !isSecondNumberHashed)
        {
            if ((firstNumber->size() != secondNumber->size()) ||
                (isEvenNbOfDigits(firstNumber) != isEvenNbOfDigits(secondNumber)))
            {
                return false;
            }

            if (!isEvenNbOfDigits(firstNumber))
            {
                // Replace filler as top 4 bits of last digit.
                firstNumber->back() &= s_oddDigitMask;
                secondNumber->back() &= s_oddDigitMask;
            }

            const auto *firstNumberPtr = firstNumber->data() + 2;
            const auto *secondNumberPtr = secondNumber->data() + 2;

            return std::memcmp(firstNumberPtr, secondNumberPtr, firstNumber->size() - 2) == 0;
        }

        return isFirstNumberHashed ? areHashedEqual(firstNumber, secondNumber)
                                   : areHashedEqual(secondNumber, firstNumber);
    }

    bool areHashedEqual(const Uvr::OctetStringPtr &hashedFirstNumber, const Uvr::OctetStringPtr &secondNumber)
    {
        const uint64_t *firstNumberHash = getNumberHashPtr(hashedFirstNumber);

        hasher_t hasher;
        md_p secondNumberMD = hasher.do_hash<8>(toStdString(secondNumber));

        auto secondNumberHash = reinterpret_cast<const uint64_t *>(secondNumberMD->m_value);

        return *firstNumberHash == *secondNumberHash;
    }

} // TRecQ763Number
