// Own include.
#include "Validator.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"

// STL includes.
#include <cstdint>

namespace
{
    constexpr uint64_t s_codesExtractor = 1'000'00'00;
    constexpr uint64_t s_areaCodeExtractor = 1'000;
    constexpr uint64_t s_countryCodeExtractor = s_codesExtractor * s_areaCodeExtractor;
    constexpr uint64_t s_russiaCountryCode = 7;

    Log::Tag s_logTag("TRecQ763NumberValidator");
    const std::array<std::pair<uint16_t, uint16_t>, 4> s_areaCodeRanges{{{300, 399},
                                                                         {400, 499},
                                                                         {800, 899},
                                                                         {900, 999}}};
}

namespace TRecQ763Number
{
    bool Validator::verify(const Uvr::OctetStringPtr &number,
                           const bool isEmptyValid,
                           const bool trimLongNumber)
    {
        // Check empty number.
        if (number == nullptr || number->empty())
        {
            if (!isEmptyValid)
            {
                LOGW(s_logTag, "Number is empty.");
                return false;
            }

            return true;
        }

        const size_t size = number->size();

        // Check hashed number.
        if (TRecQ763Number::isNumberHashed(number))
        {
            // Check size.
            if (size < TRecQ763Number::hashedNumberSize)
            {
                LOGW(s_logTag, "Octet string with hashed number = \"", TRecQ763Number::toStdString(number),
                     "\" is too short. Expected size = 10 bytes, actual size = ", size, " bytes.");
                return false;
            }

            if (size > TRecQ763Number::hashedNumberSize)
            {
                LOGW(s_logTag, "Octet string with hashed number = \"", TRecQ763Number::toStdString(number, hashedNumberTrimSize),
                     "...\" is too long. Expected size = 10 bytes, actual size = ", size, " bytes.");

                if (trimLongNumber)
                {
                    LOGW(s_logTag, "Too long number will be trimmed to ", size, " bytes.");
                    number->resize(hashedNumberTrimSize);
                }

                return false;
            }

            return true;
        }

        // Check not hashed number.
        // Check size.
        if (size < TRecQ763Number::notHashedNumberSize)
        {
            LOGW(s_logTag, "Octet string with number = \"", TRecQ763Number::toStdString(number),
                 "\" is too short. Expected size = 10 bytes, actual size = ", size, " bytes.");
            return false;
        }

        if (size > TRecQ763Number::notHashedNumberSize)
        {
            LOGW(s_logTag, "Octet string with number = \"", TRecQ763Number::toStdString(number, notHashedNumberTrimSize),
                 "...\" is too long. Expected size = 8 bytes, actual size = ", size, " bytes.");

            if (trimLongNumber)
            {
                LOGW(s_logTag, "Too long number will be trimmed to ", size, " bytes.");
                number->resize(notHashedNumberTrimSize);
            }

            return false;
        }

        if (TRecQ763Number::isEvenNbOfDigits(number))
        {
            LOGW(s_logTag, "Octet string with number = \"", TRecQ763Number::toStdString(number),
                 "\" has 12 digits. Must be 11.");
            return false;
        }

        // Check that the number starts with the "7".
        const uint64_t numberInt = TRecQ763Number::toUInt64(number);
        const uint8_t countryCode = static_cast<uint8_t>(numberInt / s_countryCodeExtractor);
        if (countryCode != s_russiaCountryCode)
        {
            LOGW(s_logTag, "Number = \"", TRecQ763Number::toStdString(number),
                 "\" has invalid country code = \"", std::to_string(countryCode), "\". Must be \"7\".");
            return false;
        }

        // Check that the area code is valid.
        const uint16_t areaCode = static_cast<uint16_t>((numberInt / s_codesExtractor) % s_areaCodeExtractor);
        bool isAreaCodeValid = false;
        for (const auto &areaCodeRange : s_areaCodeRanges)
        {
            if (areaCodeRange.first <= areaCode && areaCode <= areaCodeRange.second)
            {
                isAreaCodeValid = true;
                break;
            }
        }

        if (!isAreaCodeValid)
        {
            LOGW(s_logTag, "Number = \"", TRecQ763Number::toStdString(number),
                 "\" has invalid area code = \"", std::to_string(areaCode),
                 "\". Must be one of \"3XX\", \"4XX\", \"8XX\" or \"9XX\".");
            return false;
        }

        return true;
    }
}
