#include "sequencer.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"

// Boost includes.
#include <boost/bind.hpp>

int sequencer::id_cntr{0};
int sequencer::number_of_ongoing_sequences{0};
std::recursive_mutex sequencer::m_guard{};

sequencer::sequencer(const std::string &name,
                     const TypeInst &owner,
                     boost::asio::io_context *ioc_p) : m_name(name),
                                                       m_owner(owner),
                                                       m_ioc_p(ioc_p)

{
    std::lock_guard<std::recursive_mutex> guard(m_guard);
    number_of_ongoing_sequences++;
    m_id = ++id_cntr;
}

sequencer::~sequencer()
{
    std::lock_guard<std::recursive_mutex> guard(m_guard);
    number_of_ongoing_sequences--;
    LOGDEBUG(m_owner,
             " Sequence #", m_id, " \"", m_name, "\" DESTROYED. Left: ", number_of_ongoing_sequences);
}

int sequencer::get_number_of_ongoing_sequences()
{
    std::lock_guard<std::recursive_mutex> guard(m_guard);
    return number_of_ongoing_sequences;
}

void sequencer::addFunction(std::function<bool(void)> const &func)
{
    m_funcs.push_back(func);
}

void sequencer::start()
{
    std::lock_guard<std::recursive_mutex> guard(m_guard);
    LOGD(m_owner,
         " Sequence #", m_id, " \"", m_name, "\" START");
    int n;
    if ((n = get_number_of_ongoing_sequences()) > 10)
    {
        LOGWARN(m_owner,
                " There are ", n, " sequences ongoing");
    }
    boost::asio::post(*m_ioc_p, std::bind(&sequencer::do_next_step, shared_from_this()));
}

void sequencer::do_next_step(void)
{
    LOGTRACE(m_owner,
             " Sequence #", m_id, " \"", m_name, "\", step #", m_step + 1, "/", m_funcs.size(), " is started");
    bool result = m_funcs[m_step]();
    LOGTRACE(m_owner,
             " Sequence #", m_id, " \"", m_name, "\", step #", m_step + 1, "/", m_funcs.size(), " is completed with result=", result);
    m_step++;
    std::string stopReason;
    if (result)
    {
        if (m_step < m_funcs.size())
        {
            boost::asio::post(*m_ioc_p, std::bind(&sequencer::do_next_step, shared_from_this()));
            return;
        }
        else
        {
            stopReason = "COMPLETE";
        }
    }
    else
    {
        stopReason = "ABORTED";
    }
    LOGTRACE(m_owner,
             " Sequence #", m_id, " \"", m_name, "\" ", stopReason, ". Left: ", get_number_of_ongoing_sequences() - 1);
}
