#include "timers/src/timer_manager.hpp"
#include "timers/src/timer_oneshot.hpp"
#include "timers/src/timer_periodic.hpp"

namespace timers
{
    manager::manager(boost::asio::io_context &ioc)
        : m_ioc(ioc)
    {
        m_timerId_cntr.init();
    }

    id manager::find_free_timerId(void)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        if (!m_rel_timerIds.empty())
        {
            timers::id timer_id = m_rel_timerIds.front();
            m_rel_timerIds.pop_front();
            return timer_id;
        }
        return m_timerId_cntr++;
    }

    void manager::delTimerAndStop(timers::id &id)
    {
        if (id.undefined())
        {
            LOGWARN("Cannot stop timer: timer ID is undefined");
            return;
        }

        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto iter = m_timers.find(id);
        if (iter == m_timers.end())
        {
            LOGWARN("Cannot stop timer: timer with ID=", id, " not found");
            id.reset();
            return;
        }

        if (iter->second)
        {
            iter->second->stop();
        }
        m_timers.erase(iter);
        m_rel_timerIds.push_front(id);
        LOGDEBUG(iter->second, " was deleted, ", id, " was returned to free timer IDs");
        id.reset();
    }

    void manager::delAndStopAll(void)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        interfaces::manager::server::stop_server();
        for (auto iter = m_timers.begin(); iter != m_timers.end(); iter++)
        {
            if (iter->second)
            {

                iter->second->stop();
            }
        }
        m_timers.clear();
        m_rel_timerIds.clear();
    }

    void manager::restartTimer(const timers::id &id)
    {
        if (id.undefined())
        {
            LOGWARN("Cannot restart timer: timer ID is undefined");
            return;
        }

        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto iter = m_timers.find(id);
        if (iter == m_timers.end())
        {
            LOGWARN("Cannot restart timer: timer with ID=", id, " not found");
            return;
        }

        switch (iter->second->getType())
        {
            case timers::type::ONESHOT:
            {
                timers::restartTimer<timers::oneshot>(iter->second);
                break;
            }
            case timers::type::PERIODIC:
            {
                timers::restartTimer<timers::periodic>(iter->second);
                break;
            }
            default:
                LOGWARN("Cannot restart ", iter->second.get(), ": unknown timer type");
        }
    }

    manager::~manager()
    {
        delAndStopAll();
    }

    void manager::manage_timeout(const ::timers::id &id,
                                 const ::timers::time_point &startTime)
    {
        interfaces::notifier::client::notify_timeout(id, startTime);
    }

    void manager::manage_stopped(const ::timers::id &id)
    {
        if (id.undefined())
        {
            LOGWARN("Cannot stop timer: timer ID is undefined");
            return;
        }

        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto iter = m_timers.find(id);
        if (iter == m_timers.end())
        {
            LOGWARN("Cannot stop timer: timer with ID=", id, " not found");
            return;
        }

        m_timers.erase(iter);
        m_rel_timerIds.push_front(id);
        LOGDEBUG(iter->second, " was managed to stop, ", id, " was returned to free timer IDs");
    }
} // namespace timers
