#include "timer_types.hpp"
#include "components/log/src/Log.hpp"

// STL includes.
#include <iomanip>
#include <regex>

namespace timers
{

    time_point::time_point(const std::string &timeline) : timespec{0, 0}
    {
        if (!timeline.empty())
        {
            // Get time zone from timestamp
            int hours = 0, minutes = 0;
            char sign = '+';
            const std::regex tzRegex("([+-])(0[0-9]|1[0-4]):([0-5][0-9])");
            std::smatch tzMatches;
            if (std::regex_search(timeline, tzMatches, tzRegex) &&
                tzMatches.size() == 4)
            {
                sign = tzMatches[1].str()[0];
                hours = std::stoull(tzMatches[2].str());
                minutes = std::stoull(tzMatches[3].str());
                LOGDEBUG("Timestamp=\'", timeline, "\' contains TZ: hours=",
                         hours, ", minutes=", minutes, ", sign=\'", sign, "\'");
            }
            else
            {
                LOGWARN("Timestamp=\'", timeline,
                        "\' does not contain TZ or corrupted, consider it as UTC");
            }
            // Get seconds since the Epoch as UTC
            std::istringstream ss(timeline);
            struct std::tm tmVal = {0};
            ss >> std::get_time(&tmVal, "%Y-%m-%dT%T");
            tv_sec = timegm(&tmVal);
            // Apply TZ offset
            int offset = hours * 3600 + minutes * 60;
            if (sign == '+')
                tv_sec -= offset;
            else
                tv_sec += offset;
        }
    }

    time_point::time_point(const std::string &timeline,
                           const char *format_str) : timespec{0, 0}
    {
        if (!timeline.empty())
        {
            std::istringstream ss(timeline);
            struct std::tm tm = {0};
            ss >> std::get_time(&tm, format_str);
            tv_sec = timegm(&tm);
        }
    }

    std::string time_point::format() const
    {
        std::string timestamp = format("%Y-%m-%dT%T%z");
        timestamp.insert(timestamp.size() - 2, 1, ':');
        return timestamp;
    }

    std::string time_point::format(const char *format_str) const
    {
        std::ostringstream os_timestamp;
        struct tm tmVal;
        localtime_r(&tv_sec, &tmVal);
        os_timestamp << std::put_time(&tmVal, format_str);
        return os_timestamp.str();
    }

    std::string time_point::utcFormat(const char *format_str) const
    {
        std::ostringstream os_timestamp;
        struct tm tmVal;
        gmtime_r(&tv_sec, &tmVal);
        os_timestamp << std::put_time(&tmVal, format_str);
        return os_timestamp.str();
    }

    time_point time_point::operator-(const time_point &other) const
    {
        time_point ret;
        ret.tv_sec = tv_sec - other.tv_sec;
        if (tv_nsec >= other.tv_nsec)
        {
            ret.tv_nsec = tv_nsec - other.tv_nsec;
        }
        else
        {
            ret.tv_nsec = (giga + tv_nsec) - other.tv_nsec;
            ret.tv_sec--;
        }
        return ret;
    }

    time_point time_point::operator+(const time_point &other) const
    {
        time_point ret;
        ret.tv_sec = tv_sec + other.tv_sec;
        ret.tv_nsec = tv_nsec + other.tv_nsec;
        if (ret.tv_nsec >= giga)
        {
            ret.tv_nsec -= giga;
            ret.tv_sec++;
        }
        return ret;
    }

    bool time_point::operator>(const time_point &other) const
    {
        if (tv_sec > other.tv_sec)
        {
            return true;
        }
        else if (tv_sec == other.tv_sec)
        {
            return (tv_nsec > other.tv_nsec);
        }
        else
        {
            return false;
        }
    }

    bool time_point::operator>=(const time_point &other) const
    {
        return !(*this < other);
    }

    bool time_point::operator<(const time_point &other) const
    {
        if (tv_sec < other.tv_sec)
        {
            return true;
        }
        else if (tv_sec == other.tv_sec)
        {
            return (tv_nsec < other.tv_nsec);
        }
        else
        {
            return false;
        }
    }

    bool time_point::operator<=(const time_point &other) const
    {
        return !(*this > other);
    }

    bool time_point::operator==(const time_point &other) const
    {
        return (tv_sec == other.tv_sec) && (tv_nsec == other.tv_nsec);
    }

    bool time_point::operator!=(const time_point &other) const
    {
        return !(*this == other);
    }

    std::ostream &operator<<(std::ostream &os, const time_point &obj)
    {
        os << obj.tv_sec << " s " << std::setw(3) << (obj.tv_nsec / timers::mega) << " ms";
        return os;
    }

    void now(time_point &obj)
    {
        clock_gettime(CLOCK_REALTIME, &obj);
    }

    time_point now()
    {
        time_point ret;
        now(ret);
        return ret;
    }

}
