#include "components/tools/src/Tools.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"

// STL includes.
#include <cstring>
#include <iostream>
#include <array>
#include <fstream>

std::string doText(const void *buffer, const size_t size)
{
    static const std::string map("0123456789abcdef");
    std::string ret(2 * size, 0);
    auto ptr = reinterpret_cast<const uint8_t *>(buffer);
    char *ptr1 = &ret.front();
    for (uint i = 0; i < size; i++, ptr++, ptr1 += 2)
    {
        ptr1[1] = map[*ptr & 0x0F];
        ptr1[0] = map[*ptr >> 4];
    }
    return ret;
}

std::string doText(const std::string &str)
{
    return doText(&str[0], str.size());
}

std::string doText(Uvr::OctetStringPtr str)
{
    if (str == nullptr)
    {
        return "";
    }

    return doText(&(*str)[0], str->size());
}

int hexCharToInt(unsigned char c)
{
    switch (c)
    {
        case '0':
            return 0;
        case '1':
            return 1;
        case '2':
            return 2;
        case '3':
            return 3;
        case '4':
            return 4;
        case '5':
            return 5;
        case '6':
            return 6;
        case '7':
            return 7;
        case '8':
            return 8;
        case '9':
            return 9;
        case 'a':
        case 'A':
            return 0x0A;
        case 'b':
        case 'B':
            return 0x0B;
        case 'c':
        case 'C':
            return 0x0C;
        case 'd':
        case 'D':
            return 0x0D;
        case 'e':
        case 'E':
            return 0x0E;
        case 'f':
        case 'F':
            return 0x0F;
        default:
            return 0;
    }
}

Uvr::OctetString undoText(const std::string &md_text)
{
    if (md_text.size() % 2 != 0)
    {
        return Uvr::OctetString();
    }

    Uvr::OctetString ret(md_text.size() / 2);

    unsigned char *q;
    unsigned char ch;
    unsigned char cl;
    int chi;
    int cli;
    const unsigned char *p;
    size_t cnt;

    for (p = reinterpret_cast<const unsigned char *>(md_text.c_str()), q = ret.data(), cnt = 0; *p;)
    {
        ch = *p++;
        cl = *p++;
        cli = hexCharToInt(cl);
        chi = hexCharToInt(ch);
        cnt++;
        *q++ = (unsigned char)((chi << 4) | cli);
    }

    return ret;
}

namespace console
{

    output getCmdOutput(const std::string &cmd)
    {
        output out;
        constexpr int maxOutputLineSize = 8192;

        LOGDEBUG("Execute command '", cmd, "'");
        std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(cmd.c_str(), "r"), &pclose);
        std::array<char, maxOutputLineSize> buffer;
        if (!pipe)
        {
            LOGERROR(" Cannot execute command=\"", cmd, "\": failed to open pipe");
        }
        else
        {
            while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr)
            {
                out.emplace_back(buffer.data(), std::strlen(buffer.data()) - 1);
                LOGDEBUG(buffer.data());
            }
        }
        return out;
    }

} // namespace console

namespace env
{

    std::string get(const std::string &key)
    {
        std::optional<std::string> valStr = getOptional(key);

        if (!valStr.has_value())
        {
            throw std::invalid_argument("environment var with key=" + key + " is missing");
        }

        return valStr.value();
    }

    size_t getSizeT(const std::string &key)
    {
        std::optional<std::string> valStr = getOptional(key);

        if (!valStr.has_value())
        {
            throw std::invalid_argument("environment var with key=" + key + " is missing");
        }

        size_t ret = strtoul(valStr.value().c_str(), nullptr, 10);
        if (ret == 0)
        {
            throw std::invalid_argument("environment var with key=" + key + "is invalid with " + valStr.value());
        }
        return ret;
    }

    size_t getSizeT(const std::string &key,
                    const std::string &defaultValue)
    {
        return strtoul(get(key, defaultValue).c_str(), nullptr, 10);
    }

    std::string get(const std::string &key,
                    const std::string &defaultValue)
    {
        std::optional<std::string> value = getOptional(key);

        if (value.has_value())
        {
            return value.value();
        }
        else
        {
            return defaultValue;
        }
    }

    std::optional<std::string> getOptional(const std::string &key)
    {
        std::optional<std::string> ret;
        char *envVar = std::getenv(key.c_str());
        if (envVar)
        {
            std::string var(envVar);
            if (!var.empty())
            {
                ret.emplace(var);
            }
        }
        return ret;
    }

} // namespace env

namespace Json
{
    Errors parseFile(const std::string &filename,
                     boost::json::value &fileValue)
    {
        std::ifstream in(filename);
        boost::json::stream_parser parser;
        boost::json::error_code err;

        if (!in.is_open())
        {
            LOGERROR("Cannot open input file=\"", filename, "\"");
            return Errors::FAIL;
        }

        std::array<char, 1024> buf;
        while (in && !err)
        {
            auto const nread = in.read(buf.data(), buf.size()).gcount();
            parser.write(buf.data(), nread, err);
        }
        if (err)
        {
            LOGERROR("Json failed to read file=\"", filename, "\": ", err.message());
            in.close();
            return Errors::FAIL;
        }

        parser.finish(err);
        in.close();
        if (err)
        {
            LOGERROR("Json failed to parse file=\"", filename, "\": ", err.message());
            return Errors::FAIL;
        }

        try
        {
            fileValue = parser.release();
        }
        catch (const boost::json::system_error &ex)
        {
            LOGERROR("Cannot get Json value for file=\"", filename, "\": ", ex.what());
            return Errors::FAIL;
        }

        return Errors::OK;
    }

    void prettyPrint(std::ostream &out,
                     boost::json::value const &jv,
                     std::string *indent_p)
    {
        std::string indent;
        if (!indent_p)
            indent_p = &indent;
        switch (jv.kind())
        {
            case boost::json::kind::object:
            {
                out << "{\n";
                indent_p->append(4, ' ');
                auto const &obj = jv.get_object();
                if (!obj.empty())
                {
                    auto it = obj.begin();
                    for (;;)
                    {
                        out << *indent_p << boost::json::serialize(it->key()) << " : ";
                        prettyPrint(out, it->value(), indent_p);
                        if (++it == obj.end())
                            break;
                        out << ",\n";
                    }
                }
                out << "\n";
                indent_p->resize(indent_p->size() - 4);
                out << *indent_p << "}";
                break;
            }

            case boost::json::kind::array:
            {
                out << "[\n";
                indent_p->append(4, ' ');
                auto const &arr = jv.get_array();
                if (!arr.empty())
                {
                    auto it = arr.begin();
                    for (;;)
                    {
                        out << *indent_p;
                        prettyPrint(out, *it, indent_p);
                        if (++it == arr.end())
                            break;
                        out << ",\n";
                    }
                }
                out << "\n";
                indent_p->resize(indent_p->size() - 4);
                out << *indent_p << "]";
                break;
            }

            case boost::json::kind::string:
            {
                out << boost::json::serialize(jv.get_string());
                break;
            }

            case boost::json::kind::uint64:
                out << jv.get_uint64();
                break;

            case boost::json::kind::int64:
                out << jv.get_int64();
                break;

            case boost::json::kind::double_:
                out << jv.get_double();
                break;

            case boost::json::kind::bool_:
                if (jv.get_bool())
                    out << "true";
                else
                    out << "false";
                break;

            case boost::json::kind::null:
                out << "null";
                break;
        }

        if (indent_p->empty())
            out << "\n";
    }
} // namespace Json

namespace network
{
    namespace interface
    {
        addresses get_address(const std::string &if_name,
                              sa_family_t sa_family)
        {
            addresses res;
            ifaddrs *ifs;
            if (getifaddrs(&ifs))
            {
                return res;
            }
            for (auto addr = ifs; addr != nullptr; addr = addr->ifa_next)
            {
                // No address? Skip.
                if (addr->ifa_addr == nullptr)
                    continue;

                // Interface isn't active? Skip.
                if (!(addr->ifa_flags & IFF_UP))
                    continue;

                if (if_name != std::string(addr->ifa_name))
                    continue;

                if ((sa_family == addr->ifa_addr->sa_family)
                    && (addr->ifa_addr->sa_family == AF_INET))
                {
                    auto addr_p = reinterpret_cast<sockaddr_in *>(addr->ifa_addr);
                    res.emplace_back(boost::asio::ip::make_address_v4(ntohl(addr_p->sin_addr.s_addr)));
                }
                else if ((sa_family == addr->ifa_addr->sa_family)
                         && (addr->ifa_addr->sa_family == AF_INET6))
                {
                    auto addr_p = reinterpret_cast<sockaddr_in6 *>(addr->ifa_addr);
                    boost::asio::ip::address_v6::bytes_type buf;
                    memcpy(buf.data(), addr_p->sin6_addr.s6_addr, sizeof(addr_p->sin6_addr));
                    res.emplace_back(boost::asio::ip::make_address_v6(buf, addr_p->sin6_scope_id));
                }
                else
                    continue;
            }
            freeifaddrs(ifs);
            return res;
        }
    } // namespace interface

    address make_address(const std::string &address,
                         sa_family_t sa_family,
                         boost::system::error_code &err)
    {
        network::address ipAddress = boost::asio::ip::make_address(address, err);
        if (err)
        {
            LOGDEBUG(" Failed make address from [", address, "].",
                     " Try to treat it as an interface name");

            const std::string &if_name = address;

            addresses ips = interface::get_address(if_name, sa_family);
            if (ips.empty())
            {
                LOGERROR(" Failed make address from interface [", if_name, "]");
                return ipAddress;
            }
            err.clear();
            ipAddress = ips.front();
            LOGINFO("Resolved interface [", if_name, "] address of ", ipAddress);
        }
        return ipAddress;
    }

} // namespace network

namespace io_service
{
    Context::~Context()
    {
        stop();
    }

    void Context::start(int threadPoolSize)
    {
        workers_guard = std::make_shared<work_guard>(boost::asio::make_work_guard(*this));

        auto aux_thread_main{[this]()
                             {
                                 LOGDEBUG(" Aux thread started");
                                 boost::asio::io_service::run();
                                 LOGDEBUG(" Aux thread stopped");
                             }};

        // Create a pool of threads to run all of the io_contexts.
        for (int i = 0; i < threadPoolSize; ++i)
        {
            auto t = std::make_shared<std::thread>(aux_thread_main);
            m_threads.push_back(t);
        }
    }

    void Context::stop()
    {
        if (workers_guard)
        {
            workers_guard->reset();
        }
    }

    void Context::wait_stop()
    {
        // Wait for all threads in the pool to exit.
        for (std::size_t i = 0; i < m_threads.size(); ++i)
        {
            if (m_threads[i]->joinable())
            {
                m_threads[i]->join();
            }
        }
    }

} // namespace io_service


namespace Digest
{

    bool shouldSendVerifyRequest(Uvr::UvrId_t uvrId)
    {
        if ((uvrId < Uvr::UvrId::serviceMin)
            || (uvrId == Uvr::UvrId::uvzPingContinue)
            || (uvrId == Uvr::UvrId::uvzPingRelease))
        {
            return true;
        }

        return false;
    }

    bool shouldIncreaseCounters(Uvr::UvrId_t uvrId)
    {
        switch (uvrId)
        {
            case Uvr::UvrId::uvzPingContinue:
            case Uvr::UvrId::uvzPingRelease:
            {
                return false;
            }
            default:
            {
                return true;
            }
        }
    }

    bool shouldMakeIncident(Uvr::UvrId_t uvrId)
    {
        switch (uvrId)
        {
            case Uvr::UvrId::uvzPingContinue:
            case Uvr::UvrId::uvzPingRelease:
            {
                return false;
            }
            default:
            {
                return true;
            }
        }
    }

} // namespace Digest
