#pragma once

// STL includes.
#include <cstdint>
#include <functional>
#include <memory>
#include <optional>
#include <string>
#include <time.h>
#include <vector>

namespace Uvr
{
    //--------------------------------------------------------------------------
    // Type aliases:
    //--------------------------------------------------------------------------

    /// Status of KVR.
    using NodeStatus_t = uint8_t;
    enum class NodeStatus : NodeStatus_t
    {
        NotInitialized,
        Initalized,
        Starting,
        Started,
        Stopping,
        Stopped
    };

    /// String with list of pairs: MNC operator code and OrgCode organization code
    /// according to BDPN (ex. "MNC=ORG,MNC2=ORG2").
    using BdpnCode = std::string;
    using OptionalBdpnCode = std::optional<BdpnCode>;

    /// Call Identifier
    using CallId_t = std::string;

    /// Time interval in seconds to match call within.
    using CallMatchInterval_t = uint16_t;
    using OptionalCallMatchInterval = std::optional<CallMatchInterval_t>;
    namespace CallMatchInterval
    {
        constexpr CallMatchInterval_t min = 1;
        constexpr CallMatchInterval_t max = 900;
        constexpr CallMatchInterval_t def = 180;
    } // namespace CallMatchInterval

    /// Call duration in seconds
    using CallDuration_t = uint32_t;
    using OptionalCallDuration = std::optional<CallDuration_t>;

    using Date = std::string;

    /// Domain name.
    using Dns_t = std::string;

    /// Q.850 cause code.
    using ReleaseCode_t = uint8_t;
    using OptionalReleaseCode = std::optional<ReleaseCode_t>;

    /// Global Title.
    using Gt_t = std::string;

    /// Individual taxpayer number. A sequence of numbers up to 12 characters long.
    /// An empty string - if there is no number.
    using Inn = std::string;

    // IP address.
    using Ip_t = std::string;

    namespace MetaInfo
    {
        /// Represents meta information in the form of bit flags.
        /// The structure of the flags depends on the specific meta information
        /// that was converted.
        using BitFlags_t = uint32_t;

        constexpr BitFlags_t emptyBitFlags = 0;

        /// Represents meta information as a string with a set of key-value pairs
        /// in the following format "KEY1=VALUE1, KEY2=VALUE2,...".
        using String = std::string;
    }

    /// Represents names of various kinds.
    using Name = std::string;
    using OptionalName = std::optional<Name>;

    /// String representation of a phone number (ex. "79871112233").
    using Number = std::string;

    /// Identifier of the operator, according to the digest. Possible values:
    /// 0..4294967295. The range from 0 to 9999 is reserved for service purposes.
    using OperatorId_t = uint32_t;
    using OptionalOperatorId = std::optional<OperatorId_t>;
    constexpr OperatorId_t undefinedOperatorId = 0;

    using OctetString = std::vector<uint8_t>;
    using OctetStringPtr = std::shared_ptr<OctetString>;

    using RequestId_t = uint64_t;

    /// Response code on call information request.
    using ResponseCode_t = uint16_t;
    namespace ResponseCode
    {
        constexpr ResponseCode_t badRequest = 0;  ///< Bad request received.
        constexpr ResponseCode_t found = 1;       ///< Information about the requested call was found.
        constexpr ResponseCode_t notFound = 2;    ///< Information about the requested call was not found.
        constexpr ResponseCode_t systemError = 3; ///< System error during request processing.
        constexpr ResponseCode_t undefined = 4;
    } // namespace ResponseCode

    using SessionId_t = uint64_t;
    constexpr SessionId_t undefinedSessionId = 0;

    using Strings = std::vector<std::string>;

    /// Call action code with a transit operator.
    using TransitActionCode_t = uint8_t;
    using OptionalTransitActionCode = std::optional<TransitActionCode_t>;
    namespace TransitActionCode
    {
        constexpr TransitActionCode_t pass = 0;   ///< Call was transited.
        constexpr TransitActionCode_t reject = 1; ///< Call was rejected.
        constexpr TransitActionCode_t undefined = 2;
    } // namespace TransitActionCode

    using TimePoint = timespec;

    using UserId_t = uint32_t;
    constexpr UserId_t undefinedUserId = 0;

    using UvrId_t = uint16_t;
    using OptionalUvrId = std::optional<UvrId_t>;
    namespace UvrId
    {
        constexpr UvrId_t undefined = 0;
        constexpr UvrId_t serviceMin = 16001;
        constexpr UvrId_t serviceMax = 16383;

        constexpr UvrId_t numbersNotServiced = 16001;
        constexpr UvrId_t uvzPingContinue    = 16002;
        constexpr UvrId_t uvzPingRelease     = 16003;
    } // namespace UvrId

    using VerdictTrigger_t = uint8_t;
    namespace VerdictTrigger
    {
        const VerdictTrigger_t undefined = 0;
        const VerdictTrigger_t antifraud = 1;
    } // namespace VerdictTrigger

    /// Verdict on call verification request.
    using VerificationVerdict_t = int16_t;
    namespace VerificationVerdict
    {
        constexpr VerificationVerdict_t success                  = 0; ///< Call has been confirmed.
        constexpr VerificationVerdict_t failure                  = 1; ///< Call isn't found.
        constexpr VerificationVerdict_t timeout                  = 2; ///< Timeout has been exceeded.
        constexpr VerificationVerdict_t callingNumberNotRussian  = 3; ///< Calling number doesn't belong to Russian numeration plan.
        constexpr VerificationVerdict_t callingNumberNotServiced = 4; ///< Calling number isn't serviced.
        constexpr VerificationVerdict_t callingNumberNotFound    = 5; ///< Calling number isn't found.
        constexpr VerificationVerdict_t blocked                  = 6; ///< Calling number is serviced by UVR with a blocking policy.
        constexpr VerificationVerdict_t allowed                  = 7; ///< Calling number is serviced by UVR with a allowing policy.
        constexpr VerificationVerdict_t undefined                = 8;
    } // namespace VerificationVerdict

    /// Verification response on call information request.
    using VerificationResponse_t = int16_t;
    using OptionalVerificationRsp = std::optional<VerificationResponse_t>;
    namespace VerificationResponse
    {
        constexpr VerificationResponse_t notVerified              = -1;                                            ///< Verification was not performed.
        constexpr VerificationResponse_t success                  = VerificationVerdict::success;                  ///< Call is confirmed.
        constexpr VerificationResponse_t failure                  = VerificationVerdict::failure;                  ///< Call isn't found.
        constexpr VerificationResponse_t timeout                  = VerificationVerdict::timeout;                  ///< Timeout has been exceeded.
        constexpr VerificationResponse_t callingNumberNotRussian  = VerificationVerdict::callingNumberNotRussian;  ///< Calling number doesn't belong to Russian numeration plan.
        constexpr VerificationResponse_t callingNumberNotServiced = VerificationVerdict::callingNumberNotServiced; ///< Calling number isn't serviced.
        constexpr VerificationResponse_t callingNumberNotFound    = VerificationVerdict::callingNumberNotFound;    ///< Calling number isn't found.
        constexpr VerificationResponse_t blocked                  = VerificationVerdict::blocked;                  ///< Calling number is serviced by UVR with a blocking policy.
        constexpr VerificationResponse_t allowed                  = VerificationVerdict::allowed;                  ///< Calling number is serviced by UVR with a allowing policy.
        constexpr VerificationResponse_t undefined                = 8;
    } // namespace VerificationResponse

    /// Call type
    using CallType_t = uint8_t;
    namespace CallType
    {
        constexpr CallType_t undefined = 0; ///< Call type was not set
        constexpr CallType_t outgoing = 1;  ///< Outgoing call
        constexpr CallType_t incoming = 2;  ///< Incoming call

    } // namespace CallType

    //--------------------------------------------------------------------------
    // Structs:
    //--------------------------------------------------------------------------

    /// Stores call request.
    struct CallInformationRequest
    {
        Number numA;                        ///< Calling number.
        Number numB;                        ///< Called number.
        Number numC;                        ///< Origin called number.
        Number numD;                        ///< Generic number.
        Date date;                          ///< Date of connection attempt.
        RequestId_t idReq;                  ///< Call info request ID.
        OptionalOperatorId idSrc;           ///< Source operator ID.
        OptionalOperatorId idDst;           ///< Destination operator ID.
        OptionalCallMatchInterval interval; ///< Time interval in sec to match call within.
        CallId_t callId;                    ///< Call ID.
        OptionalCallDuration callDuration;  ///< Call duration.

        CallInformationRequest(const Number &numA = Number(),
                               const Number &numB = Number(),
                               const Number &numC = Number(),
                               const Number &numD = Number(),
                               const Date &date = Date(),
                               const RequestId_t &idReq = 0,
                               const OptionalOperatorId &idSrc = std::nullopt,
                               const OptionalOperatorId &idDst = std::nullopt,
                               const OptionalCallMatchInterval &interval = std::nullopt,
                               const CallId_t &callId = CallId_t(),
                               const OptionalCallDuration &callDuration = std::nullopt);

        CallInformationRequest(const char *req, const char delimeter = ';');
    };

    /// Stores requested call information.
    struct CallInformationResponse
    {
        Number numA;                           ///< Calling number.
        Number numB;                           ///< Called number.
        Number numC;                           ///< Origin called number.
        Number numD;                           ///< Generic number.
        Date date;                             ///< Date of connection attempt.
        Date dateAct;                          ///< Date of connection attempt written by UVR.
        RequestId_t idReq;                     ///< Call info request ID.
        OptionalOperatorId idSrc;              ///< Source operator ID.
        OptionalOperatorId idDst;              ///< Destination operator ID.
        ResponseCode_t rspCode;                ///< Response code for call info lookup.
        OptionalVerificationRsp vrfRsp;        ///< Verification procedure response code.
        SessionId_t sessionId;                 ///< Verification request (session) identifier.
        OptionalTransitActionCode tActionCode; ///< Call action code with a transit operator.
        OptionalReleaseCode releaseCode;       ///< Q.850 cause code.
        OptionalCallDuration callDuration;     ///< Call duration.
        OptionalUvrId idUvrT;                  ///< UVR ID which verify request was sent to.
        CallId_t callId;                       ///< Call ID.
        Uvr::VerdictTrigger_t idRel;           ///< Identifier of system made verdict.

        CallInformationResponse(const Number &numA = Number(),
                                const Number &numB = Number(),
                                const Number &numC = Number(),
                                const Number &numD = Number(),
                                const Date &date = Date(),
                                const Date &dateAct = Date(),
                                const RequestId_t &idReq = 0,
                                const OptionalOperatorId &idSrc = std::nullopt,
                                const OptionalOperatorId &idDst = std::nullopt,
                                const ResponseCode_t &rspCode = ResponseCode::undefined,
                                const OptionalVerificationRsp &vrfRsp = std::nullopt,
                                const SessionId_t &sessionId = undefinedSessionId,
                                const OptionalTransitActionCode &tActionCode = std::nullopt,
                                const OptionalReleaseCode &releaseCode = std::nullopt,
                                const OptionalCallDuration &callDuration = std::nullopt,
                                const OptionalUvrId &idUvrT = std::nullopt,
                                const CallId_t &callId = CallId_t(),
                                const Uvr::VerdictTrigger_t &idRel = Uvr::VerdictTrigger::undefined);

        bool operator==(const CallInformationResponse &rhs) const;
    };
    using CallInformationResponses = std::vector<CallInformationResponse>;

    /// Stores information about HUB.
    struct HubInfo
    {
        Gt_t id = "";            ///< Global Title of HUB
        Ip_t ipPrimary = "";     ///< IP address of primary host of HUB
        Ip_t ipSecondary = "";   ///< IP address of secondary host of HUB
        Dns_t dnsPrimary = "";   ///< Domain name of primary host of HUB
        Dns_t dnsSecondary = ""; ///< Domain name of secondary host of HUB
    };

    /// Stores information about Operator.
    struct OperatorInfo
    {
        OperatorId_t id = 0;                      ///< Identifier of the operator, according to the digest.
        Name name = "";                           ///< The name of the telecom operator. Text field with >= 1 character.
        Name nick = "";                           ///< The short name of the telecom operator. Text field with >= 1 character.
        Inn inn = "";                             ///< Individual taxpayer number of the telecom operator.
        OptionalBdpnCode bdpnCode = std::nullopt; ///< String with list of pairs: MNC operator code and OrgCode organization
                                                  ///< code according to BDPN (ex. "MNC=ORG,MNC2=ORG2").
        OptionalName brand = std::nullopt;        ///< The brand name of the telecom operator.
    };

    /// Stores data for call information by request in DB.
    struct CallRecord
    {
        Number numA = "";                                                  ///< Calling number.
        Number numB = "";                                                  ///< Called number.
        Number numC = "";                                                  ///< Origin called number.
        Number numD = "";                                                  ///< Generic number.
        TimePoint date;                                                    ///< Date of connection attempt.
        TimePoint dateAct;                                                 ///< Date of connection attempt written by UVR.
        OptionalOperatorId idSrc = std::nullopt;                           ///< Source operator ID.
        OptionalOperatorId idDst = std::nullopt;                           ///< Destination operator ID.
        SessionId_t sessionId = undefinedSessionId;                        ///< Verification request (session) identifier.
        VerdictTrigger_t idRel = VerdictTrigger::antifraud;                ///< Identifier of system made verdict.
        VerificationResponse_t response = VerificationResponse::undefined; ///< Response on verification request.
        OptionalReleaseCode releaseCode = std::nullopt;                    ///< Q.850 cause code.
        OptionalCallDuration callDuration = std::nullopt;                  ///< Call duration in seconds.
        OptionalUvrId idUvrT = std::nullopt;                               ///< ID of the UVR to which the verification request was sent.
        CallId_t callId = "";                                              ///< Call ID.
        CallType_t callType = CallType::undefined;                         ///< Call type.
        
        /// Updates the fields of this record with values from @a other if they
        /// are not empty. Only the following fields are updated:
        ///   - idSrc
        ///   - idDst
        ///   - releaseCode
        ///   - callDuration
        void update(const CallRecord &other);
    };
    using CallRecords = std::vector<CallRecord>;

    /// Stores a pair of UVR IDs, primary and secondary.
    struct UvrIds
    {
        UvrId_t primary = 0;
        UvrId_t secondary = 0;
    };

    /// Stores information about UVR.
    struct UvrInfo
    {
        /// UVR placement.
        enum class Place
        {
            /// The UVR is located outside the GRFC network.
            External = 1,

            /// The UVR is located inside the GRFC network.
            Internal = 2,

            /// The UVR is located inside the GRFC network and connected through
            /// a hardware coordinator via the ViPNet client.
            Vipnet = 3
        };
        using OptionalPlace = std::optional<Place>;

        /// UVR status.
        enum class Status
        {
            /// Normal UVR status. Does not have any effect on the response.
            /// Other flags will be applied.
            Active = 0,

            /// Indicates that the UVR is under maintenance. A Failure verdict
            /// will be replaced to Success. An incident will be created.
            /// Other flags will be ignored.
            Maintenance = 1
        };
        using OptionalStatus = std::optional<Status>;

        /// Bit mask of supported modes.
        enum Capability
        {
            Universal     = 0b0000, ///< Universal UVR.
            UvrT          = 0b0001, ///< UVR operates in UVR_T mode.
            UvrO          = 0b0010, ///< UVR operates in UVR_O mode.
            Retrospective = 0b0100, ///< UVR processes historical data.
            Transit       = 0b1000, ///< UVR for transit operator.
            All           = 0b1111
        };
        using OptionalCapability = std::optional<Capability>;

        /// UVR_O response processing policy.
        enum class DefPolicy
        {
            /// Received verdict will be sent without changes.
            NoChange = 0,

            /// The UVR_O should act as if a Failure response to the verification
            /// request has been received. In this case, an incident is generated
            /// with a @ref VerificationVerdict::blocked code.
            Blocked = 1,

            /// The UVR_O should act as if a Success response to the verification
            /// request has been received. In this case, an incident is generated
            /// with a @ref VerificationVerdict::allowed code.
            Allowed = 2,
        };
        using OptionalDefPolicy = std::optional<DefPolicy>;

        static MetaInfo::BitFlags_t parseMetaInfo(const MetaInfo::String &value);

        /// @brief Extracts @ref Place value from Meta Info.
        /// @returns @c std::nullopt if Meta Info does not contain a value for
        ///          @ref Place, otherwise returns one of the allowed values.
        static OptionalPlace getPlace(const MetaInfo::BitFlags_t metaInfoBitFlags);

        /// @brief Extracts @ref Status value from Meta Info.
        /// @returns @c std::nullopt if Meta Info does not contain a value for
        ///          @ref Status, otherwise returns one of the allowed values.
        static OptionalStatus getStatus(const MetaInfo::BitFlags_t metaInfoBitFlags);

        /// @brief Extracts @ref Capability value from Meta Info.
        /// @returns @c std::nullopt if Meta Info does not contain a value for
        ///          @ref Capability, otherwise returns one of the allowed values.
        static OptionalCapability getCapability(const MetaInfo::BitFlags_t metaInfoBitFlags);

        /// @brief Extracts @ref DefPolicy value from Meta Info.
        /// @returns @c std::nullopt if Meta Info does not contain a value for
        ///          @ref DefPolicy, otherwise returns one of the allowed values.
        static OptionalDefPolicy getDefPolicy(const MetaInfo::BitFlags_t metaInfoBitFlags);

        UvrId_t id = 0;                           ///< Identifier of UVR.
        Gt_t gt = "";                             ///< Global Title of UVR used for SCCP routing.
        Ip_t ipPrimary = "";                      ///< Primary IP address of UVR.
        Ip_t ipSecondary = "";                    ///< Secondary IP address of UVR.
        Dns_t dns = "";                           ///< Domain name of UVR.
        Gt_t hubIdPrimary = "";                   ///< Global Title of primary HUB.
        Gt_t hubIdSecondary = "";                 ///< Global Title of secondary HUB.
        Gt_t gtUvr1 = "";                         ///< Global Title for SCCP routing.
        Gt_t gtUvr2 = "";                         ///< Global Title for SCCP routing.
        OperatorId_t idSrc = undefinedOperatorId; ///< Source operator ID.
        MetaInfo::BitFlags_t metaInfo = 0;        ///< Meta Info stored as bit flags.
    };

    /// Contains result of call verification.
    struct VerificationResult
    {
        VerificationVerdict_t verificationVerdict = VerificationVerdict::success; ///< Verdict of call verification.
        SessionId_t sessionId = undefinedSessionId;                               ///< Session ID of call verification.
        UvrId_t idUvrT = UvrId::undefined;                                        ///< UVR ID which verify request was sent to.
        CallId_t callId = "";                                                     ///< Call ID.
    };


    //--------------------------------------------------------------------------
    // Callbacks:
    //--------------------------------------------------------------------------

    using CallInformationRequestCallback =
        std::function<void(const Uvr::UvrId_t uvrId,
                           const CallInformationRequest &)>;

    using StoreRecordCallback =
        std::function<bool(const Uvr::UvrId_t uvrId,
                           const CallRecord &record)>;

    using UpdateRecordsCallback =
        std::function<size_t(const Uvr::UvrId_t uvrId,
                             const CallRecords &records)>;

    using VerifyCallVerdictCallback =
        std::function<void(const VerificationResult &)>;

    //--------------------------------------------------------------------------
    // Auxiliary functions:
    //--------------------------------------------------------------------------

    template <typename T>
    std::string toString(const std::optional<T>& optValue, const std::string& defValue = "")
    {
        static_assert(std::is_integral<T>::value, "Optional value should be integer.");
        return (optValue.has_value() ? std::to_string(static_cast<int>(optValue.value())) : defValue);
    }

    std::ostream &operator<<(std::ostream &os, const CallInformationResponse &r);
    std::ostream &operator<<(std::ostream &os, const NodeStatus &status);
    std::ostream &operator<<(std::ostream &os, const CallRecord &callRecord);

} // namespace Uvr
