#include "components/worker_node/src/Codec.hpp"

// AFAP includes.
#include "components/tools/src/Tools.hpp"
#include "pm/common/src/pm_primitive.hpp"

#define HEADER_TAG "h"
#define BODY_TAG "b"
#define ID_TAG "i"
#define FROM_TAG "f"
#define TO_TAG "t"
#define ULPKEY_TAG "u"
#define FILENAME_TAG "f"
#define FILENAMES_LIST_TAG "f"
#define NUMBERS_TAG "n"
#define VERDICT_TAG "v"
#define RLC_TAG "r"
#define ID_SRC_TAG "s"
#define NUM_FILE_TAG "n"
#define DELTA_FILES_TAG "d"
#define NODES_FILES_TAG "n"
#define OPERATORS_FILES_TAG "o"
#define PKEYS_HUB_FILES_TAG "h"
#define PKEYS_NODE_FILES_TAG "n"
#define PKEYS_TOCOPY_TAG "cp"
#define PKEYS_TOREMOVE_TAG "rm"
#define NUM_A_TAG "a"
#define NUM_B_TAG "b"
#define NUM_C_TAG "c"
#define NUM_D_TAG "d"
#define ID_CONN_TAG "i"
#define SERVICE_KEY_TAG "s"
#define UVR_O_TAG "o"
#define UVR_T_TAG "t"
#define SESSION_KEY_TAG "k"
#define VERIFY_CALL_INDICATION_TAG "i"
#define RELEASE_CODE_TAG "r"
#define ADDITIONAL_CODE_TAG "a"
#define VERIFY_CALL_QUERY_TAG "q"
#define UVR_P_TAG "p"
#define UVR_S_TAG "s"
#define ERRORCODE_TAG "e"
#define CALL_ID_TAG "i"
#define NAME_TAG "nm"
#define NICK_TAG "ni"
#define INN_TAG "i"
#define BDPN_TAG "b"
#define BRAND_TAG "br"
#define ARGS_TAG "a"
#define CMD_TAG "c"
#define RSP_TAG "r"
#define STATUS_TAG "s"
#define VERIFICATION_VERDICT_TAG "v"
#define INCIDENT_VERDICT_TAG "i"

// QUERY_UVR/HUB_REQ/RSP
#define GT_TAG "g"
#define IP_P_TAG "p"
#define IP_S_TAG "s"
#define DNS_TAG "d"
#define ID_HUB_P_TAG "p"
#define ID_HUB_S_TAG "s"
#define GT_UVR_1_TAG "o"
#define GT_UVR_2_TAG "t"
#define DNS_P_TAG "p"
#define DNS_S_TAG "s"

namespace primitive
{
    namespace Codec
    {
        std::string typeToStr(const Type type)
        {
            switch (type)
            {
            case Type::Binary:
                return "Binary";
            case Type::Json:
                return "Json";
            default:
                return "Unknown(" + std::to_string(static_cast<uint16_t>(type)) + ")";
            }
        }

        raw_data_t encode(const generic *prim, const Type codecType, const bool prependSize)
        {
            static constexpr auto sizeOfCodecType = sizeof(Type);
            static constexpr auto sizeOfDataSize = sizeof(uint16_t);

            raw_data_t data(prependSize ? sizeOfDataSize + sizeOfCodecType : sizeOfCodecType,
                            static_cast<char>(codecType));

            switch (codecType)
            {
            case Type::Json:
            {
                data += Json::encode(prim);
                break;
            }
            default:
            {
                LOGERROR("primitive::Codec::encode doesn't support codec: ",
                         typeToStr(codecType));
                return "";
            }
            }

            if (prependSize)
            {
                uint8_t *buff = reinterpret_cast<uint8_t *>(&data[0]);
                uint16_t dataSize = static_cast<uint16_t>(data.size() - sizeOfDataSize);
                Binary::writeToBuff(buff, dataSize);
            }

            LOGDEBUG("Codec::encode: ", data);

            return data;
        }

        GenericPtr decode(const raw_data_t &rawData)
        {
            if (rawData.empty())
            {
                LOGERROR("primitive::Codec::decode can't decode empty data.");
                return nullptr;
            }

            const uint8_t *buff = reinterpret_cast<const uint8_t *>(&rawData[0]);
            Type codecType = Type::Unknown;
            Binary::readFromBuff(buff, codecType);

            switch (codecType)
            {
            case Type::Json:
            {
                return Json::decode(buff, rawData.size() - sizeof(Type));
            }
            default:
            {
                LOGERROR("primitive::Codec::decode doesn't support codec: ",
                         typeToStr(codecType));
                return nullptr;
            }
            }
        }

        namespace Json
        {
            raw_data_t encode(const generic *prim)
            {
                // Encode primitive header.
                boost::json::object obj;
                obj[HEADER_TAG] = encodeHeader(*prim->m_header);

                // Encode primitive body.
                switch (prim->m_header->id)
                {
                case type::PM_PROVIDER_REGISTER_REQ:
                {
                    obj[BODY_TAG] = encodePmProviderReqisterReq(prim);
                    break;
                }
                case type::GET_COUNTERS_REQ:
                {
                    obj[BODY_TAG] = encodeGetCountersReq(prim);
                    break;
                }
                case type::GET_COUNTERS_RSP:
                {
                    obj[BODY_TAG] = encodeGetCountersRsp(prim);
                    break;
                }
                case type::GET_STATISTICS_REQ:
                {
                    obj[BODY_TAG] = encodeGetStatisticsReq(prim);
                    break;
                }
                case type::GET_STATISTICS_RSP:
                {
                    obj[BODY_TAG] = encodeGetStatisticsRsp(prim);
                    break;
                }
                case type::GET_INCIDENTS_RSP:
                {
                    obj[BODY_TAG] = encodeGetIncidentsRsp(prim);
                    break;
                }
                case type::SEND_INCIDENTSA_IND:
                {
                    obj[BODY_TAG] = encodeGetIncidentsARsp(prim);
                    break;
                }
                case type::UPLOAD_NUMBERS_IND:
                {
                    obj[BODY_TAG] = encodeUploadNumbersInd(prim);
                    break;
                }
                case type::UPLOAD_NODES_IND:
                {
                    obj[BODY_TAG] = encodeUploadNodesInd(prim);
                    break;
                }
                case type::UPLOAD_OPERATORS_IND:
                {
                    obj[BODY_TAG] = encodeUploadOperatorsInd(prim);
                    break;
                }
                case type::UPLOAD_PUBLIC_KEYS_IND:
                {
                    obj[BODY_TAG] = encodeUploadPublicKeysInd(prim);
                    break;
                }
                case type::NOTIFY_CALL_INFO_REQ:
                {
                    obj[BODY_TAG] = encodeNotifyCallInfoRequest(prim);
                    break;
                }
                case type::NOTIFY_CALL_INFO_RSP:
                {
                    obj[BODY_TAG] = encodeNotifyCallInfoResponse(prim);
                    break;
                }
                case type::VERIFY_CALL_QUERY:
                {
                    obj[BODY_TAG] = encodeVerifyCallQuery(prim);
                    break;
                }
                case type::VERIFY_CALL_REQ:
                {
                    obj[BODY_TAG] = encodeVerifyCallRequest(prim);
                    break;
                }
                case type::VERIFY_CALL_IND:
                {
                    obj[BODY_TAG] = encodeVerifyCallIndication(prim);
                    break;
                }
                case type::VERIFY_CALL_CNF:
                {
                    obj[BODY_TAG] = encodeVerifyCallConfirm(prim);
                    break;
                }
                case type::VERIFY_CALL_RSP:
                {
                    obj[BODY_TAG] = encodeVerifyCallResponse(prim);
                    break;
                }
                case type::VERIFY_CALL_VRD:
                {
                    obj[BODY_TAG] = encodeVerifyCallVerdict(prim);
                    break;
                }
                case type::OUTGOING_CALL_IND:
                {
                    obj[BODY_TAG] = encodeOutgoingCallIndication(prim);
                    break;
                }
                case type::QUERY_UVR_IDS_REQ:
                {
                    obj[BODY_TAG] = encodeQueryUvrIdsReq(prim);
                    break;
                }
                case type::QUERY_UVR_IDS_RSP:
                {
                    obj[BODY_TAG] = encodeQueryUvrIdsRsp(prim);
                    break;
                }
                case type::QUERY_UVR_INFO_REQ:
                {
                    obj[BODY_TAG] = encodeQueryUvrInfoReq(prim);
                    break;
                }
                case type::QUERY_UVR_INFO_RSP:
                {
                    obj[BODY_TAG] = encodeQueryUvrInfoRsp(prim);
                    break;
                }
                case type::QUERY_HUB_INFO_REQ:
                {
                    obj[BODY_TAG] = encodeQueryHubInfoReq(prim);
                    break;
                }
                case type::QUERY_HUB_INFO_RSP:
                {
                    obj[BODY_TAG] = encodeQueryHubInfoRsp(prim);
                    break;
                }
                case type::QUERY_OPERATOR_INFO_REQ:
                {
                    obj[BODY_TAG] = encodeQueryOperatorInfoReq(prim);
                    break;
                }
                case type::QUERY_OPERATOR_INFO_RSP:
                {
                    obj[BODY_TAG] = encodeQueryOperatorInfoRsp(prim);
                    break;
                }
                case type::EXEC_CMD_REQ:
                {
                    obj[BODY_TAG] = encodeExecCmdReq(prim);
                    break;
                }
                case type::EXEC_CMD_RSP:
                {
                    obj[BODY_TAG] = encodeExecCmdRsp(prim);
                    break;
                }
                case type::STATUS_CHANGE_IND:
                {
                    obj[BODY_TAG] = encodeStatusChangeInd(prim);
                    break;
                }
                case type::WHOAMI:
                case type::GET_INCIDENTS_REQ:
                case type::VERIFY_REQ_ABORT:
                {
                    // Body is empty, nothing to encode
                    break;
                }
                default:
                {
                    LOGERROR("primitive::Codec::Json::encode doesn't support primitive: ",
                             id2str(prim->m_header->id));
                    return "";
                }
                }

                LOGDEBUG("Json::encode: ", boost::json::serialize(obj));
                return boost::json::serialize(obj);
            }

            GenericPtr decode(const uint8_t *buff, const size_t buffSize)
            {
                std::string data(reinterpret_cast<const char *>(buff), buffSize);

                boost::json::error_code errorCode;
                boost::json::value dataValue = boost::json::parse(data, errorCode);
                if (errorCode)
                {
                    LOGERROR("primitive::Codec::Json::decode can't parse data.",
                             "\nData: ", data);
                    return nullptr;
                }

                // Decode primitive header.
                const boost::json::object &dataObject = dataValue.get_object();
                const boost::json::value *headerValue = dataObject.if_contains(HEADER_TAG);
                if (headerValue == nullptr)
                {
                    LOGERROR("primitive::Codec::Json::decode can't parse primitive header.",
                             "\nData: ", data);
                    return nullptr;
                }
                auto header = decodeHeader(headerValue);

                // Decode primitive body.
                const boost::json::value *bodyValue = dataObject.if_contains(BODY_TAG);

                GenericPtr prim;
                switch (header->id)
                {
                case type::WHOAMI:
                {
                    prim = std::make_shared<Whoami>();
                    break;
                }
                case type::GET_STATISTICS_REQ:
                {
                    prim = decodeGetStatisticsReq(bodyValue);
                    break;
                }
                case type::GET_STATISTICS_RSP:
                {
                    prim = decodeGetStatisticsRsp(bodyValue);
                    break;
                }
                case type::EXEC_CMD_REQ:
                {
                    prim = decodeExecCmdReq(bodyValue);
                    break;
                }
                case type::EXEC_CMD_RSP:
                {
                    prim = decodeExecCmdRsp(bodyValue);
                    break;
                }
                default:
                {
                    LOGERROR("primitive::Codec::Json::decode doesn't support primitive: ",
                             id2str(header->id));
                    return nullptr;
                }
                }

                if (prim == nullptr)
                {
                    LOGERROR("primitive::Codec::Json::decode can't parse body.",
                             "\nId: ", id2str(header->id), ";",
                             " From: ", header->from, ";",
                             " To: ", header->to, ";",
                             "\nBody: ", boost::json::serialize(*bodyValue));
                    return nullptr;
                }

                prim->m_header = header;
                return prim;
            }

            boost::json::value encodeHeader(const generic::Header &header)
            {
                boost::json::object headerObject;
                headerObject[ID_TAG] = header.id;
                headerObject[FROM_TAG] = boost::json::value_from(header.from);
                headerObject[TO_TAG] = boost::json::value_from(header.to);
                headerObject[ULPKEY_TAG] = header.ulpKey;

                return headerObject;
            }

            generic::HeaderPtr decodeHeader(const boost::json::value *headerValue)
            {
                boost::json::object headerObject = headerValue->as_object();
                return std::make_shared<generic::Header>(boost::json::value_to<uint16_t>(headerObject.at(ID_TAG)),
                                                         boost::json::value_to<TypeInst>(headerObject.at(FROM_TAG)),
                                                         boost::json::value_to<TypeInst>(headerObject.at(TO_TAG)),
                                                         boost::json::value_to<generic::ulpKey_t>(headerObject.at(ULPKEY_TAG)));
            }

            boost::json::value encodePmProviderReqisterReq(const generic *prim)
            {
                auto req = dynamic_cast<const pm::primitive::ProviderRegisterReq *>(prim);

                boost::json::object bodyObject;
                bodyObject[COUNTERS_TAG] = boost::json::value_from(req->counters);

                return bodyObject;
            }

            boost::json::value encodeGetCountersReq(const generic *prim)
            {
                auto req = dynamic_cast<const pm::primitive::GetCountersReq *>(prim);

                boost::json::object bodyObject;
                bodyObject[RESET_TAG] = req->reset;
                bodyObject[COUNTERS_TAG] = boost::json::value_from(req->counters);

                return bodyObject;
            }

            boost::json::value encodeGetCountersRsp(const generic *prim)
            {
                auto rsp = dynamic_cast<const pm::primitive::GetCountersRsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[COUNTERS_TAG] = boost::json::value_from(rsp->counters);

                return bodyObject;
            }
            boost::json::value encodeGetStatisticsReq(const generic *prim)
            {
                auto req = dynamic_cast<const pm::primitive::GetStatisticsReq *>(prim);

                boost::json::object bodyObject;
                bodyObject[RESET_TAG] = req->reset;
                bodyObject[COUNTERS_TAG] = boost::json::value_from(req->counters);

                return bodyObject;
            }

            boost::json::value encodeGetStatisticsRsp(const generic *prim)
            {
                auto rsp = dynamic_cast<const pm::primitive::GetStatisticsRsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[COUNTERS_TAG] = boost::json::value_from(rsp->counters);
                if (rsp->errorCode != pm::primitive::GetStatisticsRsp::ErrorCode::noError)
                {
                    bodyObject[ERRORCODE_TAG] = boost::json::value_from(rsp->errorCode);
                }

                return bodyObject;
            }

            boost::json::value encodeGetIncidentsRsp(const generic *prim)
            {
                auto req = dynamic_cast<const ::primitive::GetIncidentsRsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[FILENAME_TAG] = req->filename;

                return bodyObject;
            }

            boost::json::value encodeGetIncidentsARsp(const generic *prim)
            {
                auto req = dynamic_cast<const ::primitive::GetIncidentsARsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[FILENAME_TAG] = req->filename;

                return bodyObject;
            }

            boost::json::value encodeUploadNumbersInd(const generic *prim)
            {
                auto ind = dynamic_cast<const ::primitive::UploadNumbersInd *>(prim);

                boost::json::object bodyObject;
                bodyObject[NUM_FILE_TAG] = boost::json::value_from(ind->numFile);
                bodyObject[DELTA_FILES_TAG] = boost::json::value_from(ind->deltas);

                return bodyObject;
            }

            boost::json::value encodeUploadNodesInd(const generic *prim)
            {
                auto ind = dynamic_cast<const ::primitive::UploadNodesInd *>(prim);

                boost::json::object bodyObject;
                bodyObject[NODES_FILES_TAG] = boost::json::value_from(ind->files);

                return bodyObject;
            }

            boost::json::value encodeUploadOperatorsInd(const generic *prim)
            {
                auto ind = dynamic_cast<const ::primitive::UploadOperatorsInd *>(prim);

                boost::json::object bodyObject;
                bodyObject[OPERATORS_FILES_TAG] = boost::json::value_from(ind->files);

                return bodyObject;
            }

            boost::json::value encodeUploadPublicKeysInd(const generic *prim)
            {
                auto ind = dynamic_cast<const ::primitive::UploadPublicKeysInd *>(prim);

                boost::json::object bodyObject;
                bodyObject[PKEYS_HUB_FILES_TAG].emplace_object()[PKEYS_TOCOPY_TAG] = boost::json::value_from(ind->hubFilesToCopy);
                bodyObject[PKEYS_HUB_FILES_TAG].get_object()[PKEYS_TOREMOVE_TAG] = boost::json::value_from(ind->hubFilesToRemove);
                bodyObject[PKEYS_NODE_FILES_TAG].emplace_object()[PKEYS_TOCOPY_TAG] = boost::json::value_from(ind->nodeFilesToCopy);
                bodyObject[PKEYS_NODE_FILES_TAG].get_object()[PKEYS_TOREMOVE_TAG] = boost::json::value_from(ind->nodeFilesToRemove);

                return bodyObject;
            }

            boost::json::value encodeNotifyCallInfoRequest(const generic *prim)
            {
                auto req = dynamic_cast<const ::primitive::NotifyCallInfoRequest *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_TAG] = boost::json::value_from(req->uvrId);
                bodyObject[FILENAMES_LIST_TAG] = boost::json::value_from(req->fileNamesList);

                return bodyObject;
            }

            boost::json::value encodeNotifyCallInfoResponse(const generic *prim)
            {
                auto rsp = dynamic_cast<const primitive::NotifyCallInfoResponse *>(prim);

                boost::json::object bodyObject;
                bodyObject[FILENAME_TAG] = boost::json::value_from(rsp->filename);

                return bodyObject;
            }

            boost::json::value encodeVerifyCallQuery(const generic *prim)
            {
                auto query = dynamic_cast<const primitive::VerifyCallQuery *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_SRC_TAG] = boost::json::value_from(query->idSrc);
                bodyObject[NUM_A_TAG] = boost::json::value_from(doText(query->callingNumber));
                bodyObject[NUM_B_TAG] = boost::json::value_from(doText(query->calledNumber));
                if (query->originalCalledNumber)
                {
                    bodyObject[NUM_C_TAG] = boost::json::value_from(doText(query->originalCalledNumber));
                }
                if (query->genericNumber)
                {
                    bodyObject[NUM_D_TAG] = boost::json::value_from(doText(query->genericNumber));
                }
                bodyObject[CALL_ID_TAG] = boost::json::value_from(query->callId);

                return bodyObject;
            }

            boost::json::value encodeVerifyCallRequest(const generic *prim)
            {
                auto req = dynamic_cast<const primitive::VerifyCallRequest *>(prim);

                boost::json::object bodyObject;
                bodyObject[VERIFY_CALL_QUERY_TAG] = encodeVerifyCallQuery(req->verifyCallQuery.get());
                bodyObject[UVR_P_TAG] = boost::json::value_from(req->uvrP);
                bodyObject[UVR_S_TAG] = boost::json::value_from(req->uvrS);

                return bodyObject;
            }

            boost::json::value encodeVerifyCallIndication(const generic *prim)
            {
                auto ind = dynamic_cast<const primitive::VerifyCallIndication *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_CONN_TAG] = boost::json::value_from(ind->connId);
                bodyObject[SERVICE_KEY_TAG] = boost::json::value_from(ind->serviceKey);
                bodyObject[UVR_O_TAG] = boost::json::value_from(ind->uvrO);
                bodyObject[UVR_T_TAG] = boost::json::value_from(ind->uvrT);
                bodyObject[SESSION_KEY_TAG] = boost::json::value_from(ind->sessionKey);
                bodyObject[NUM_A_TAG] = boost::json::value_from(doText(ind->callingNumber));
                bodyObject[NUM_B_TAG] = boost::json::value_from(doText(ind->calledNumber));
                if (ind->originalCalledNumber)
                {
                    bodyObject[NUM_C_TAG] = boost::json::value_from(doText(ind->originalCalledNumber));
                }

                return bodyObject;
            }

            boost::json::value encodeVerifyCallConfirm(const generic *prim)
            {
                auto cnf = dynamic_cast<const primitive::VerifyCallConfirm *>(prim);

                boost::json::object bodyObject;
                bodyObject[VERIFY_CALL_INDICATION_TAG] = encodeVerifyCallIndication(cnf->verifyCallIndication.get());
                bodyObject[RELEASE_CODE_TAG] = boost::json::value_from(cnf->releaseCode);
                bodyObject[ADDITIONAL_CODE_TAG] = boost::json::value_from(cnf->additionalCode);
                bodyObject[ERRORCODE_TAG] = boost::json::value_from(cnf->errorCode);

                return bodyObject;
            }

            boost::json::value encodeVerifyCallResponse(const generic *prim)
            {
                auto rsp = dynamic_cast<const primitive::VerifyCallResponse *>(prim);

                boost::json::object bodyObject;
                bodyObject[VERIFICATION_VERDICT_TAG] = boost::json::value_from(rsp->verificationVerdict);
                bodyObject[INCIDENT_VERDICT_TAG] = boost::json::value_from(rsp->incidentVerdict);
                bodyObject[UVR_T_TAG] = boost::json::value_from(rsp->uvrT);

                return bodyObject;
            }

            boost::json::value encodeVerifyCallVerdict(const generic *prim)
            {
                auto verdict = dynamic_cast<const primitive::VerifyCallVerdict *>(prim);

                boost::json::object bodyObject;
                bodyObject[RLC_TAG] = boost::json::value_from(verdict->rlc);
                bodyObject[UVR_T_TAG] = boost::json::value_from(verdict->uvrT);

                return bodyObject;
           }

            boost::json::value encodeOutgoingCallIndication(const generic *prim)
            {
                auto ind = dynamic_cast<const primitive::OutgoingCallIndication *>(prim);

                boost::json::object bodyObject;
                bodyObject[NUM_A_TAG] = boost::json::value_from(doText(ind->callingNumber));
                bodyObject[NUM_B_TAG] = boost::json::value_from(doText(ind->calledNumber));

                return bodyObject;
            }

            boost::json::value encodeQueryUvrIdsReq(const generic *prim)
            {
                auto req = dynamic_cast<const ::primitive::QueryUvrIdsReq *>(prim);

                boost::json::object bodyObject;
                bodyObject[NUM_A_TAG] = boost::json::value_from(doText(req->number));

                return bodyObject;
            }

            boost::json::value encodeQueryUvrIdsRsp(const generic *prim)
            {
                auto rsp = dynamic_cast<const ::primitive::QueryUvrIdsRsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[UVR_P_TAG] = boost::json::value_from(rsp->uvrIds.primary);
                bodyObject[UVR_S_TAG] = boost::json::value_from(rsp->uvrIds.secondary);
                return bodyObject;
            }

            boost::json::value encodeQueryUvrInfoReq(const generic *prim)
            {
                auto req = dynamic_cast<const ::primitive::QueryUvrInfoReq *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_TAG] = boost::json::value_from(req->uvrId);

                return bodyObject;
            }

            boost::json::value encodeQueryUvrInfoRsp(const generic *prim)
            {
                auto rsp = dynamic_cast<const ::primitive::QueryUvrInfoRsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_TAG]       = boost::json::value_from(rsp->uvrInfo.id);
                bodyObject[GT_TAG]       = boost::json::value_from(rsp->uvrInfo.gt);
                bodyObject[IP_P_TAG]     = boost::json::value_from(rsp->uvrInfo.ipPrimary);
                bodyObject[IP_S_TAG]     = boost::json::value_from(rsp->uvrInfo.ipSecondary);
                bodyObject[DNS_TAG]      = boost::json::value_from(rsp->uvrInfo.dns);
                bodyObject[ID_HUB_P_TAG] = boost::json::value_from(rsp->uvrInfo.hubIdPrimary);
                bodyObject[ID_HUB_S_TAG] = boost::json::value_from(rsp->uvrInfo.hubIdSecondary);
                bodyObject[GT_UVR_1_TAG] = boost::json::value_from(rsp->uvrInfo.gtUvr1);
                bodyObject[GT_UVR_2_TAG] = boost::json::value_from(rsp->uvrInfo.gtUvr2);
                bodyObject[ID_SRC_TAG]   = boost::json::value_from(rsp->uvrInfo.idSrc);

                return bodyObject;
            }

            boost::json::value encodeQueryHubInfoReq(const generic *prim)
            {
                auto req = dynamic_cast<const ::primitive::QueryHubInfoReq *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_TAG] = boost::json::value_from(req->hubId);

                return bodyObject;
            }

            boost::json::value encodeQueryHubInfoRsp(const generic *prim)
            {
                auto rsp = dynamic_cast<const ::primitive::QueryHubInfoRsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_TAG]    = boost::json::value_from(rsp->hubInfo.id);
                bodyObject[IP_P_TAG]  = boost::json::value_from(rsp->hubInfo.ipPrimary);
                bodyObject[IP_S_TAG]  = boost::json::value_from(rsp->hubInfo.ipSecondary);
                bodyObject[DNS_P_TAG] = boost::json::value_from(rsp->hubInfo.dnsPrimary);
                bodyObject[DNS_S_TAG] = boost::json::value_from(rsp->hubInfo.dnsSecondary);

                return bodyObject;
            }

            boost::json::value encodeQueryOperatorInfoReq(const generic *prim)
            {
                auto req = dynamic_cast<const ::primitive::QueryOperatorInfoReq *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_TAG] = boost::json::value_from(req->operatorId);

                return bodyObject;
            }

            boost::json::value encodeQueryOperatorInfoRsp(const generic *prim)
            {
                auto rsp = dynamic_cast<const ::primitive::QueryOperatorInfoRsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[ID_TAG]    = boost::json::value_from(rsp->operatorInfo.id);
                bodyObject[NAME_TAG]  = boost::json::value_from(rsp->operatorInfo.name);
                bodyObject[NICK_TAG]  = boost::json::value_from(rsp->operatorInfo.nick);
                bodyObject[INN_TAG]   = boost::json::value_from(rsp->operatorInfo.inn);
                bodyObject[BDPN_TAG]  = boost::json::value_from(rsp->operatorInfo.bdpnCode);
                bodyObject[BRAND_TAG] = boost::json::value_from(rsp->operatorInfo.brand);

                return bodyObject;
            }

            boost::json::value encodeExecCmdReq(const generic *prim)
            {
                auto req = dynamic_cast<const ::primitive::ExecCmdReq *>(prim);

                boost::json::object bodyObject;
                bodyObject[CMD_TAG] = req->cmd;
                bodyObject[ARGS_TAG] = boost::json::value_from(req->cmdArgs);

                return bodyObject;
            }

            boost::json::value encodeExecCmdRsp(const generic *prim)
            {
                auto rsp = dynamic_cast<const ::primitive::ExecCmdRsp *>(prim);

                boost::json::object bodyObject;
                bodyObject[RSP_TAG] = boost::json::value_from(rsp->responses);

                return bodyObject;
            }

            boost::json::value encodeRegisterCmdInd(const generic *prim)
            {
                auto ind = dynamic_cast<const ::primitive::RegisterCmdInd *>(prim);

                boost::json::object bodyObject;
                bodyObject[CMD_TAG] = ind->cmd;

                return bodyObject;
            }

            boost::json::value encodeStatusChangeInd(const generic *prim)
            {
                auto ind = dynamic_cast<const ::primitive::StatusChangeInd *>(prim);

                boost::json::object bodyObject;
                bodyObject[STATUS_TAG] = (Uvr::NodeStatus_t)ind->status;

                return bodyObject;
            }

            GenericPtr decodeGetStatisticsReq(const boost::json::value *bodyValue)
            {
                bool reset = false;
                pm::counter::vectorFullIDs counters;
                if (!parseValue(bodyValue, RESET_TAG, reset) ||
                    !parseValue(bodyValue, COUNTERS_TAG, counters))
                {
                    return nullptr;
                }

                return std::make_shared<pm::primitive::GetStatisticsReq>(reset, counters);
            }

            GenericPtr decodeGetStatisticsRsp(const boost::json::value *bodyValue)
            {
                pm::counter::container counters;
                pm::primitive::GetStatisticsRsp::ErrorCode errorCode{pm::primitive::GetStatisticsRsp::ErrorCode::noError};
                if (!parseValue(bodyValue, COUNTERS_TAG, counters))
                {
                    return nullptr;
                }
                parseOptionalValue(bodyValue, ERRORCODE_TAG, errorCode);

                return std::make_shared<pm::primitive::GetStatisticsRsp>(counters, errorCode);
            }

            GenericPtr decodeExecCmdReq(const boost::json::value *bodyValue)
            {
                std::string cmd;
                Uvr::Strings cmdArgs;
                if (!parseValue(bodyValue, CMD_TAG, cmd) ||
                    !parseValue(bodyValue, ARGS_TAG, cmdArgs))
                {
                    return nullptr;
                }

                return std::make_shared<primitive::ExecCmdReq>(cmd, cmdArgs);
            }

            GenericPtr decodeExecCmdRsp(const boost::json::value *bodyValue)
            {
                Uvr::Strings rsps;
                if (!parseValue(bodyValue, RSP_TAG, rsps))
                {
                    return nullptr;
                }

                return std::make_shared<primitive::ExecCmdRsp>(std::move(rsps));
            }

            GenericPtr decodeRegisterCmdInd(const boost::json::value *bodyValue)
            {
                std::string cmd;
                if (!parseValue(bodyValue, CMD_TAG, cmd))
                {
                    return nullptr;
                }

                return std::make_shared<primitive::RegisterCmdInd>(cmd);
            }

        } // namespace Json

    } // namespace Codec

} // namespace primitive
