#include "oam/src/node.hpp"

// AFAP includes
#include "components/worker_node/src/Codec.hpp"

namespace oam
{

    node::node(const TypeInst &ownInst,
               boost::asio::io_context &ioc,
               const configuration::manager &config)
        : ::pm::node(ownInst, ioc, config),
          m_pm_trans_mngr(ownInst, PM_PROVIDER_MNGR_ID, m_uvrId)
    {
    }

    void node::start()
    {
        ::pm::node::start();
        registerCliCommands();
    }

    void node::prepareCounters()
    {
        addCounter(::pm::counter::oam::providers, ::pm::counter::indexing::no);
    }

    void node::handleSpecificPrimitive(const primitive::GenericPtr &prim)
    {
        switch (prim->m_header->id)
        {
        case ::primitive::PM_PROVIDER_REGISTER_REQ:
            handle_pmProviderRegisterReq(prim);
            break;
        case ::primitive::GET_STATISTICS_REQ:
            handle_getStatisticsReq(prim);
            break;
        case ::primitive::GET_COUNTERS_RSP:
            handle_getCountersRsp(prim);
            break;
        case ::primitive::EXEC_CMD_REQ:
            handle_execCmdReq(prim);
            break;
        default:
            ::pm::node::handleSpecificPrimitive(prim);
        }
    }

    void node::handle_pmProviderRegisterReq(const primitive::GenericPtr &prim)
    {
        auto req = std::dynamic_pointer_cast<::pm::primitive::ProviderRegisterReq>(prim);
        if (!req)
        {
            LOGERROR(m_ownTypeInst, " Cannot add ", prim->m_header->from, " as PM Provider");
            return;
        }

        std::lock_guard<std::recursive_mutex> guard(m_guard);
        m_pm_providers.emplace(req->m_header->from, std::make_shared<pm::provider>(req));
        LOGDEBUG(m_ownTypeInst, " The node ", req->m_header->from, " was added as PM Provider");

        setCounter(::pm::counter::oam::providers,
                   ::pm::counter::value(static_cast<::pm::counter::value_t>(m_pm_providers.size())));
    }

    void node::delete_pmProvider(TypeInst inst)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        m_pm_providers.erase(inst);

        setCounter(::pm::counter::oam::providers,
                   ::pm::counter::value(static_cast<::pm::counter::value_t>(m_pm_providers.size())));
    }

    void node::handle_getStatisticsReq(const primitive::GenericPtr &prim, OutputFormat outputFormat)
    {
        auto req = std::dynamic_pointer_cast<::pm::primitive::GetStatisticsReq>(prim);
        if (!req)
        {
            LOGERROR(m_ownTypeInst,
                     " Cannot handle ", ::primitive::id2str(prim->m_header->id), " from ", prim->m_header->from);
            return;
        }

        std::vector<std::shared_ptr<::pm::primitive::GetCountersReq>> prims;

        std::unique_lock<std::recursive_mutex> guard(m_guard);
        for (auto const &provider : m_pm_providers)
        {
            std::shared_ptr<::pm::primitive::GetCountersReq> msg =
                provider.second->buildGetCountersReq(req);
            if (msg != nullptr)
            {
                prims.push_back(msg);
            }
        }
        guard.unlock();

        if (prims.size())
        {
            ::transaction::errors err;
            ::transaction::id transId =
                m_pm_trans_mngr.openTransaction(pm::transaction::data(prim->m_header,
                                                                      (int)prims.size(),
                                                                      outputFormat),
                                                err);
            if (err == ::transaction::errors::OK)
            {
                for (auto const &msg : prims)
                {
                    msg->m_header->ulpKey = transId;
                    workers::node::send(msg);
                }
            }
            else
            {
                auto msg = std::shared_ptr<::pm::primitive::GetStatisticsRsp>(
                    new ::pm::primitive::GetStatisticsRsp({},
                                                          ::pm::primitive::GetStatisticsRsp::ErrorCode::transactionError));
                msg->m_header->to = prim->m_header->from;
                msg->m_header->ulpKey = prim->m_header->ulpKey;
                sendRsp(msg);
            }
        }
        else
        {
            auto msg = std::shared_ptr<::pm::primitive::GetStatisticsRsp>(
                new ::pm::primitive::GetStatisticsRsp({},
                                                      ::pm::primitive::GetStatisticsRsp::ErrorCode::noProvidersError));
            msg->m_header->to = prim->m_header->from;
            msg->m_header->ulpKey = prim->m_header->ulpKey;
            sendRsp(msg);
        }
    }

    void node::handle_getCountersRsp(const primitive::GenericPtr &prim)
    {
        auto rsp = std::dynamic_pointer_cast<::pm::primitive::GetCountersRsp>(prim);
        if (!rsp)
        {
            LOGERROR(m_ownTypeInst,
                     " Cannot handle ", ::primitive::id2str(prim->m_header->id), " from ", prim->m_header->from);
            return;
        }

        pm::transaction::data *pm_data_p = m_pm_trans_mngr.get_transaction_data(prim->m_header->ulpKey);
        OutputFormat outputFormat = pm_data_p->m_outputFormat;
        if (pm_data_p)
        {
            std::unique_lock<std::recursive_mutex> guard(pm_data_p->m_guard);
            for (auto counterIter = rsp->counters.begin(); counterIter != rsp->counters.end(); counterIter++)
            {
                if (counterIter->second.first == ANY_INDEX)
                {
                    continue;
                }

                auto rangeIter = pm_data_p->m_accum_storage.equal_range(counterIter->first);
                // The counter is received first time
                if (rangeIter.first == pm_data_p->m_accum_storage.end())
                {
                    pm_data_p->m_accum_storage.emplace(counterIter->first, counterIter->second);
                    continue;
                }
                auto indexIter = rangeIter.first;
                for (; indexIter != rangeIter.second; indexIter++)
                {
                    // The counter with such index already exists
                    if (indexIter->second.first == counterIter->second.first)
                    {
                        indexIter->second.second += counterIter->second.second;
                        break;
                    }
                }
                // The counter with such index is received first time
                if (indexIter == rangeIter.second)
                {
                    pm_data_p->m_accum_storage.emplace(counterIter->first, counterIter->second);
                }
            }
            pm_data_p->m_numOfWaitProviders--;
            if (pm_data_p->m_numOfWaitProviders == 0)
            {
                guard.unlock();

                ::transaction::errors err;
                pm::transaction::data pm_data = m_pm_trans_mngr.closeTransaction(prim->m_header->ulpKey, err);
                if (err == ::transaction::errors::OK)
                {
                    auto msg = std::make_shared<::pm::primitive::GetStatisticsRsp>(pm_data.m_accum_storage);
                    msg->m_header->to = pm_data.m_header->from;
                    msg->m_header->ulpKey = pm_data.m_header->ulpKey;
                    sendRsp(msg, outputFormat);
                }
            }
        }
    }

    void node::handle_execCmdReq(const primitive::GenericPtr &prim)
    {
        auto req = std::dynamic_pointer_cast<::primitive::ExecCmdReq>(prim);

        if (req->cmd == CliCmds::printCounters)
        {
            handleCmdPrintCounters(req);
            return;
        }

        sendExecCmdRsp(req->m_header->from, {"OAM: not known command."});
    }

    void node::handleCmdPrintCounters(const std::shared_ptr<primitive::ExecCmdReq> &req)
    {
        Uvr::Strings rsps;
        Uvr::Strings args = req->cmdArgs;
        std::string rspStr = std::string("Wrong '") + CliCmds::printCounters + std::string("' format!");
        std::string usage = std::string("Usage: ... ") + CliCmds::printCounters +
                            std::string(" <Text|Json> <counters_request_as_json_string>");

        if ((args.size() >= 1) &&
            ((args.at(0) == "-h") || (args.at(0) == "-help") || (args.at(0) == "--help")))
        {
            rsps.emplace_back(usage);
            sendExecCmdRsp(req->m_header->from, std::move(rsps));
            return;
        }

        rsps.emplace_back(rspStr);

        if (args.size() != 2)
        {
            rsps.emplace_back("Wrong number of arguments!");
            sendExecCmdRsp(req->m_header->from, std::move(rsps));
            return;
        }

        std::string outputFormatStr = args.at(0);
        OutputFormat outputFormat;
        if ((outputFormatStr == "Text") || (outputFormatStr == "text"))
        {
            outputFormat = OutputFormat::Text;
        }
        else if ((outputFormatStr == "Json") || (outputFormatStr == "json"))
        {
            outputFormat = OutputFormat::Json;
        }
        else
        {
            rsps.emplace_back("Counters output format can be 'Text|text' or 'Json|json' only.");
            sendExecCmdRsp(req->m_header->from, std::move(rsps));
            return;
        }

        std::string countersRequest = args.at(1);
        boost::json::error_code errorCode;
        boost::json::value dataValue = boost::json::parse(countersRequest, errorCode);
        if (errorCode)
        {
            LOGE("Json can't parse string.", "\nData: ", countersRequest);
            rsps.emplace_back("Json can't parse string. Data:");
            rsps.emplace_back(countersRequest);
            sendExecCmdRsp(req->m_header->from, std::move(rsps));
            return;
        }

        bool reset = false;
        ::pm::counter::vectorFullIDs counters;
        if (!primitive::Codec::Json::parseValue(&dataValue, RESET_TAG, reset))
        {
            LOGE("Failed get RESET_TAG");
            rsps.emplace_back("Failed get RESET_TAG");
            sendExecCmdRsp(req->m_header->from, std::move(rsps));
            return;
        }
        configuration::counter::list countersList;
        if (!primitive::Codec::Json::parseValue(&dataValue, COUNTERS_TAG, countersList))
        {
            LOGE("Failed get COUNTERS_TAG");
            rsps.emplace_back("Failed get COUNTERS_TAG");
            sendExecCmdRsp(req->m_header->from, std::move(rsps));
            return;
        }

        ::pm::counter::list2fullIDs(countersList, counters);
        auto prim = std::make_shared<::pm::primitive::GetStatisticsReq>(reset, counters);
        prim->m_header->to = TypeInst();
        prim->m_header->from = req->m_header->from;
        handle_getStatisticsReq(prim, outputFormat);
    }

    void node::sendExecCmdRsp(const TypeInst &to, Uvr::Strings &&rsps)
    {
        auto rsp = std::make_shared<primitive::ExecCmdRsp>(std::move(rsps));
        rsp->m_header->to = to;
        workers::node::send(rsp);
    }

    void node::sendRsp(const std::shared_ptr<::pm::primitive::GetStatisticsRsp> &getStatRsp,
                       OutputFormat outputFormat)
    {
        std::string rspStr;
        Uvr::Strings rsps;
        std::stringstream ss;

        if (getStatRsp->m_header->to.getType() != workers::type::CLIGW)
        {
            workers::node::send(getStatRsp);
            return;
        }

        if (outputFormat == OutputFormat::Text)
        {
            for (auto iter = getStatRsp->counters.begin(); iter != getStatRsp->counters.end(); iter++)
            {
                ss << std::setw(40) << ::pm::counter::all_ids.at(iter->first) << "|"
                   << std::setw(15) << iter->second.first << "|"
                   << std::setw(10) << iter->second.second << "|";
                rspStr = ss.str();
                rsps.emplace_back(rspStr);
                ss.clear();
            }
        }
        else if (outputFormat == OutputFormat::Json)
        {
            boost::json::object bodyObject;
            bodyObject[COUNTERS_TAG] = boost::json::value_from(getStatRsp->counters);
            ss << "||" << boost::json::serialize(bodyObject);
            rspStr = ss.str();
            rsps.emplace_back(rspStr);
        }

        sendExecCmdRsp(getStatRsp->m_header->to, std::move(rsps));
    }

    void node::registerCliCommands()
    {
        auto ind = std::make_shared<::primitive::RegisterCmdInd>(CliCmds::printCounters);
        ind->m_header->to = TypeInst(workers::type::CLIGW, 0);
        workers::node::send(ind);
    }

} // namespace oam
