#include "pm/manager/src/pm_manager.hpp"

// AFAP includes.
#include "components/csv/src/CsvFile.hpp"
#include "components/filesystem/src/filesystem.hpp"
#include "components/worker_node/src/config_parameters.hpp"
#include "pm/common/src/counters_map.hpp"

// STL includes.
#include <fstream>

namespace pm
{
    manager::manager(const TypeInst &nodeId,
                     ::transaction::manager_id trans_mngr_id,
                     boost::asio::io_context &ioc,
                     const configuration::manager &config)
        : m_ioc(ioc),
          m_config(config),
          m_trans_mngr(nodeId, trans_mngr_id, 0) // To be set later in constructor of a parent object
    {
        m_uvrId = m_config.getValue<Uvr::UvrId_t>(UVR_ID,
                                                  Uvr::undefinedUserId,
                                                  Log::Level::Error);
        if (m_uvrId == Uvr::undefinedUserId)
        {
            LOGERROR(" Cannot read UVR_ID");
            return;
        }
        m_trans_mngr.setUvrId(m_uvrId);
    }

    manager::~manager()
    {
        stop_and_del_all_jobs();
    }

    std::shared_ptr<job> manager::create_job(const std::string &job_name)
    {
        // Get duration for pm job
        job::period duration = m_config.getValue<timers::Duration>(PMJOB_PERIOD(job_name));
        if (duration == job::period::zero())
        {
            LOGERROR(getTypeInst(), " Cannot get duration for \"", job_name, "\" pm job");
            return nullptr;
        }
        // Get counters list
        configuration::counter::list countersList =
            m_config.getList<configuration::counter::entity>(PMJOB_COUNTERS(job_name));
        if (countersList.empty())
        {
            LOGERROR(getTypeInst(), " Cannot get counters for \"", job_name, "\" pm job");
            return nullptr;
        }
        counter::vectorFullIDs counters;
        counter::list2fullIDs(countersList, counters);
        return std::make_shared<job>(m_ioc, job_name, duration, counters);
    }

    errors manager::generate_jobs(const std::vector<std::string> &jobs_list)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        for (auto name : jobs_list)
        {
            std::shared_ptr<job> a_job = create_job(name);
            if (a_job)
            {
                a_job->interfaces::notifier::client::bind(this);
                m_jobs.emplace(name, a_job);
            }
            else
            {
                LOGERROR(getTypeInst(), " Cannot add \"", name, "\" to jobs list");
                return ::pm::errors::FAIL;
            }
        }
        return ::pm::errors::OK;
    }

    std::shared_ptr<job> manager::get_job(const std::string &job_name)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto jobIter = m_jobs.find(job_name);
        if (jobIter == m_jobs.end())
        {
            LOGERROR(getTypeInst(), " Cannot find PM job with name=\"", job_name, "\"");
            return nullptr;
        }
        return jobIter->second;
    }

    void manager::start_job(std::shared_ptr<job> a_job) const
    {
        a_job->start();
    }

    void manager::stop_job(std::shared_ptr<job> a_job) const
    {
        a_job->stop();
    }

    void manager::notify_timeout(const pm::job &job,
                                 const ::timers::time_point &startTime)
    {
        auto prim = std::make_shared<primitive::GetStatisticsReq>(true, job.counters());
        prim->m_header->to = TypeInst(workers::type::OAM, 0);
        ::transaction::errors err;
        transaction::data data(job.name(), startTime);
        prim->m_header->ulpKey = m_trans_mngr.openTransaction(std::move(data), err);
        if (err == ::transaction::errors::OK)
        {
            workers::interfaces::sender::client::send(prim);
        }
    }

    void manager::start_all_jobs()
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        for (auto const &iter : m_jobs)
        {
            start_job(iter.second);
        }
    }

    void manager::stop_and_del_all_jobs()
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        for (auto const &iter : m_jobs)
        {
            stop_job(iter.second);
        }
        m_jobs.clear();
    }

    bool manager::generate_report(std::shared_ptr<primitive::GetStatisticsRsp> rsp,
                                  std::string &report_name)
    {
        ::transaction::errors err;
        pm::transaction::data pm_job_data = m_trans_mngr.closeTransaction(rsp->m_header->ulpKey, err);
        if (::transaction::errors::OK != err)
        {
            LOGERROR(getTypeInst(),
                     " Received ", ::primitive::id2str(rsp->m_header->id), " for unknown transactionId=", rsp->m_header->ulpKey);
            return false;
        }

        std::multimap<::pm::counter::index, ::pm::counter::id_value> counter_data;
        for (auto counterIter = rsp->counters.begin(); counterIter != rsp->counters.end(); counterIter++)
        {
            if (counterIter->second.first != ANY_INDEX)
            {
                counter_data.emplace(counterIter->second.first,
                                     pm::counter::id_value(counterIter->first,
                                                           counterIter->second.second));
            }
        }

        std::shared_ptr<pm::job> job_p = get_job(pm_job_data.m_jobName);
        if (!job_p)
        {
            LOGERROR(getTypeInst(),
                     " Cannot generate report due to absence of pm job with name=\"", pm_job_data.m_jobName, "\"");
            return false;
        }

        std::string reportDirPath = m_config.getValue<std::string>(CCNGW_REPORT_PATH + STATS_REPORT,
                                                                   STATS_REPORT_DEFAULT_DIR)
                                    + "/";

        fs::local localFs(getTypeInst());
        localFs.createDir(reportDirPath);

        timers::time_point currTime(timers::now());
        report_name = reportDirPath + "STAT_" + std::to_string(m_uvrId) + "_" + currTime.utcFormat(Csv::FileTsFormat) + ".csv";
        std::ofstream out(report_name);
        if (!out.is_open())
        {
            LOGERROR(getTypeInst(),
                     " Cannot open file=\"", report_name, "\" to generate report");
            return false;
        }

        // Print header
        out << "ID_SRC" << Csv::delimiter << "START_DATE" << Csv::delimiter << "DUR";
        for (auto orderIter = job_p->counters().begin();
             orderIter != job_p->counters().end();
             orderIter++)
        {
            out << Csv::delimiter << pm::counter::all_ids.at(orderIter->first);
        }
        out << Csv::new_line;

        // Print rows
        auto indexIter = counter_data.begin();
        while (indexIter != counter_data.end())
        {
            std::string timestamp = pm_job_data.m_startTime.format();
            auto secs = std::chrono::duration_cast<std::chrono::seconds>(job_p->get_period());
            std::ostringstream os_record;
            os_record << indexIter->first << Csv::delimiter << timestamp << Csv::delimiter << secs.count();
            auto range = counter_data.equal_range(indexIter->first);
            // Write the counters in order which they were requested
            for (auto orderIter = job_p->counters().begin(); orderIter != job_p->counters().end(); orderIter++)
            {
                auto counterIter = range.first;
                for (; counterIter != range.second; counterIter++)
                {
                    if (orderIter->first == counterIter->second.first)
                    {
                        os_record << Csv::delimiter << counterIter->second.second;
                        break;
                    }
                }
                if (counterIter == range.second)
                {
                    LOGDEBUG(getTypeInst(),
                             " Counter \"", pm::counter::all_ids.at(orderIter->first),
                             "\" is not found for index=\"", indexIter->first, "\"");
                    os_record << Csv::delimiter << pm::counter::value();
                }
            } // for (; orderIter != job_p->counters().end(); orderIter++)
            // Write record to report
            out << os_record.str() << Csv::new_line;
            // Skip all elements with the same key, go to the next index
            indexIter = range.second;
        } // while (indexIter != counter_data.end())

        out.close();
        return true;
    }

} // namespace pm
