#include "pm/provider/src/pm_provider.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"
#include "pm/common/src/pm_primitive.hpp"

namespace pm
{
    provider::provider() = default;

    provider::~provider()
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        m_counters.clear();
    }

    void provider::registerCounters()
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        counter::vectorIDs counterIDs;
        counter::id prevCounterId = counter::last;
        for (auto counterIter = m_counters.begin(); counterIter != m_counters.end(); counterIter++)
        {
            if (counterIter->first != prevCounterId)
            {
                LOGDEBUG(getTypeInst(), " Registering counter with id=", counterIter->first);
                counterIDs.push_back(counterIter->first);
                prevCounterId = counterIter->first;
            }
        }
        auto request = std::make_shared<pm::primitive::ProviderRegisterReq>(counterIDs);
        request->m_header->to = TypeInst(workers::type::OAM, 0);
        workers::interfaces::sender::client::send(request);
    }

    void provider::addCounter(counter::id key, counter::indexing isIndexing)
    {
        if (isIndexing == counter::indexing::yes)
        {
            addCounter(key, ANY_INDEX);
        }
        else
        {
            addCounter(key, NO_INDEX);
        }
        return;
    }

    void provider::addCounter(counter::id key, counter::index index)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto range = m_counters.equal_range(key);
        if (range.first != m_counters.end())
        {
            for (auto counterIter = range.first; counterIter != range.second; counterIter++)
            {
                if (counterIter->second.first == index)
                {
                    LOGERROR(getTypeInst(),
                             " The counter [", key, ";", index, "] already exists");
                    return;
                }
            }
        }
        m_counters.emplace(key, counter::index_value(index, counter::value()));
        LOGDEBUG(getTypeInst(), " The counter[", key, ";", index, "] was added");
        return;
    }

    void provider::increaseCounter(counter::id key, counter::value delta)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto counterIter = m_counters.find(key);
        if (counterIter == m_counters.end())
        {
            LOGERROR(getTypeInst(),
                     " Cannot increase counter[", key, ";", NO_INDEX, "]: no counter with such ID");
            return;
        }

        if (counterIter->second.first != NO_INDEX)
        {
            LOGERROR(getTypeInst(),
                     " Cannot increase counter[", key, ";",
                     counterIter->second.first, "]: counter should be non-indexing");
            return;
        }
#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
        counter::value prev_value = counterIter->second.second;
#endif
        counterIter->second.second += delta;
        LOGDEBUG(getTypeInst(),
                 " The counter[", key, ";", NO_INDEX, "] was increased on delta=",
                 delta, " from previous value=", prev_value, " to value=", counterIter->second.second);
    }

    void provider::increaseCounter(counter::id key, counter::index index, counter::value delta)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto range = m_counters.equal_range(key);
        if (range.first == m_counters.end())
        {
            LOGERROR(getTypeInst(),
                     " Cannot increase counter[", key, ";", index, "]: no counter with such ID");
            return;
        }

        auto counterIter = range.first;
        for (; counterIter != range.second; counterIter++)
        {
            // Try to find already inserted counter-index in map
            if (counterIter->second.first != ANY_INDEX &&
                counterIter->second.first != NO_INDEX &&
                counterIter->second.first != index)
            {
                continue;
            }
            // If index is any, it is the first index for this counter
            // Assign the index to it
            if (counterIter->second.first == ANY_INDEX)
            {
                counterIter->second.first = index;
            }
#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
            counter::value prev_value = counterIter->second.second;
#endif
            counterIter->second.second += delta;
            LOGDEBUG(getTypeInst(),
                     " The counter[", key, ";", index, "] was increased on delta=", delta,
                     " from previous value=", prev_value, " to value=", counterIter->second.second);
            break;
        }
        // New counter-index appears, add it to map
        if (counterIter == range.second)
        {
            m_counters.emplace(key, counter::index_value(index, delta));
            LOGDEBUG(getTypeInst(),
                     " New index=\"", index, "\" is added for counter id=", key, " with value=", delta);
        }
        return;
    }

    void provider::resetCounterNoIndex(counter::id key)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto counterIter = m_counters.find(key);
        if (counterIter == m_counters.end())
        {
            LOGERROR(getTypeInst(),
                     " Cannot reset counter[", key, ";", NO_INDEX, "]: no counter with such ID");
            return;
        }

        if (counterIter->second.first != NO_INDEX)
        {
            LOGERROR(getTypeInst(),
                     " Cannot reset counter[", key, ";", counterIter->second.first,
                     "]: counter should be non-indexing");
            return;
        }

        counterIter->second.second = counter::value();
        LOGDEBUG(getTypeInst(), " The counter[", key, ";", NO_INDEX, "] was reset");
    }

    void provider::resetCounter(counter::id key, const counter::index &index)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto range = m_counters.equal_range(key);
        if (range.first == m_counters.end())
        {
            LOGERROR(getTypeInst(),
                     " Cannot reset counter[", key, ";", index, "]: no counter with such ID");
        }

        for (auto counterIter = range.first; counterIter != m_counters.end(); counterIter++)
        {
            if (counterIter->second.first == index ||
                index == ALL_INDEX)
            {
                counterIter->second.second = counter::value();
                LOGDEBUG(getTypeInst(), " Counter[", key, ";", index, "] was reset");
            }
        }
        return;
    }

    void provider::setCounter(counter::id counter, counter::value value)
    {
        setCounter(counter, NO_INDEX, value);
    }
    void provider::setCounter(counter::id counter, counter::index index, counter::value value)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);
        auto range = m_counters.equal_range(counter);
        if (range.first == m_counters.end())
        {
            LOGERROR(getTypeInst(),
                     " Cannot set counter[", counter, ";", index, "]: no counter with such ID");
            return;
        }

        auto counterIter = range.first;
        for (; counterIter != range.second; counterIter++)
        {
            // Try to find already inserted counter-index in map
            if (counterIter->second.first != ANY_INDEX &&
                counterIter->second.first != NO_INDEX &&
                counterIter->second.first != index)
            {
                continue;
            }
            // If index is any, it is the first index for this counter
            // Assign the index to it
            if (counterIter->second.first == ANY_INDEX)
            {
                counterIter->second.first = index;
            }
#if LOG_MACRO_LEVEL >= LOG_LEVEL_DEBUG
            counter::value prev_value = counterIter->second.second;
#endif
            counterIter->second.second = value;
            LOGDEBUG(getTypeInst(),
                     " The counter[", counter, ";", index, "] was set to=", value,
                     " from previous value=", prev_value, " to value=", counterIter->second.second);
            break;
        }
        // New counter-index appears, add it to map
        if (counterIter == range.second)
        {
            m_counters.emplace(counter, counter::index_value(index, value));
            LOGDEBUG(getTypeInst(),
                     " New index=\"", index, "\" is added for counter id=", counter, " with value=", value);
        }
        return;
    }

    void provider::handleGetCountersReq(const ::primitive::GenericPtr &prim)
    {
        std::lock_guard<std::recursive_mutex> guard(m_guard);

        auto req = std::dynamic_pointer_cast<pm::primitive::GetCountersReq>(prim);
        auto rsp = std::make_shared<pm::primitive::GetCountersRsp>();

        for (auto idIter = req->counters.cbegin(); idIter != req->counters.cend(); ++idIter)
        {
            auto range = m_counters.equal_range(idIter->first);
            if (range.first == m_counters.end())
            {
                LOGDEBUG(getTypeInst(),
                         " Cannot find counter[", idIter->first, ";", idIter->second, "]: no counter with such ID");
                rsp->counters.emplace(idIter->first, counter::index_value(idIter->second, counter::value::no_value()));
                continue;
            }

            bool isIndexNotFound = true;
            for (auto counterIter = range.first; counterIter != range.second; counterIter++)
            {
                if (idIter->second == ALL_INDEX ||
                    counterIter->second.first == idIter->second)
                {
                    rsp->counters.emplace(idIter->first, counterIter->second);
                    if (req->reset)
                    {
                        counterIter->second.second = counter::value();
                    }
                    isIndexNotFound = false;

                    if (counterIter->second.first == idIter->second)
                        break;
                }
            }
            if (isIndexNotFound)
            {
                LOGDEBUG(getTypeInst(),
                         " Cannot find counter[", idIter->first, ";", idIter->second, "]: no counter with such index");
                rsp->counters.emplace(idIter->first, counter::index_value(idIter->second, counter::value::no_value()));
            }
        }

        rsp->m_header->to = prim->m_header->from;
        rsp->m_header->ulpKey = prim->m_header->ulpKey;
        workers::interfaces::sender::client::send(rsp);
    }

} // namespace pm
