#pragma once

// UVR includes.
#include "UvrTypes.hpp"

namespace Uvr
{
    namespace API
    {
        /// @brief Initializes UVR with values from the configuration file.
        ///        Must be called before the start() method.
        /// @param[in] configPath Path to the configuration file.
        /// @param[in] asyncMode Indicates in which mode the calls will be
        ///                      verified. If it is set to @c true then @ref
        ///                      verifyCall() function should be used to
        ///                      verify call. If it is set to @c false then @ref
        ///                      verifyCallSynchronously() function should
        ///                      be used to verify call. @c true by default.
        /// @returns @c true if the initialization was successful,
        ///          @c false otherwise.
        /// @see start()
        /// @see verifyCall()
        /// @see verifyCallSynchronously()
        bool init(const std::string &configPath,
                  const bool asyncMode = true);

        /// @brief Starts UVR's io_service loop.
        /// @param[in] threadPoolSize
        /// @see init()
        /// @see getNodeStatus()
        /// @see stop()
        void start(int threadPoolSize = 20);

        /// @brief Returns the status of UVR. Can be called at any time.
        /// @returns The status of UVR:
        ///          - @c Uvr::NodeStatus::NotInitialized initial state, UVR isn't initialized
        ///          - @c Uvr::NodeStatus::Initalized initialization is successfully finished,
        ///                                           @ref init() returned @c true
        ///          - @c Uvr::NodeStatus::Starting activation is in process, @ref start() is working,
        ///                                         numeration dictionaries are being loaded
        ///          - @c Uvr::NodeStatus::Started UVR is ready to work, numeration dictionaries are loaded
        ///          - @c Uvr::NodeStatus::Stopping deactivation is started, @ref stop() was called
        ///          - @c Uvr::NodeStatus::Stopped deactivation is finished.
        Uvr::NodeStatus getNodeStatus();

        void stop();
        void waitStop();

        /// @brief Verifies the call asynchronously. The verification verdict will
        ///        be passed to the registered callback. Calls @ref StoreRecordCallback
        ///        if it was registered for a User with passed @a userId.
        /// @returns The session ID that will be passed to the registered callback.
        ///          It can be used to understand for which call the response was
        ///          passed to the callback.
        /// @see registerVerifyCallVerdictCallback()
        /// @see registerStoreRecordCallback()
        SessionId_t
        verifyCall(UserId_t userId,
                   OperatorId_t idSrc,
                   const OctetStringPtr &callingNumber,
                   const OctetStringPtr &calledNumber,
                   const OctetStringPtr &originalCalledNumber = OctetStringPtr(),
                   const OctetStringPtr &genericNumber = OctetStringPtr(),
                   const CallId_t &callId = CallId_t(),
                   bool isTrusted = false);

        /// @brief Verifies the call synchronously. Calls all @ref StoreRecordCallback
        ///        that have been registered.
        /// @returns Verification result, see @ref VerificationResult for more details.
        ///          If @a isTrusted was set to @c true then return @ref VerificationResult value
        ///          with verdict equal to @ref Uvr::VerificationVerdict::success.
        /// @see registerStoreRecordCallback()
        VerificationResult
        verifyCallSynchronously(const OperatorId_t idSrc,
                                const OctetStringPtr &callingNumber,
                                const OctetStringPtr &calledNumber,
                                const OctetStringPtr &originalCalledNumber = OctetStringPtr(),
                                const OctetStringPtr &genericNumber = OctetStringPtr(),
                                const CallId_t &callId = CallId_t(),
                                const bool isTrusted = false);

        /// @brief Creates a temporary entry in the calls registry for an outgoing call.
        ///        Calls all @ref StoreRecordCallback that have been registered.
        /// @param[in] callingNumber Calling number in ITU-T Recommendation Q.763 format.
        /// @param[in] calledNumber Called number in ITU-T Recommendation Q.763 format.
        /// @param[in] callId Call ID. Empty by default.
        /// @param[in] idDst Destination Operator ID. std::nullopt by default.
        /// @see TRecQ763Number
        /// @see registerStoreRecordCallback()
        void
        notifyAboutOutgoingCall(const OctetStringPtr &callingNumber,
                                const OctetStringPtr &calledNumber,
                                const Uvr::CallId_t &callId = "",
                                const Uvr::OptionalOperatorId &idDst = std::nullopt);

        UserId_t
        registerVerifyCallVerdictCallback(VerifyCallVerdictCallback callback,
                                          UserId_t userId = undefinedUserId);

        void
        reportCallInformation(const CallInformationResponses &responses);

        UserId_t
        registerCallInformationRequestCallback(CallInformationRequestCallback callback,
                                               UserId_t userId = undefinedUserId);

        /// @brief Registers a callback for storing records. If the callback was
        ///        registered, then it will be called at the end of call verification
        ///        and at the end of outgoing call notification.
        /// @param[in] callback A function that implements saving the record in the storage.
        /// @param[in] userId The User ID for which the callback is being registered.
        ///                   @ref undefinedUserId by default.
        /// @returns User ID passed to @a userId, or new User ID if @a userId = @ref undefinedUserId.
        /// @see verifyCall()
        /// @see verifyCallSynchronously()
        /// @see notifyAboutOutgoingCall()
        UserId_t
        registerStoreRecordCallback(const StoreRecordCallback callback,
                                    const UserId_t userId = undefinedUserId);

        /// @brief Registers a callback for updating records. If the callback was
        ///        registered, then it will be called at "update-records" command.
        /// @param[in] callback A function that implements updating records in the storage.
        /// @param[in] userId The User ID for which the callback is being registered.
        ///                   @ref undefinedUserId by default.
        /// @returns User ID passed to @a userId, or new User ID if @a userId = @ref undefinedUserId.
        UserId_t
        registerUpdateRecordsCallback(const UpdateRecordsCallback callback,
                                      const UserId_t userId = undefinedUserId);

        namespace Digest
        {
            namespace Numbers
            {
                /// @brief The function returns IDs of primary and secondary UVRs to which the given @a number belongs according to the NUMBERS digest.
                /// @param[in] number The number for which primary and secondary UVR IDs are requested.
                /// @returns UVR IDs for primary and secondary UVRs.
                UvrIds
                getUvrIds(const OctetStringPtr &number);

            } // namespace Numbers

            namespace Uvrs
            {
                /// @brief Gets information about UVR for passed UVR ID.
                /// @param[in] uvrId UVR ID.
                /// @returns Information about UVR.
                UvrInfo
                getUvrInfo(const UvrId_t &uvrId);

                /// @brief The function returns own UVR ID.
                /// @returns Own VFN ID.
                UvrId_t
                getOwnUvrId();
            } // namespace Uvrs

            namespace Hubs
            {
                /// @brief Gets information about HUB for passed HUB ID.
                /// @param[in] hubId HUB ID.
                /// @returns Information about HUB.
                HubInfo
                getHubInfo(const Gt_t &hubId);
            } // namespace Hubs

            namespace Operators
            {
                /// @brief Gets the information about Operator for the passed Operator ID.
                /// @param[in] operatorId The Operator ID for which information is being requested.
                /// @returns Information about Operator.
                OperatorInfo
                getOperatorInfo(const OperatorId_t &operatorId);
            } // namespace Operators

        } // namespace Digest

    } // namespace API

    namespace Troubleshoot
    {
        std::string doVersions();
    } // namespace Troubleshoot

} // namespace Uvr
