#pragma once

// AFAP includes.
#include "components/db/src/types.hpp"
#include "components/filesystem/src/filesystem.hpp"
#include "components/support/variadic.hpp"
#include "components/worker_node/src/typeInst.hpp"

// STL includes.
#include <shared_mutex>

namespace vfc
{

    namespace subscriber
    {
        struct info_t
        {
            uint16_t m_idUvrP = 0;
            uint16_t m_idUvrS = 0;
            Uvr::MetaInfo::BitFlags_t m_metaInfoFlags = Uvr::MetaInfo::emptyBitFlags;
        };
    } // namespace subscriber

#define LEVEL_BITS 12
#define LEVEL_MASK MASK(LEVEL_BITS)
    struct base_level
    {
        const static int level_size{1 << LEVEL_BITS};

        int m_depth;
        explicit base_level(int depth) : m_depth(depth) {}
        virtual ~base_level() = default;
        virtual void match(uint64_t number,
                           subscriber::info_t **ret_pp) = 0;
        virtual uint16_t push(uint64_t number,
                              subscriber::info_t &info) = 0;
        virtual void pop(uint64_t number,
                         subscriber::info_t &info) = 0;
    };
    template <class sub_level, int depth>
    struct level : public base_level
    {
        std::array<std::shared_ptr<sub_level>, level_size> nodes = {0};
        level() : base_level(depth) {}
        level(level &str) = delete;

        level &operator=(level &str) = delete;

        uint16_t push(uint64_t number,
                      subscriber::info_t &info) override
        {
            int indx = (number & LEVEL_MASK);
            if (nodes[indx] == nullptr)
            {
                nodes[indx] = std::make_shared<sub_level>();
            }
            return nodes[indx]->push(number >> LEVEL_BITS, info);
        }
        void pop(uint64_t number, subscriber::info_t &info) override
        {
            int indx = (number & LEVEL_MASK);
            if (nodes[indx] == nullptr)
            {
                info = {};
            }
            else
            {
                nodes[indx]->pop(number >> LEVEL_BITS, info);
            }
        }
        void match(uint64_t number,
                   subscriber::info_t **ret_pp) override
        {
            int indx = (number & LEVEL_MASK);
            if (nodes[indx] == nullptr)
            {
                *ret_pp = nullptr;
            }
            else
            {
                nodes[indx]->match(number >> LEVEL_BITS, ret_pp);
            }
        }
    };
    template <>
    struct level<subscriber::info_t, 1> : public base_level
    {
        std::array<subscriber::info_t, level_size> nodes = {};
        level() : base_level(1) {}
        uint16_t push(uint64_t number,
                      subscriber::info_t &info) override
        {
            int indx = (number & LEVEL_MASK);
            uint16_t ret = nodes[indx].m_idUvrP;
            nodes[indx] = info;
            return ret;
        }
        void pop(uint64_t number, subscriber::info_t &info) override
        {
            int indx = (number & LEVEL_MASK);
            info = nodes[indx];
            nodes[indx] = {};
        }
        void match(uint64_t number,
                   subscriber::info_t **ret_pp) override
        {
            int indx = (number & LEVEL_MASK);
            if (nodes[indx].m_idUvrP == 0) // Primary UVR couldn't be zero
            {
                *ret_pp = nullptr;
            }
            else
            {
                *ret_pp = &nodes[indx];
            }
        }
    };

#define MAX_LEVELS 3

#define _MAX_LEVEL(N) CAT(level, N)
#define MAX_LEVEL_TYPE EVAL(DEFER(_MAX_LEVEL)(MAX_LEVELS))

#define _LEVEL(N, N0) using CAT(level, N) = level<CAT(level, N0), N>
#define DECLARE_LEVEL(N) EVAL(DEFER(_LEVEL)(N, DEC(N)))

    using level1 = level<subscriber::info_t, 1>;
    DECLARE_LEVEL(2);
    DECLARE_LEVEL(3);

    class matcher
    {
        mutable std::shared_mutex m_guard; ///< Multiple-reader/single-writer mutex
        const TypeInst m_owner;
        std::shared_ptr<MAX_LEVEL_TYPE> lvls_p;
        db::limit_t m_number_of_subs;
        fs::local m_fs_local;

        boost::asio::io_context &m_ioc;

        void reset();

        using StringsBuff = std::vector<std::string>;

        struct NumberInfo
        {
            subscriber::info_t subInfo;
            uint64_t number;
        };
        using NumberInfoBuff = std::vector<NumberInfo>;

        using StrandType = boost::asio::strand<boost::asio::io_context::executor_type>;

        class PopulateGuard
        {
            std::promise<bool> &m_promise;

        public:
            PopulateGuard(std::promise<bool> &promiseObj);

            ~PopulateGuard();
        };

    public:
        enum class result_t
        {
            do_match = 1,
            no_match = 0
        };

        explicit matcher(const TypeInst &owner,
                         boost::asio::io_context &ioc);

        result_t match(const Uvr::OctetStringPtr &CallingPN,
                       subscriber::info_t &);

        void sub_insert(const db::subscriber::record &);
        void subscribersParse(const StringsBuff &buff,
                              std::shared_ptr<PopulateGuard> guard,
                              StrandType strand);
        void subscribersInsert(const std::shared_ptr<NumberInfoBuff> buff,
                               std::shared_ptr<PopulateGuard> guard);
        void sub_update(const db::subscriber::record &);
        void sub_remove(const db::subscriber::record &);

        void populate(const std::string &zipFilename);
        void update(const std::string &filename);

        db::limit_t number_of_subs() const;

        static uint64_t reverseNumberBytes(uint64_t &number);
        static uint64_t convertToNumber(const Uvr::Number &callingPartyNumber);
        static uint64_t convertToNumber(const Uvr::OctetStringPtr &callingPartyNumber);
    };

    std::ostream &operator<<(std::ostream &os, const matcher::result_t &obj);

} // namespace vfc
