#pragma once

// AFAP includes.
#include "components/config/src/config_manager.hpp"
#include "components/db/src/interfaces.hpp"
#include "components/db/src/types.hpp"
#include "components/http/src/connection_manager.hpp"
#include "components/http/src/listener.hpp"
#include "pm/provider/src/interfaces.hpp"

// STL includes.
#include <shared_mutex>

namespace hubs
{
    struct host
    {
        Uvr::Ip_t m_ipAddress; ///< Host IP address
        Uvr::Dns_t m_dnsName;  ///< Host domain name

        host(const Uvr::Ip_t &ipAddress = "",
             const Uvr::Dns_t &dnsName = "")
            : m_ipAddress(ipAddress),
              m_dnsName(dnsName)
        {
        }
    };

    struct node
    {
        Uvr::Gt_t m_hubId;            ///< Hub node identifier
        std::vector<host> m_hosts;    ///< List of hub hosts stored in priority order (first element has highest priority)
        unsigned int m_nextHostIx{0}; ///< Index of the next host to send request

        node(const Uvr::Gt_t &hubId = "",
             const std::vector<host> &hosts = {})
            : m_hubId(hubId),
              m_hosts(hosts)
        {
        }

        /// @brief Checks that the IP address belongs to hub.
        /// @param[in] address IP address that should be cheched.
        /// @returns @c true if the address belongs to one of the hub hosts,
        ///          @c false otherwise.
        bool hasIpAddress(const std::string &address) const
        {
            for (const auto &host : m_hosts)
            {
                if (address == host.m_ipAddress)
                {
                    return true;
                }
            }

            return false;
        }

        /// @brief Returns IP address of next host to send http request.
        ///        Host is choosen by round robin algorith among
        ///        configured hosts for this HUB node.
        /// @param
        /// @return Pointer to host IP address,
        ///         @c nullptr if no hosts are configured
        const std::string *getIpOfNextHost(void)
        {
            if (m_hosts.empty())
            {
                return nullptr;
            }
            unsigned int ix = m_nextHostIx;
            m_nextHostIx++;
            if (m_nextHostIx == m_hosts.size())
                m_nextHostIx = 0;
            return &m_hosts[ix].m_ipAddress;
        }

        /// @brief Returns IP address of the host after the specified one.
        /// @param ipAddress IP address of the specified host
        /// @return Pointer to host IP address after the specified one,
        ///         @c nullptr if the specified host is not found
        const std::string *getIpOfAfterHost(const std::string &ipAddress)
        {
            std::size_t hostIx = 0;
            for (; hostIx < m_hosts.size(); hostIx++)
            {
                if (ipAddress == m_hosts[hostIx].m_ipAddress)
                    break;
            }
            // The current host is not found
            if (hostIx == m_hosts.size())
                return nullptr;
            // Only one host is configured, no next host
            if (m_hosts.size() == 1)
                return nullptr;
            // Go to the nest host
            hostIx = (hostIx == m_hosts.size() - 1) ? 0 : hostIx + 1;
            return &m_hosts[hostIx].m_ipAddress;
        }

        /// @brief Returns domain name of next host to send http request.
        ///        Host is choosen by round robin algorith among
        ///        configured hosts for this HUB node.
        /// @param
        /// @return Pointer to host domain name,
        ///         @c nullptr if no hosts are configured
        const std::string *getDnsOfNextHost(void)
        {
            if (m_hosts.empty())
            {
                return nullptr;
            }
            unsigned int ix = m_nextHostIx;
            m_nextHostIx++;
            if (m_nextHostIx == m_hosts.size())
                m_nextHostIx = 0;
            return &m_hosts[ix].m_dnsName;
        }

        /// @brief Returns domain name of the host after the specified one.
        /// @param domain Domain name of the specified host
        /// @return Pointer to host domain name after the specified one,
        ///         @c nullptr if the specified host is not found
        const std::string *getDnsOfAfterHost(const std::string &domain)
        {
            std::size_t hostIx = 0;
            for (; hostIx < m_hosts.size(); hostIx++)
            {
                if (domain == m_hosts[hostIx].m_dnsName)
                    break;
            }
            // The current host is not found
            if (hostIx == m_hosts.size())
                return nullptr;
            // Only one host is configured, no next host
            if (m_hosts.size() == 1)
                return nullptr;
            // Go to the nest host
            hostIx = (hostIx == m_hosts.size() - 1) ? 0 : hostIx + 1;
            return &m_hosts[hostIx].m_dnsName;
        }
    };

    class manager : public ::http::connections::manager,
                    public http::interfaces::messages::handler::client,
                    public pm::interfaces::provider::client,
                    public db::interfaces::tables::uvrs::client
    {
    public:
        manager(const TypeInst &owner,
                boost::asio::io_context &ioc,
                const configuration::manager &config);
        ~manager() override;

        /// @brief Starts HUB manager:
        ///        - reads configuration parameters
        ///        - becomes listen to incoming http connections
        /// @param
        void start(void);

        /// @brief Stops HUB manager:
        ///        - destroyes existing http connections
        ///        - clear HUB information
        /// @param
        void stop(void);

        /// @brief Updates received HUB configuration and
        ///        recreate HUB node entities with changed data if needed
        /// @param hubs[in] Records with HUB data from DB
        void updateHubs(::db::hub::rows &hubs);

        /// @brief Gets HUB ID by specified IP address
        /// @param[in] ipAddress Specified IP address
        /// @param[out] hubId Returns HUB ID value
        /// @return @c true if hub is found, @c false otherwise
        bool getHubIdByIpAddress(const Uvr::Ip_t &ipAddress, Uvr::Gt_t &hubId);

        /// @brief Starts session and sends http request towards primary HUB
        /// @param[in] reqId Http session ID
        /// @param[in] target Http request target
        /// @param[in] payloadData Http request payload data
        /// @param[in] user_data Specific http user data for this session
        /// @param[in] keepAlive The flag indicated whether connection should be kept
        ///                      after reply will be received
        /// @return @c true if new session is initiated
        ///         @c false otherwise
        bool startAndSendRequest(Uvr::SessionId_t reqId,
                                 const std::string &target,
                                 const std::string &payloadData,
                                 const std::shared_ptr<void> &user_data,
                                 bool keepAlive = false);

        /// @brief Sends http response over connection, where request was received
        /// @param[in] result Http response result code
        /// @param[in] reqId Http session ID
        /// @param[in] payloadData Http response payload data
        /// @param[in] keepAlive The flag indicated whether connection should be kept
        /// @return @c true if replay is sent
        ///         @c false otherwise
        bool sendReply(http::messages::result result,
                       const http::connections::id &connId,
                       const std::string &payloadData,
                       bool keepAlive = false);

        /// @brief Resends http request on next host
        /// @param[in] connPtr Pointer to http connection entity
        /// @return @c true if new session to next host is initiated
        ///         @c false if no available hosts
        bool resendRequestOnNextHost(const http::connection<http::role::CLIENT> *connPtr) const;

    protected:
        // http::connections::manager overrides

        /// @brief Checks that the connection IP address belongs to one of the hubs
        ///        and calls the parent method if so.
        /// @returns @c true in case of success, @c false otherwise.
        /// @see http::connections::manager::connectionAccept
        bool connectionAccept(http::connection_ptr<http::role::SERVER> connPtr) override;
        void connectFailure(http::connection_ptr<http::role::CLIENT> connPtr) override;
        void connectionUp(http::connection_ptr<http::role::CLIENT> connPtr) override;
        void connectionUp(http::connection_ptr<http::role::SERVER> connPtr) override;
        void connectionDown(http::connections::id connId, http::role role) override;

        void getAddress(uint attemptNumber,
                        const std::string &prevAddress,
                        std::string &nextAddress) override;

        // Implement virtual functions of http::connections::manager class
        void requestSpecificHandling(http::connections::id connId,
                                     const std::string_view &target,
                                     const std::string &payloadData) override;
        void responseSpecificHandling(const http::connection<http::role::CLIENT> *connPtr,
                                      http::messages::result result,
                                      const std::string &payloadData,
                                      bool keepAlive) override;

    private:
        void addHub(const db::hub::record &hub);
        void insertHub(size_t pos, const db::hub::record &hub);
        void replaceHub(size_t pos, const db::hub::record &hub);
        void swapHubs(size_t first, size_t second);
        void removeHub(size_t pos);

        const configuration::manager &m_config; ///< Configuration
        std::string m_remPort;                  ///< Port of the remote HTTP server
        http::listener m_listener;              ///< Listener of incomming connections
        mutable std::shared_mutex m_hubGuard;   ///< Protector of shared HUB data in multiple threads (multiple-reader/single-writer mutex)
        db::hub::rows m_hubConfig;              ///< Hubs information from DB
        std::vector<hubs::node> m_hubNodes;     ///< List of hub nodes stored in priority order (first element has highest priority)
        bool m_isRunning{false};                ///< Indicator whether hub manager in running now
        timers::duration m_connectTimeout;      ///< TCP connect timeout
    };

} // namespace hubs
