#include "vfgw/src/node.hpp"

// AFAP includes.
#include "components/asn1/src/asn1_tools.hpp"
#include "components/db/src/mimic.hpp"
#include "components/worker_node/src/config_parameters.hpp"
#include "vfgw/src/types.hpp"

// OpenSSL includes.
#include <openssl/rand.h>

namespace vfgw
{

    node::node(const TypeInst &ownInst,
               boost::asio::io_context &ioc,
               const configuration::manager &config) : pm::node(ownInst, ioc, config),
                                                       m_hubMngr(m_ownTypeInst, ioc, config),
                                                       m_fsLocal(ownInst),
                                                       m_rsa(config)
    {
        m_dbClient = std::make_shared<db::stub::mimic>(m_uvrId);
        db::interfaces::tables::client::bind(m_dbClient.get());
        db::interfaces::tables::uvrs::client::bind(m_dbClient.get());
        db::interfaces::tables::hubs::client::bind(m_dbClient.get());
        db::interfaces::tables::Operators::client::bind(m_dbClient.get());
        m_hubMngr.http::interfaces::messages::handler::client::bind(this);
        m_hubMngr.pm::interfaces::provider::client::bind(this);
        m_hubMngr.db::interfaces::tables::uvrs::client::bind(m_dbClient.get());
        m_httpDataEncrypt = m_config.getValue<bool>(HTTP_DATA_ENCRYPT, false);
    }

    node::~node()
    {
        clearPubKeys();
    }

    void node::stop(void)
    {
        clearPubKeys();
        m_hubMngr.stop();
        pm::node::stop();
    }

    void node::prepareCounters()
    {
        addCounter(pm::counter::calls::ERR1, pm::counter::indexing::yes);
        addCounter(pm::counter::calls::ERR2, pm::counter::indexing::yes);
        addCounter(pm::counter::http::http_up, pm::counter::indexing::no);
        addCounter(pm::counter::http::http_down, pm::counter::indexing::no);
        addCounter(pm::counter::http::http_auth_reject, pm::counter::indexing::no);
        addCounter(pm::counter::http::bad_request, pm::counter::indexing::no);
        addCounter(pm::counter::http::not_found, pm::counter::indexing::no);
        addCounter(pm::counter::http::bad_response, pm::counter::indexing::no);
        addCounter(pm::counter::http::request_failed, pm::counter::indexing::no);
        addCounter(pm::counter::http::request_routing_to_hub_failed, pm::counter::indexing::no);
        addCounter(pm::counter::http::request_routing_to_uvr_failed, pm::counter::indexing::no);
        addCounter(pm::counter::http::gen_bad_request_reply, pm::counter::indexing::no);
        addCounter(pm::counter::http::gen_error_response, pm::counter::indexing::no);
        addCounter(pm::counter::vfgw::hubs, pm::counter::indexing::no);
        addCounter(pm::counter::vfgw::uvrs, pm::counter::indexing::no);
        addCounter(pm::counter::vfgw::operators, pm::counter::indexing::no);
        addCounter(pm::counter::http::response_error, pm::counter::indexing::no);
        addCounter(pm::counter::http::response_continue, pm::counter::indexing::no);
        addCounter(pm::counter::http::response_release, pm::counter::indexing::no);
    }

    void node::handleSpecificPrimitive(const primitive::GenericPtr &prim)
    {
        switch (prim->m_header->id)
        {
            case ::primitive::UPLOAD_NODES_IND:
            {
                auto ind = std::dynamic_pointer_cast<::primitive::UploadNodesInd>(prim);
                if (ind)
                {
                    for (auto const &filename : ind->files)
                    {
                        std::string unzipFilename;
                        m_fsLocal.unzipFile(filename, unzipFilename);
                        if (filename.find("UVR_") != std::string::npos)
                        {
                            // re-create table
                            db::interfaces::tables::client::recreate(m_dbClient->db::interfaces::tables::uvrs::server::tableName,
                                                                     m_dbClient->db::interfaces::tables::uvrs::server::tableFields,
                                                                     m_dbClient->db::interfaces::tables::uvrs::server::tableSettings);
                            // read file and fill-in DB
                            LOGINFO(m_ownTypeInst, " START populate from ", unzipFilename);
                            db::interfaces::tables::client::fill_in(m_dbClient->db::interfaces::tables::uvrs::server::tableName,
                                                                    unzipFilename);
                        }
                        else if (filename.find("HUB_") != std::string::npos)
                        {
                            // re-create table
                            db::interfaces::tables::client::recreate(m_dbClient->db::interfaces::tables::hubs::server::tableName,
                                                                     m_dbClient->db::interfaces::tables::hubs::server::tableFields,
                                                                     m_dbClient->db::interfaces::tables::hubs::server::tableSettings);
                            // read file and fill-in DB
                            LOGINFO(m_ownTypeInst, " START populate from ", unzipFilename);
                            db::interfaces::tables::client::fill_in(m_dbClient->db::interfaces::tables::hubs::server::tableName,
                                                                    unzipFilename);
                        }
                        else
                        {
                            LOGERROR(m_ownTypeInst,
                                     " Unexpected file for parsing of \"", DIGESTS_NODES,
                                     "\" files updates: filename=\"", filename, "\"");
                        }
                        // cleanup
                        m_fsLocal.deleteFile(unzipFilename);
                    }

                    // Update counters
                    setCounter(pm::counter::vfgw::hubs,
                               ::pm::counter::value(static_cast<::pm::counter::value_t>(
                                   db::interfaces::tables::client::size(m_dbClient->db::interfaces::tables::hubs::server::tableName))));
                    setCounter(pm::counter::vfgw::uvrs,
                               ::pm::counter::value(static_cast<::pm::counter::value_t>(
                                   db::interfaces::tables::client::size(m_dbClient->db::interfaces::tables::uvrs::server::tableName))));

                    // Get primary and secondary hubs IDs
                    db::uvr::record uvr_rec;
                    db::interfaces::tables::uvrs::client::get(uvr_rec, m_uvrId);
                    if (uvr_rec.m_idUvr == Uvr::UvrId::undefined)
                    {
                        LOGWARN(m_ownTypeInst,
                                " Cannot get record from UVRs table about own VFN by uvrId=",
                                m_uvrId);
                        return;
                    }
                    db::gts_t hubIds{};
                    db::hub::rows hubs{};
                    hubIds.push_back(uvr_rec.m_idHubP);
                    if (!uvr_rec.m_idHubS.empty())
                    {
                        hubIds.push_back(uvr_rec.m_idHubS);
                    }
                    // Get primary and secondary hubs rows
                    db::interfaces::tables::hubs::client::get(hubs, hubIds);
                    if (hubs.empty())
                    {
                        LOGWARN(m_ownTypeInst,
                                " Cannot get records from HUBs table about hubs for uvrId=", m_uvrId,
                                " by primary hub gt=", hubIds[0],
                                " and secondary hub gt=", (hubIds.size() > 1 ? hubIds[1] : ""));
                        return;
                    }
                    m_hubMngr.updateHubs(hubs);
                    m_hubMngr.start();
                }
                break;
            }
            case ::primitive::UPLOAD_OPERATORS_IND:
            {
                auto ind = std::dynamic_pointer_cast<::primitive::UploadOperatorsInd>(prim);

                for (auto const &filename : ind->files)
                {
                    std::string unzipFilename;
                    m_fsLocal.unzipFile(filename, unzipFilename);
                    if (filename.find("OPR_") == std::string::npos)
                    {
                        LOGERROR(m_ownTypeInst,
                                 " Unexpected file for parsing of \"", DIGESTS_OPERATORS,
                                 "\" files updates: filename=\"", filename, "\"");
                        break;
                    }

                    // Re-create table.
                    db::interfaces::tables::client::recreate(m_dbClient->db::interfaces::tables::Operators::server::tableName,
                                                             m_dbClient->db::interfaces::tables::Operators::server::tableFields,
                                                             m_dbClient->db::interfaces::tables::Operators::server::tableSettings);

                    // Read file and fill-in DB.
                    LOGINFO(m_ownTypeInst, " START populate from ", unzipFilename);
                    db::interfaces::tables::client::fill_in(m_dbClient->db::interfaces::tables::Operators::server::tableName,
                                                            unzipFilename);

                    // Cleanup.
                    m_fsLocal.deleteFile(unzipFilename);
                }

                // Update counters
                setCounter(pm::counter::vfgw::operators,
                           pm::counter::value(
                               static_cast<::pm::counter::value_t>(
                                   db::interfaces::tables::client::size(m_dbClient->db::interfaces::tables::Operators::server::tableName))));

                break;
            }
            case ::primitive::UPLOAD_PUBLIC_KEYS_IND:
            {
                auto ind = std::dynamic_pointer_cast<::primitive::UploadPublicKeysInd>(prim);
                if (ind)
                {
                    std::unique_lock<std::shared_mutex> hubKeysGuard(m_hubKeysGuard);
                    for (const auto &filename : ind->hubFilesToCopy)
                    {
                        Uvr::Gt_t hubId = getHubIdFromFilename(filename);
                        if (hubId.empty())
                            continue; // go to read next file
                        RSA *pubKey;
                        pubKey = m_rsa.readPublicKey(filename);
                        if (!pubKey)
                            continue; // go to read next file
                        auto result = m_hubPubKeys.emplace(hubId, pubKey);
                        // If key for HUB already exists, update it
                        if (!result.second)
                        {
                            result.first->second = pubKey;
                            LOGDEBUG(m_ownTypeInst,
                                     " Public Key for HUB with id=", hubId, " was updated");
                        }
                        else
                        {
                            LOGDEBUG(m_ownTypeInst,
                                     " Public Key for HUB with id=", hubId, " was inserted");
                        }
                    }
                    for (const auto &filename : ind->hubFilesToRemove)
                    {
                        Uvr::Gt_t hubId = getHubIdFromFilename(filename);
                        if (hubId.empty())
                            continue; // go to read next file
                        auto hubIter = m_hubPubKeys.find(hubId);
                        if (hubIter == m_hubPubKeys.end())
                        {
                            LOGDEBUG(m_ownTypeInst,
                                     " Public Key for HUB with id=", hubId,
                                     " was not found to be deleted");
                            continue;
                        }
                        RSA_free(hubIter->second);
                        m_hubPubKeys.erase(hubIter);
                    }
                    hubKeysGuard.unlock();
                    std::unique_lock<std::shared_mutex> nodeKeysGuard(m_nodeKeysMutex);
                    for (const auto &filename : ind->nodeFilesToCopy)
                    {
                        Uvr::UvrId_t nodeId = getNodeIdFromFilename(filename);
                        if (nodeId == Uvr::UvrId::undefined)
                            continue; // go to next file
                        RSA *pubKey;
                        pubKey = m_rsa.readPublicKey(filename);
                        if (!pubKey)
                            continue; // go to next file
                        auto result = m_nodePubKeys.emplace(nodeId, pubKey);
                        // If key for NODE already exists, update it
                        if (!result.second)
                        {
                            result.first->second = pubKey;
                            LOGDEBUG(m_ownTypeInst,
                                     " Public Key for NODE with id=", nodeId, " was updated");
                        }
                        else
                        {
                            LOGDEBUG(m_ownTypeInst,
                                     " Public Key for NODE with id=", nodeId, " was inserted");
                        }
                    }
                    for (const auto &filename : ind->nodeFilesToRemove)
                    {
                        Uvr::UvrId_t nodeId = getNodeIdFromFilename(filename);
                        if (nodeId == Uvr::UvrId::undefined)
                            continue; // go to next file
                        auto nodeIter = m_nodePubKeys.find(nodeId);
                        if (nodeIter == m_nodePubKeys.end())
                        {
                            LOGDEBUG(m_ownTypeInst,
                                     " Public Key for NODE with id=", nodeId,
                                     " was not found to be deleted");
                            continue;
                        }
                        RSA_free(nodeIter->second);
                        m_nodePubKeys.erase(nodeIter);
                    }
                    nodeKeysGuard.unlock();
                }
                break;
            }
            case ::primitive::VERIFY_CALL_REQ:
            {
                handleVerifyCallRequest(prim);
            }
            case ::primitive::VERIFY_CALL_CNF:
            {
                auto conf = std::dynamic_pointer_cast<::primitive::VerifyCallConfirm>(prim);
                if (conf)
                {
                    if (conf->errorCode != 0)
                    {
                        boost::json::object errorBody;
                        boost::json::object &rspObj = errorBody[HTTP_RESPONSE_TAG].emplace_object();
                        boost::json::error_code err;
                        rspObj[HTTP_SERVICEKEY_TAG] = conf->verifyCallIndication->serviceKey;
                        rspObj[HTTP_UVRO_TAG] = conf->verifyCallIndication->uvrO;
                        rspObj[HTTP_UVRT_TAG] = conf->verifyCallIndication->uvrT;
                        rspObj[HTTP_SESSIONID_TAG] = doText(&conf->m_header->ulpKey, 8);
                        rspObj[HTTP_OPCODE_TAG] = HTTP_OPCODE_ERROR;
                        errorBody[HTTP_ERROR_TAG].emplace_object()[HTTP_ERRORCODE_TAG] = conf->errorCode;

                        m_hubMngr.sendReply(http::messages::result::ok,
                                            http::connections::id(conf->verifyCallIndication->connId),
                                            boost::json::serialize(errorBody));

                        return;
                    }

                    ::primitive::ServiceKey_t serviceKey = conf->verifyCallIndication->serviceKey;
                    Uvr::UvrId_t uvrO = conf->verifyCallIndication->uvrO;
                    Uvr::UvrId_t uvrT = conf->verifyCallIndication->uvrT;
                    ::primitive::SessionKeyOptional_t sessionKey = conf->verifyCallIndication->sessionKey;
                    ::primitive::ReleaseCode_t releaseCode = conf->releaseCode;

                    const Uvr::SessionId_t &sessionId = conf->m_header->ulpKey;

                    boost::json::object http_body;
                    boost::json::object &response = http_body[HTTP_RESPONSE_TAG].emplace_object();
                    response[HTTP_SERVICEKEY_TAG] = serviceKey;
                    response[HTTP_UVRO_TAG] = uvrO;
                    response[HTTP_UVRT_TAG] = uvrT;
                    response[HTTP_SESSIONID_TAG] = doText(&sessionId, 8);

                    uint8_t *buf_p1{nullptr};
                    size_t size_p1(0);

                    if (sessionKey.has_value())
                    {
                        buf_p1 = (uint8_t *)&(sessionKey.value());
                        size_p1 = 4;
                    }

                    std::array<char, 100> buf;
                    asn_enc_rval_t ec;
                    if (releaseCode == 0)
                    {
                        asn1::ContinueParams cont(buf_p1, size_p1);
                        ec = cont.AsnEncode(buf.data(), buf.size());
                    }
                    else
                    {
                        if (conf->additionalCode == 0)
                        {
                            asn1::ReleaseParams rels(buf_p1, size_p1,
                                                     releaseCode);
                            ec = rels.AsnEncode(buf.data(), buf.size());
                        }
                        else
                        {
                            asn1::ReleaseParams rels(buf_p1, size_p1,
                                                     releaseCode, conf->additionalCode);
                            ec = rels.AsnEncode(buf.data(), buf.size());
                        }
                    }

                    if (serviceKey == HTTP_SERVICEKEY_CYPHR)
                    {
                        auto pubKeyIter = m_nodePubKeys.find(uvrO);
                        if (pubKeyIter == m_nodePubKeys.end())
                        {
                            LOGERROR(m_ownTypeInst,
                                     " Failed to send ", (releaseCode == 0 ? "continue" : "release"),
                                     " response to UVR=", uvrO,
                                     " on request (id=", sessionId,
                                     "): public key is not found");
                            return;
                        }
                        std::string inData(buf.data(), ec.encoded);
                        std::string outData;
                        if (!m_rsa.encrypt(pubKeyIter->second, inData, outData))
                        {
                            LOGERROR(m_ownTypeInst,
                                     " Failed to send ", (releaseCode == 0 ? "continue" : "release"),
                                     " response to UVR=", uvrO,
                                     " on request (id=", sessionId,
                                     "): failed to encrypt");
                            return;
                        }
                        response[HTTP_DATA_TAG] = doText(outData.c_str(), outData.size());
                    }
                    else if (serviceKey == HTTP_SERVICEKEY_PLAIN)
                    {
                        response[HTTP_DATA_TAG] = doText(buf.data(), ec.encoded);
                    }
                    else
                    {
                        LOGERROR(m_ownTypeInst,
                                 " Failed to send ", (releaseCode == 0 ? "continue" : "release"),
                                 " response to UVR=", uvrO,
                                 " on request (id=", sessionId,
                                 "): unknown serviceKey=", serviceKey);
                        return;
                    }
                    response[HTTP_OPCODE_TAG] = releaseCode == 0 ? HTTP_OPCODE_CONTINUE : HTTP_OPCODE_RELEASE;

                    m_hubMngr.sendReply(http::messages::result::ok,
                                        http::connections::id(conf->verifyCallIndication->connId),
                                        boost::json::serialize(http_body));
                }
                break;
            }
            case ::primitive::VERIFY_REQ_ABORT:
            {
                auto abort =
                    std::dynamic_pointer_cast<::primitive::VerifyRequestAbort>(prim);
                if (abort)
                {
                    auto connPtr = m_hubMngr.getOutgoingConnection(
                        http::connections::id(abort->m_header->ulpKey));
                    if (connPtr == nullptr)
                    {
                        LOGWARN(m_ownTypeInst,
                                " Connection with id=", abort->m_header->ulpKey,
                                " does not exist.");
                        return;
                    }
                    connPtr->shutdown();
                }
                break;
            }
            case ::primitive::QUERY_UVR_INFO_REQ:
            {
                auto req = std::dynamic_pointer_cast<::primitive::QueryUvrInfoReq>(prim);
                if (req)
                {
                    db::uvr::record rec;
                    db::interfaces::tables::uvrs::client::get(rec, req->uvrId);
                    if (rec.m_idUvr == Uvr::UvrId::undefined)
                    {
                        LOGWARN(m_ownTypeInst,
                                " Cannot get record from UVRs table for UVR ID =", req->uvrId);
                    }

                    auto rsp = std::make_shared<primitive::QueryUvrInfoRsp>(rec);
                    rsp->m_header->to = prim->m_header->from;
                    rsp->m_header->ulpKey = prim->m_header->ulpKey;
                    workers::node::send(rsp);
                }
                break;
            }
            case ::primitive::QUERY_HUB_INFO_REQ:
            {
                auto req = std::dynamic_pointer_cast<::primitive::QueryHubInfoReq>(prim);
                if (req)
                {
                    db::hub::rows recs;
                    db::gts_t hubIds = {req->hubId};
                    db::interfaces::tables::hubs::client::get(recs, hubIds);
                    if (recs.size() == 0)
                    {
                        LOGWARN(m_ownTypeInst,
                                " Cannot get record for HUB with ID = ", req->hubId);
                        recs.emplace_back("", "", "", "", "");
                    }

                    auto rsp = std::make_shared<primitive::QueryHubInfoRsp>(recs.front());
                    rsp->m_header->to = prim->m_header->from;
                    rsp->m_header->ulpKey = prim->m_header->ulpKey;
                    workers::node::send(rsp);
                }
                break;
            }
            case ::primitive::QUERY_OPERATOR_INFO_REQ:
            {
                auto req = std::dynamic_pointer_cast<::primitive::QueryOperatorInfoReq>(prim);
                if (req)
                {
                    db::Operator::Rows recs;
                    const db::OperatorsIds operatorsIds = {req->operatorId};
                    db::interfaces::tables::Operators::client::get(recs, operatorsIds);
                    if (recs.size() == 0)
                    {
                        LOGWARN(m_ownTypeInst,
                                " Can't get record for Operator with ID = ", req->operatorId);
                        recs.emplace_back(0, "", "", "", "", "");
                    }

                    const auto rsp = std::make_shared<primitive::QueryOperatorInfoRsp>(recs.front());
                    rsp->m_header->to = prim->m_header->from;
                    rsp->m_header->ulpKey = prim->m_header->ulpKey;
                    workers::node::send(rsp);
                }
                break;
            }
            default:
            {
                pm::node::handleSpecificPrimitive(prim);
            }
        }
    }

    void node::handleVerifyCallRequest(const primitive::GenericPtr &prim)
    {
        const auto req = std::dynamic_pointer_cast<::primitive::VerifyCallRequest>(prim);
        const auto sessionId = req->m_header->ulpKey;

        auto policyResponse = checkPreVerifyPolicy(req->uvrP, sessionId);
        if (policyResponse != nullptr)
        {
            workers::node::send(policyResponse);
            return;
        }

        ::primitive::SessionKey_t sessionKey;
        RAND_bytes(reinterpret_cast<unsigned char *>(&sessionKey), sizeof(sessionKey));
        LOGD(m_logTag,
             "sessionKey=", sessionKey,
             " has been generated for HTTP request(id=", sessionId, ").");

        std::string sessionIdText = doText(&sessionId, 8);

        asn1::RequestParams r1((uint8_t *)&sessionKey, 4,
                               req->verifyCallQuery->callingNumber->data(), req->verifyCallQuery->callingNumber->size(),
                               req->verifyCallQuery->calledNumber->data(), req->verifyCallQuery->calledNumber->size(),
                               req->verifyCallQuery->originalCalledNumber ? req->verifyCallQuery->originalCalledNumber->data() : nullptr,
                               req->verifyCallQuery->originalCalledNumber ? req->verifyCallQuery->originalCalledNumber->size() : 0);

        std::array<char, 100> buf;
        asn_enc_rval_t ec = r1.AsnEncode(buf.data(), buf.size());
        std::string requestData(buf.data(), ec.encoded);

        // Store session information.
        const auto dataPtr =
            std::make_shared<http::VerifyCallData>(req->verifyCallQuery->idSrc,
                                                   req->uvrS,
                                                   sessionKey,
                                                   sessionIdText,
                                                   requestData);

        // Encrypt data if needed.
        Uvr::UvrId_t uvrId = req->uvrP;
        if (m_httpDataEncrypt)
        {
            // First try primary UVR.
            std::string encryptError = encryptRequestData(uvrId, requestData);
            if (encryptError.empty())
            {
                dataPtr->m_usedUvrS = false;
            }
            else
            {
                // Failed to encrypt with primary UVR, try secondary.
                LOGW(m_logTag,
                     "Can't encrypt verify request (id=", sessionIdText, ")",
                     " for primary UVR=", uvrId, ".",
                     " Reason: ", encryptError,
                     " Trying secondary UVR.");

                policyResponse = checkPreVerifyPolicy(req->uvrS, sessionId);
                if (policyResponse != nullptr)
                {
                    workers::node::send(policyResponse);
                    return;
                }

                uvrId = req->uvrS;
                encryptError = encryptRequestData(uvrId, requestData);
                if (encryptError.empty())
                {
                    dataPtr->m_usedUvrS = true;
                }
                else
                {
                    LOGE(m_logTag,
                         "Failed to send verify request (id=", sessionIdText, ")",
                         " to secondary UVR=", uvrId, ".",
                         " Reason: ", encryptError);
                    increaseCounter(pm::counter::calls::ERR2);
                    return;
                }
            }
        }

        boost::json::object httpBody;
        genVerifyRequestHttpBody(httpBody, uvrId, sessionIdText, requestData);
        m_hubMngr.startAndSendRequest(sessionId,
                                      HTTP_VERIFY_CALL_TARGET,
                                      boost::json::serialize(httpBody),
                                      std::static_pointer_cast<void>(dataPtr));
    }

    std::string node::encryptRequestData(const Uvr::UvrId_t uvrId,
                                         std::string &requestData)
    {
        if (uvrId == Uvr::UvrId::undefined)
        {
            return "UVR ID is undefined.";
        }

        std::shared_lock<std::shared_mutex> nodeKeysLock(m_nodeKeysMutex);
        const auto nodeIter = m_nodePubKeys.find(uvrId);
        if (nodeIter == m_nodePubKeys.end())
        {
            return "Public key is not found.";
        }

        std::string tmpData;
        if (!m_rsa.encrypt(nodeIter->second, requestData, tmpData))
        {
            return "Can't encrypt with provided public key.";
        }

        requestData = std::move(tmpData);
        return "";
    }

    void node::handleRequest(http::connections::id connId,
                             const std::string_view &target,
                             const std::string &payloadData)
    {
        boost::json::error_code errorCode;
        boost::json::value bodyVal = boost::json::parse(payloadData, errorCode);
        if (errorCode)
        {
            std::string errText = std::string("json parsing failed: ") + errorCode.message();
            HTTP_REPLY_BAD_REQUEST(connId, target, errText);
            return;
        }

        if (target.compare(HTTP_VERIFY_CALL_TARGET) == 0)
        {
            boost::json::object *bodyObjPtr = bodyVal.if_object();
            if (!bodyObjPtr)
            {
                std::string errText = std::string("invalid structure of json body: ") + boost::json::serialize(bodyVal);
                HTTP_REPLY_BAD_REQUEST(connId, target, errText);
                return;
            }
            const boost::json::value *reqValPtr = bodyObjPtr->if_contains(HTTP_REQUEST_TAG);
            if (!reqValPtr)
            {
                std::string errText = std::string("request section is not found: ") + boost::json::serialize(bodyVal);
                HTTP_REPLY_BAD_REQUEST(connId, target, errText);
                return;
            }
            GET_MANDATORY_HTTP_REQUEST_BODY_FIELD(connId, target, reqValPtr, bodyVal,
                                                  "/" HTTP_REQUEST_TAG "/" HTTP_SERVICEKEY_TAG,
                                                  ::primitive::ServiceKey_t, serviceKey);
            GET_MANDATORY_HTTP_REQUEST_BODY_FIELD(connId, target, reqValPtr, bodyVal,
                                                  "/" HTTP_REQUEST_TAG "/" HTTP_UVRO_TAG,
                                                  Uvr::UvrId_t, uvrO);
            GET_MANDATORY_HTTP_REQUEST_BODY_FIELD(connId, target, reqValPtr, bodyVal,
                                                  "/" HTTP_REQUEST_TAG "/" HTTP_UVRT_TAG,
                                                  Uvr::UvrId_t, uvrT);
            if (uvrT != m_uvrId)
            {
                std::string errText =
                    "wrong " HTTP_UVRT_TAG "=" + std::to_string(uvrT)
                    + ", expected " HTTP_UVRT_TAG "=" + std::to_string(m_uvrId);
                HTTP_RESPONSE_ERROR(connId, target, HTTP_UVR_OUT_OF_SERVICE, reqValPtr, errText);
                return;
            }

            GET_MANDATORY_HTTP_REQUEST_BODY_FIELD(connId, target, reqValPtr, bodyVal,
                                                  "/" HTTP_REQUEST_TAG "/" HTTP_SESSIONID_TAG,
                                                  ::primitive::raw_data_t, sessionIdHex);
            std::string sessionIdRaw = asn1::UndoText(sessionIdHex);
            Uvr::SessionId_t sessionId =
                *(reinterpret_cast<Uvr::SessionId_t *>(&(sessionIdRaw[0])));
            GET_MANDATORY_HTTP_REQUEST_BODY_FIELD(connId, target, reqValPtr, bodyVal,
                                                  "/" HTTP_REQUEST_TAG "/" HTTP_DATA_TAG,
                                                  ::primitive::raw_data_t, data);
            GET_MANDATORY_HTTP_REQUEST_BODY_FIELD(connId, target, reqValPtr, bodyVal,
                                                  "/" HTTP_REQUEST_TAG "/" HTTP_OPCODE_TAG,
                                                  std::string, opCode);

            if (opCode != HTTP_OPCODE_REQUEST)
            {
                std::string errText =
                    std::string("wrong ") + HTTP_OPCODE_TAG + "=" + opCode;
                HTTP_RESPONSE_ERROR(connId, target, HTTP_UVR_ERRORCODE_ILLEGAL_VALUE, reqValPtr, errText);
                return;
            }

            std::string asnData;
            if (serviceKey == HTTP_SERVICEKEY_CYPHR) // Decrypt meaningful data of incoming HTTP request
            {
                std::string undoData = asn1::UndoText(data);
                if (!m_rsa.decrypt(undoData, asnData))
                {
                    std::string errText = "Failed to decrypt received data by own private key";
                    HTTP_RESPONSE_ERROR(connId, target, HTTP_UVR_ERRORCODE_ILLEGAL_VALUE, reqValPtr, errText);
                    return;
                }
            }
            else if (serviceKey == HTTP_SERVICEKEY_PLAIN)
            {
                asnData = asn1::UndoText(data);
            }
            else
            {
                std::string errText =
                    std::string(HTTP_SERVICEKEY_TAG) + "=" + std::to_string(serviceKey) + " is not supported";
                HTTP_RESPONSE_ERROR(connId, target, HTTP_UVR_ERRORCODE_ILLEGAL_VALUE, reqValPtr, errText);
                return;
            }

            asn_dec_rval_t rval;
            asn1::RequestParamsPtr req =
                asn1::RequestParams::AsnDecode(asnData.c_str(), asnData.size(), &rval);

            if (rval.code != RC_OK)
            {
                std::string errText =
                    std::string("wrong data as ASN1 decode failed with binary=") + doText(data.c_str(), data.size());
                HTTP_RESPONSE_ERROR(connId, target, HTTP_UVR_ERRORCODE_ILLEGAL_VALUE, reqValPtr, errText);
                return;
            }

            std::shared_ptr<Uvr::OctetString> originalCalledNumber_p;
            if (req->originalCalledNumber)
            {
                originalCalledNumber_p =
                    std::make_shared<Uvr::OctetString>(req->originalCalledNumber->buf,
                                                       req->originalCalledNumber->buf + req->originalCalledNumber->size);
            }

            ::primitive::SessionKeyOptional_t sessionKey;
            if (req->sessionKey != nullptr)
            {
                sessionKey = *(reinterpret_cast<::primitive::SessionKey_t *>(req->sessionKey->buf));
            }

            // This implementation of UVR will not generate HTTP_UVR_NUMBER_OUT_OF_SERVICE error:
            // According to UVR-UVR v10.6 protocol UVR could not check
            // that calling number is served by this UVR,
            // It could just send release response if call is absent in calls registry
            auto ind = std::make_shared<primitive::VerifyCallIndication>(
                connId.m_value,
                serviceKey, uvrO, uvrT, sessionKey,
                std::make_shared<Uvr::OctetString>(req->callingPartyNumber.buf,
                                                   req->callingPartyNumber.buf + req->callingPartyNumber.size),
                std::make_shared<Uvr::OctetString>(req->calledPartyNumber.buf,
                                                   req->calledPartyNumber.buf + req->calledPartyNumber.size),
                originalCalledNumber_p);

            ind->m_header->to = TypeInst(workers::type::VFS, 0);
            ind->m_header->ulpKey = sessionId;

            workers::node::send(ind);
        }
        else
        {
            // TODO: specify is it correct or 404 should be used when URI is absent in HTTP message???
            std::string errText = "URI is not found, unknown target=" + std::string(target.data(), target.size());
            HTTP_REPLY_NOT_FOUND(connId, target, errText);
        }
    }

    void node::handleResponse(http::messages::result result,
                              const http::connection<http::role::CLIENT> *connPtr,
                              const std::string &payloadData)
    {
        const std::string_view target = connPtr->outgoingRequest().target();
#if LOG_MACRO_LEVEL >= LOG_LEVEL_WARN
        const http::connections::id &connId = connPtr->connId();
#endif
        const Uvr::SessionId_t &sessionIdC = connPtr->sessionId();

        // Http responses with NOT OK result has text content
        if (result != http::messages::result::ok)
        {
            LOGW(m_ownTypeInst,
                 " Http request (connId=", connId, ";target=", target, ") failed due to: ",
                 payloadData,
                 ".Trying to resent request on another HUB host.");
            increaseCounter(pm::counter::http::request_failed,
                            ::pm::counter::value(1));
            if (!m_hubMngr.resendRequestOnNextHost(connPtr))
            {
                LOGW(m_ownTypeInst,
                     " Failed to resend http request (connId=",
                     connId, ";target=", target,
                     ") on another HUB host.");
            }
            return;
        }

        // Http responses with OK result has json content
        boost::json::error_code errorCode;
        boost::json::value bodyVal = boost::json::parse(payloadData, errorCode);
        if (errorCode)
        {
            LOGW(m_ownTypeInst,
                 " Bad HTTP response (connId=", connId, ";target=", target, ";res=", result,
                 "): json parsing failed: ", errorCode.message());
            increaseCounter(pm::counter::http::bad_response,
                            ::pm::counter::value(1));
            /* ERR1 to be increased after timeout from VFC */
            return;
        }

        if (target.compare(HTTP_VERIFY_CALL_TARGET) == 0)
        {
            std::shared_ptr<http::VerifyCallData> dataPtr =
                std::static_pointer_cast<http::VerifyCallData>(connPtr->userData());
            assert(dataPtr != nullptr);

            GET_MANDATORY_HTTP_RESPONSE_BODY_FIELD(bodyVal, "/" HTTP_RESPONSE_TAG "/" HTTP_SERVICEKEY_TAG,
                                                   ::primitive::ServiceKey_t, serviceKey);
            GET_MANDATORY_HTTP_RESPONSE_BODY_FIELD(bodyVal, "/" HTTP_RESPONSE_TAG "/" HTTP_SESSIONID_TAG,
                                                   ::primitive::raw_data_t, sessionIdHex);
            std::string sessionIdRaw = asn1::UndoText(sessionIdHex);
            Uvr::SessionId_t sessionId = *(reinterpret_cast<Uvr::SessionId_t *>(&(sessionIdRaw[0])));
            if (sessionId != sessionIdC)
            {
                LOGW(m_ownTypeInst,
                     " Received http response (connId=", connId, ")",
                     " with sessionId=", sessionId,
                     " different that the one of the request sessionIdC=", sessionIdC,
                     ", payloadData:\n", payloadData);
                increaseCounter(pm::counter::http::bad_response,
                                ::pm::counter::value(1));
                /* ERR1 to be increased after timeout from VFC */
                return;
            }

            GET_MANDATORY_HTTP_RESPONSE_BODY_FIELD(bodyVal, "/" HTTP_RESPONSE_TAG "/" HTTP_UVRO_TAG, Uvr::UvrId_t, uvrO);
            GET_MANDATORY_HTTP_RESPONSE_BODY_FIELD(bodyVal, "/" HTTP_RESPONSE_TAG "/" HTTP_UVRT_TAG, Uvr::UvrId_t, uvrT);
            if (uvrO != m_uvrId)
            {
                LOGW(m_ownTypeInst,
                     " Verification response received from uvrT=", uvrT,
                     " on request (connId=", connId, ";id=", sessionId, ";target=", target,
                     ") has invalid uvrO=", uvrO, " (expected own VFN=", m_uvrId,
                     "), payloadData:\n", payloadData);
                increaseCounter(pm::counter::http::bad_response,
                                ::pm::counter::value(1));
                /* ERR1 to be increased after timeout from VFC */
                return;
            }

            GET_MANDATORY_HTTP_RESPONSE_BODY_FIELD(bodyVal, "/" HTTP_RESPONSE_TAG "/" HTTP_OPCODE_TAG,
                                                   std::string, opCode);

            Uvr::VerificationVerdict_t rlc = Uvr::VerificationVerdict::timeout;
            if (opCode == HTTP_OPCODE_ERROR)
            {
                increaseCounter(pm::counter::http::response_error,
                                ::pm::counter::value(1));

                GET_MANDATORY_HTTP_RESPONSE_BODY_FIELD(bodyVal, "/" HTTP_ERROR_TAG "/" HTTP_ERRORCODE_TAG,
                                                       int, errorCode);
                switch (errorCode)
                {
                    case HTTP_HUB_ERRORCODE_UVR_IS_NOT_FOUND:
                        /* fall through */
                    case HTTP_HUB_ERRORCODE_UVR_ROUTING_ERROR:
                        /* fall through */
                    case HTTP_HUB_ERRORCODE_TIMEOUT:
                        /* fall through */
                    case HTTP_HUB_ERRORCODE_ORIGIN_ILLEGAL_VALUE:
                        /* fall through */
                    case HTTP_HUB_ERRORCODE_ORIGIN_ILLEGAL_PARAM:
                        /* fall through */
                    case HTTP_HUB_ERRORCODE_ORIGIN_MISSING_PARAM:
                        /* fall through */
                    case HTTP_HUB_ERRORCODE_REMOTE_ILLEGAL_VALUE:
                        /* fall through */
                    case HTTP_HUB_ERRORCODE_REMOTE_ILLEGAL_PARAM:
                        /* fall through */
                    case HTTP_HUB_ERRORCODE_REMOTE_MISSING_PARAM:
                    {
                        // Check HUB signature for error generated by HUB
                        if (serviceKey == HTTP_SERVICEKEY_CYPHR)
                        {
                            CHECK_HTTP_ERROR_SIGNATURE(bodyVal, errorCode, connPtr->address());
                        }
                    } // continue handlig error
                    case HTTP_UVR_ERRORCODE_ILLEGAL_VALUE:
                        /* fall through */
                    case HTTP_UVR_ERRORCODE_ILLEGAL_PARAM:
                        /* fall through */
                    case HTTP_UVR_ERRORCODE_MISSING_PARAM:
                        /* fall through */
                    case HTTP_UVR_ERRORCODE_OVERLOAD:
                        /* fall through */
                    case HTTP_UVR_NUMBER_OUT_OF_SERVICE:
                        /* fall through */
                    case HTTP_UVR_OUT_OF_SERVICE:
                        /* fall through */
                    case HTTP_UVR_ERRORCODE_SYSTEM_ERROR:
                    {
                        // Trying to resend http request to secondary UVR.
                        LOGW(m_logTag,
                             "Received HTTP response with error code=", errorCode, ".",
                             " Trying to resend HTTP request (connId=", connId, ";id=", sessionId, ";target=", target,
                             ") to secondary UVR.");

                        // No secondary UVR.
                        if (dataPtr->m_idUvrS == 0)
                        {
                            LOGW(m_logTag,
                                 "Failed to resend HTTP request (connId=", connId, ";id=", sessionId, ";target=", target,
                                 ") to secondary UVR. Reason: UVR ID is not set.");

                            increaseCounter(pm::counter::http::request_routing_to_uvr_failed);
                            break; // Send verdict with timeout if routing is impossible.
                        }
                        // Request was already sent to it early.
                        else if (dataPtr->m_usedUvrS)
                        {
                            LOGW(m_logTag,
                                 "Failed to resend HTTP request (connId=", connId, ";id=", sessionId, ";target=", target,
                                 ") to secondary UVR. Reason: Request was already sent to secondary UVR=", dataPtr->m_idUvrS);

                            increaseCounter(pm::counter::http::request_routing_to_uvr_failed);
                            break; // Send verdict with timeout if routing is impossible.
                        }
                        else
                        {
                            // Check policy before resend.
                            const auto policyResponse = checkPreVerifyPolicy(dataPtr->m_idUvrS, sessionId);
                            if (policyResponse != nullptr)
                            {
                                workers::node::send(policyResponse);
                                return;
                            }

                            // Encrypt data by secondary UVR key.
                            std::string requestData(dataPtr->m_reqData);
                            if (serviceKey == HTTP_SERVICEKEY_CYPHR)
                            {
                                const std::string encryptError = encryptRequestData(dataPtr->m_idUvrS, requestData);
                                if (!encryptError.empty())
                                {
                                    LOGW(m_logTag,
                                         "Failed to resend HTTP request (connId=", connId, ";id=", sessionId, ";target=", target,
                                         ") to secondary UVR. Reason: ", encryptError);

                                    increaseCounter(pm::counter::http::request_routing_to_uvr_failed);
                                    break; // Send verdict with timeout if routing is impossible.
                                }
                            }

                            dataPtr->m_usedUvrS = true;

                            boost::json::object httpBody;
                            genVerifyRequestHttpBody(httpBody,
                                                     dataPtr->m_idUvrS,
                                                     dataPtr->m_sessionId,
                                                     requestData);

                            if (m_hubMngr.startAndSendRequest(sessionId,
                                                              HTTP_VERIFY_CALL_TARGET,
                                                              boost::json::serialize(httpBody),
                                                              connPtr->userData()))
                            {
                                return;
                            }
                        }

                        increaseCounter(pm::counter::http::request_routing_to_uvr_failed);
                        break; // Send verdict with timeout if routing is impossible.
                    }
                    case HTTP_HUB_ERRORCODE_OVERLOAD:
                    case HTTP_HUB_ERRORCODE_SYSTEM_ERROR:
                    {
                        // Resend http request to next HUB host
                        LOGW(m_ownTypeInst,
                             " Received http response with error code=", errorCode,
                             ". Trying to resend http request (connId=",
                             connId, ";id=", sessionId, ";target=", target,
                             ") on another HUB host.");
                        // Check HUB signature for error generated by HUB
                        if (serviceKey == HTTP_SERVICEKEY_CYPHR)
                        {
                            CHECK_HTTP_ERROR_SIGNATURE(bodyVal, errorCode, connPtr->address());
                        }
                        // Try to route on another host
                        if (m_hubMngr.resendRequestOnNextHost(connPtr))
                            return;
                        LOGW(m_ownTypeInst,
                             " Failed to resend http request (connId=",
                             connId, ";target=", target,
                             ") on another HUB host.");
                        break; // send verdict with timeout if routing is impossible
                    }
                    default:
                    {
                        LOGW(m_ownTypeInst,
                             " Http Error response was received with unknown errorCode=",
                             errorCode, "on request (connId=",
                             connId, ";id=", sessionId, ";target=", target, ")");
                        increaseCounter(pm::counter::http::bad_response,
                                        ::pm::counter::value(1));
                        /* ERR1 to be increased after timeout from VFC */
                        return;
                    }
                }
            }
            else if (opCode == HTTP_OPCODE_CONTINUE || HTTP_OPCODE_RELEASE)
            {
                ::pm::counter::id cntrId;
                if (opCode == HTTP_OPCODE_CONTINUE)
                {
                    cntrId = pm::counter::http::response_continue;
                }
                else
                {
                    cntrId = pm::counter::http::response_release;
                }
                increaseCounter(cntrId, ::pm::counter::value(1));

                GET_MANDATORY_HTTP_RESPONSE_BODY_FIELD(bodyVal, "/" HTTP_RESPONSE_TAG "/" HTTP_DATA_TAG,
                                                       ::primitive::raw_data_t, data);
                asn_dec_rval_t rval;
                std::string asnData;
                // Decrypt data if needed
                if (serviceKey == HTTP_SERVICEKEY_CYPHR)
                {
                    std::string inData = asn1::UndoText(data);
                    if (!m_rsa.decrypt(inData, asnData))
                    {
                        LOGW(m_ownTypeInst,
                             " Failed to decrypt data of ",
                             (opCode == HTTP_OPCODE_CONTINUE ? "Continue" : "Release"),
                             " message received on request (connId=",
                             connId, ";id=", sessionId, ";target=", target, ")");
                        increaseCounter(pm::counter::http::bad_response,
                                        ::pm::counter::value(1));
                        /* ERR1 to be increased after timeout from VFC */
                        return;
                    }
                }
                else if (serviceKey == HTTP_SERVICEKEY_PLAIN)
                {
                    asnData = asn1::UndoText(data);
                }
                else
                {
                    LOGW(m_ownTypeInst,
                         (opCode == HTTP_OPCODE_CONTINUE ? " Continue" : " Release"),
                         " message with unknown serviceKey=", serviceKey,
                         " was received on request (connId=",
                         connId, ";id=", sessionId, ";target=", target, ")");
                    increaseCounter(pm::counter::http::bad_response,
                                    ::pm::counter::value(1));
                    /* ERR1 to be increased after timeout from VFC */
                    return;
                }
                ::primitive::SessionKeyOptional_t sessionKey;
                if (opCode == HTTP_OPCODE_CONTINUE)
                {
                    asn1::ContinueParamsPtr cont =
                        asn1::ContinueParams::AsnDecode(asnData.c_str(), asnData.size(), &rval);
                    if (rval.code != RC_OK)
                    {
                        LOGW("ASN1 decode failed with Continue data binary: ", data);
                        increaseCounter(pm::counter::http::bad_response,
                                        ::pm::counter::value(1));
                        /* ERR1 to be increased after timeout from VFC */
                        return;
                    }
                    rlc = Uvr::VerificationVerdict::success;
                    if (cont->sessionKey != nullptr
                        && cont->sessionKey->buf != nullptr)
                    {
                        sessionKey = *(reinterpret_cast<::primitive::SessionKey_t *>(cont->sessionKey->buf));
                    }
                }
                else
                {
                    asn1::ReleaseParamsPtr rels =
                        asn1::ReleaseParams::AsnDecode(asnData.c_str(), asnData.size(), &rval);
                    if (rval.code != RC_OK)
                    {
                        LOGW("ASN1 decode failed with Release data binary: ", data);
                        increaseCounter(pm::counter::http::bad_response,
                                        ::pm::counter::value(1));
                        /* ERR1 to be increased after timeout from VFC */
                        return;
                    }
                    LOGI(m_ownTypeInst,
                         " Release with releaseCode=", rels->releaseCode,
                         " and additionalCode=",
                         (rels->additionalCode ? std::to_string(*rels->additionalCode)
                                               : "missing"),
                         " is received on http request (connId=",
                         connId, ";id=", sessionId, ";target=", target, ")");
                    rlc = Uvr::VerificationVerdict::failure;
                    if (rels->sessionKey != nullptr
                        && rels->sessionKey->buf != nullptr)
                    {
                        sessionKey = *(reinterpret_cast<::primitive::SessionKey_t *>(rels->sessionKey->buf));
                    }
                }
                // Check the authenticity of the remote side by sessionKey
                if (serviceKey == HTTP_SERVICEKEY_CYPHR
                    && (sessionKey.has_value() == false
                        || sessionKey.value() != dataPtr->m_sessionKey))
                {
                    LOGW(m_ownTypeInst,
                         " SessionKey=",
                         (sessionKey.has_value() ? std::to_string(sessionKey.value()) : "missing"),
                         " received in ",
                         (opCode == HTTP_OPCODE_CONTINUE ? "Continue" : "Release"),
                         " message doesn't match with sessionKey=", dataPtr->m_sessionKey,
                         " generated for request (connId=",
                         connId, ";id=", sessionId, ";target=", target, ")");
                    increaseCounter(pm::counter::http::bad_response,
                                    ::pm::counter::value(1));
                    /* ERR1 to be increased after timeout from VFC */
                    return;
                }
            }
            else
            {
                LOGW(HTTP_OPCODE_TAG, "=", opCode, " is invalid");
                increaseCounter(pm::counter::http::bad_response,
                                ::pm::counter::value(1));
                /* ERR1 to be increased after timeout from VFC */
                return;
            }

            // Check policy before sending response.
            auto response = checkPostVerifyPolicy(uvrT, rlc, sessionId);
            workers::node::send(response);
        }
        else
        {
            LOGW(m_ownTypeInst,
                 " Received http response on request (connId=", connId,
                 ") with unknown target=\"", target, "\"");
            increaseCounter(pm::counter::http::bad_response,
                            ::pm::counter::value(1));
        }
    }

    void node::genVerifyRequestHttpBody(boost::json::object &httpBody,
                                        Uvr::UvrId_t toVfnId,
                                        const std::string &sessionId,
                                        const std::string &reqData) const
    {
        boost::json::object &request = httpBody[HTTP_REQUEST_TAG].emplace_object();
        request[HTTP_SERVICEKEY_TAG] = m_httpDataEncrypt ? HTTP_SERVICEKEY_CYPHR : HTTP_SERVICEKEY_PLAIN;
        request[HTTP_UVRO_TAG] = m_uvrId;
        request[HTTP_UVRT_TAG] = toVfnId;
        request[HTTP_SESSIONID_TAG] = sessionId;
        request[HTTP_DATA_TAG] = doText(reqData);
        request[HTTP_OPCODE_TAG] = HTTP_OPCODE_REQUEST;
    }

    Uvr::Gt_t node::getHubIdFromFilename(const std::string &filename) const
    {
        if (filename.empty())
            return std::string();
        std::size_t posBeg = filename.rfind("hub") + std::strlen("hub") + 1;
        std::size_t posEnd = filename.rfind("-key.pub");
        if (posBeg == std::string::npos
            || posEnd == std::string::npos)
        {
            LOGWARN("Unexpected HUB public key file format: filename=\"", filename, "\"");
            return std::string();
        }
        return std::string(filename, posBeg, posEnd - posBeg);
    }

    Uvr::UvrId_t node::getNodeIdFromFilename(const std::string &filename) const
    {
        if (filename.empty())
            return Uvr::UvrId::undefined;
        std::size_t posBeg = filename.rfind("node") + std::strlen("node") + 1;
        std::size_t posEnd = filename.rfind("-key.pub");
        if (posBeg == std::string::npos
            || posEnd == std::string::npos)
        {
            LOGWARN("Unexpected NODE public key file format: filename=\"", filename, "\"");
            return Uvr::UvrId::undefined;
        }
        return static_cast<Uvr::UvrId_t>(std::atol(std::string(filename, posBeg, posEnd - posBeg).c_str()));
    }

    bool node::checkHubSignature(Uvr::Gt_t hubId,
                                 const std::string &signature)
    {
        std::string rawSignature(asn1::UndoText(signature));
        std::shared_lock<std::shared_mutex> hubKeysGuard(m_hubKeysGuard);
        auto hubIter = m_hubPubKeys.find(hubId);
        if (hubIter == m_hubPubKeys.end())
        {
            LOGERROR(m_ownTypeInst,
                     " Failed to check signature for HUB=", hubId,
                     ": HUB public key is not found");
            return false;
        }
        std::string decoded;
        if (!m_rsa.signRecover(hubIter->second, rawSignature, decoded))
            return false;
        if (hubId != decoded)
        {
            LOGERROR(m_ownTypeInst,
                     " Failed to recovery HUB id=\"", hubId,
                     "\" from signature=\"", signature,
                     "\", decoded value=\"", decoded, "\"");
            return false;
        }
        return true;
    }

    void node::clearPubKeys(void)
    {
        std::unique_lock<std::shared_mutex> hubKeysGuard(m_hubKeysGuard);
        for (auto const &hubIter : m_hubPubKeys)
        {
            RSA_free(hubIter.second);
        }
        m_hubPubKeys.clear();
        std::unique_lock<std::shared_mutex> nodeKeysGuard(m_nodeKeysMutex);
        for (auto const &nodeIter : m_nodePubKeys)
        {
            RSA_free(nodeIter.second);
        }
        m_nodePubKeys.clear();
    }

    std::shared_ptr<primitive::VerifyCallResponse>
    node::checkPreVerifyPolicy(
        const Uvr::UvrId_t uvrId,
        const Uvr::SessionId_t sessionId)
    {
        db::uvr::record uvrRecord;
        if (!db::interfaces::tables::uvrs::client::get(uvrRecord, uvrId))
        {
            return nullptr;
        }

        const auto defPolicy = Uvr::UvrInfo::getDefPolicy(uvrRecord.m_metaInfo);
        if (!defPolicy.has_value())
        {
            return nullptr;
        }

        const auto response =
            std::make_shared<primitive::VerifyCallResponse>(
                    Uvr::VerificationVerdict::undefined,
                    Uvr::VerificationVerdict::undefined,
                    uvrId);

        response->m_header->to = TypeInst(workers::type::VFC, 0);
        response->m_header->ulpKey = sessionId;

        switch (defPolicy.value())
        {
            case Uvr::UvrInfo::DefPolicy::NoChange:
            {
                return nullptr;
            }
            case Uvr::UvrInfo::DefPolicy::Blocked:
            {
                response->verificationVerdict = Uvr::VerificationVerdict::failure;
                response->incidentVerdict = Uvr::VerificationVerdict::blocked;

                return response;
            }
            case Uvr::UvrInfo::DefPolicy::Allowed:
            {
                response->verificationVerdict = Uvr::VerificationVerdict::success;
                response->incidentVerdict = Uvr::VerificationVerdict::allowed;

                return response;
            }
        }

        return nullptr;
    }

    std::shared_ptr<primitive::VerifyCallResponse>
    node::checkPostVerifyPolicy(
        const Uvr::UvrId_t uvrId,
        const Uvr::VerificationVerdict_t verificationVerdict,
        const Uvr::SessionId_t sessionId)
    {
        const auto response =
            std::make_shared<primitive::VerifyCallResponse>(
                    verificationVerdict,
                    Uvr::VerificationVerdict::undefined,
                    uvrId);

        response->m_header->to = TypeInst(workers::type::VFC, 0);
        response->m_header->ulpKey = sessionId;

        db::uvr::record uvrRecord;
        if (!db::interfaces::tables::uvrs::client::get(uvrRecord, uvrId))
        {
            return response;
        }

        const auto status = Uvr::UvrInfo::getStatus(uvrRecord.m_metaInfo);
        if (!status.has_value())
        {
            return response;
        }

        switch (status.value())
        {
            case Uvr::UvrInfo::Status::Active:
            {
                return response;
            }
            case Uvr::UvrInfo::Status::Maintenance:
            {
                response->incidentVerdict = verificationVerdict;

                if (verificationVerdict == Uvr::VerificationVerdict::failure)
                {
                    response->verificationVerdict = Uvr::VerificationVerdict::success;
                }

                return response;
            }
        }

        return response;
    }

} // namespace vfgw
