#include "Config.hpp"
#include "vfs/src/node.hpp"

// AFAP includes.
#include "components/csv/src/CsvFile.hpp"
#include "components/db/src/mimic.hpp"
#include "components/filesystem/src/filesystem.hpp"
#include "components/timers/src/timer_periodic.hpp"
#include "components/worker_node/src/config_parameters.hpp"
#include "vfgw/src/http_message.hpp"
#include "TRecQ763Number.hpp"

namespace vfs
{
    /// Times to store information about outgoing calls.
#ifndef VFS_TTL
    constexpr uint32_t s_callTTLMinValue = 60000;
#else
    constexpr uint32_t s_callTTLMinValue = VFS_TTL;
#endif
    constexpr uint32_t s_callTTLMaxValue = 3600000;

    node::node(const TypeInst &ownInst,
               boost::asio::io_context &ioc,
               const configuration::manager &config)
        : pm::node(ownInst, ioc, config)
    {
        m_db_client_p = std::make_shared<::db::stub::mimic>(m_uvrId);
        db::interfaces::tables::client::bind(m_db_client_p.get());
        db::interfaces::tables::incidentsA::client::bind(m_db_client_p.get());

        auto callTTL = std::chrono::duration_cast<std::chrono::milliseconds>(
            m_config.getValue(config::callRegistryTTL,
                              config::callRegistryTTLDefaultValue));

        if (callTTL < std::chrono::milliseconds(s_callTTLMinValue))
        {
            callTTL = std::chrono::milliseconds(s_callTTLMinValue);
            LOGW("Minumum value for 'Call registry TTL' timer is set: ",
                 std::chrono::duration_cast<std::chrono::seconds>(callTTL).count(), " seconds.");
        }
        else if (callTTL > std::chrono::milliseconds(s_callTTLMaxValue))
        {
            callTTL = std::chrono::milliseconds(s_callTTLMaxValue);
            LOGW("Maximum value for 'Call registry TTL' timer is set: ",
                 std::chrono::duration_cast<std::chrono::seconds>(callTTL).count(), " seconds.");
        }
        m_registry.setCallTTL(timers::time_point(callTTL));

        auto searchTimeout = std::chrono::duration_cast<std::chrono::milliseconds>(
            m_config.getValue(config::callRegistrySearchTimeout,
                              config::callRegistrySearchTimeoutDefaultValue));
        m_registry.setSearchTimeout(timers::time_point(searchTimeout));
    }

    void node::start(void)
    {
        start_node_timers();
        pm::node::start();

        // Create tables
        m_db_client_p->connect();
        db::interfaces::tables::client::create(m_db_client_p->db::interfaces::tables::incidentsA::server::tableName,
                                               m_db_client_p->db::interfaces::tables::incidentsA::server::tableFields,
                                               m_db_client_p->db::interfaces::tables::incidentsA::server::tableSettings);
    }

    void node::start_node_timers()
    {
        // REGISTRY_WATCHDOG
        m_registryWatchdogTimerId =
            m_timersManager->makeTimerAndStart<timers::periodic>("REGISTRY_WATCHDOG",
                                                                 std::chrono::milliseconds(100));

        m_incidentsATimerId =
            m_timersManager->makeTimerAndStart<timers::periodic>(
                "GET_INCIDENTSA",
                m_config.getValue(TIMER_GET_INCIDENTA_REQ,
                                  timers::Duration(std::chrono::hours(1))));
    }

    void node::prepareCounters()
    {
        addCounter(pm::counter::calls::outgoing, pm::counter::indexing::no);
        addCounter(pm::counter::vfs::incidentsA, pm::counter::indexing::no);
        addCounter(pm::counter::vfs::checkPass, pm::counter::indexing::no);
        addCounter(pm::counter::vfs::checkFail, pm::counter::indexing::no);
        addCounter(pm::counter::vfs::checkTimeout, pm::counter::indexing::no);
    }

    void node::handleSpecificPrimitive(const primitive::GenericPtr &prim)
    {
        switch (prim->m_header->id)
        {
            case primitive::OUTGOING_CALL_IND:
            {
                auto ind = std::dynamic_pointer_cast<primitive::OutgoingCallIndication>(prim);
                if (ind)
                {
                    increaseCounter(pm::counter::calls::outgoing,
                                    ::pm::counter::value(1));
                    m_registry.handleOutgoingCall(*ind);
                }
                break;
            }
            case primitive::VERIFY_CALL_IND:
            {
                auto ind = std::dynamic_pointer_cast<primitive::VerifyCallIndication>(prim);
                if (ind)
                {
                    ::primitive::AdditionalCode_t additionalCode;
                    ::primitive::ReleaseCode_t releaseCode;
                    ::primitive::ErrorCode_t errorCode;
                    switch (m_registry.handleVerifyCall(*ind))
                    {
                        case registry::CheckResult::pass:
                        {
                            releaseCode = 0;
                            additionalCode = 0;
                            errorCode = 0;
                            increaseCounter(pm::counter::vfs::checkPass);
                            break;
                        }
                        case registry::CheckResult::fail:
                        {
                            /// TODO need real errorCode/addictionalCode for VerifyCallConfirm
                            /// TODO from m_registry.handle_verify_call
                            releaseCode = 1;
                            additionalCode = 0;
                            errorCode = 0;
                            increaseCounter(pm::counter::vfs::checkFail);

                            doIncident(*ind);
                            break;
                        }
                        case registry::CheckResult::timeout:
                        {
                            releaseCode = 0;
                            additionalCode = 0;
                            errorCode = HTTP_UVR_ERRORCODE_OVERLOAD;
                            increaseCounter(pm::counter::vfs::checkTimeout);
                            break;
                        }
                    }

                    auto cnf = std::make_shared<primitive::VerifyCallConfirm>(ind,
                                                                              releaseCode,
                                                                              additionalCode,
                                                                              errorCode);
                    cnf->m_header->to = TypeInst(workers::type::VFGW, 0);
                    cnf->m_header->ulpKey = ind->m_header->ulpKey;
                    workers::node::send(cnf);
                }
                break;
            }
            case ::primitive::VERIFY_CALL_REQ:
            {
                auto req = std::dynamic_pointer_cast<::primitive::VerifyCallRequest>(prim);
                if (req)
                {
                    Uvr::VerificationVerdict_t rlc;
                    switch (m_registry.handleVerifyCall(*req))
                    {
                        case registry::CheckResult::pass:
                        {
                            rlc = Uvr::VerificationVerdict::success;
                            increaseCounter(pm::counter::vfs::checkPass);
                            break;
                        }
                        case registry::CheckResult::fail:
                        {
                            rlc = Uvr::VerificationVerdict::failure;
                            increaseCounter(pm::counter::vfs::checkFail);
                            doIncident(*req);
                            break;
                        }
                        case registry::CheckResult::timeout:
                        {
                            rlc = Uvr::VerificationVerdict::timeout;
                            increaseCounter(pm::counter::vfs::checkTimeout);
                            /// Need incident_a for the case ???
                            break;
                        }
                    }

                    auto res = std::make_shared<primitive::VerifyCallResponse>(rlc, Uvr::VerificationVerdict::undefined, m_uvrId);
                    res->m_header->to = TypeInst(workers::type::VFC, 0);
                    res->m_header->ulpKey = req->m_header->ulpKey;
                    workers::node::send(res);
                }
                break;
            }
            default:
                pm::node::handleSpecificPrimitive(prim);
        }
    }

    void node::handle_timeout(const ::timers::id &id,
                              const ::timers::time_point &start_time)
    {
        if (id == m_registryWatchdogTimerId)
        {
            m_registry.do_sanity();
        }
        else if (id == m_incidentsATimerId)
        {
            getIncidentsA();
        }
    }

    void node::getIncidentsA()
    {
        db::incidentA::rows incidents;
        db::interfaces::tables::incidentsA::client::get(incidents,
                                                        db::last_get,
                                                        db::latest);
        std::string report_name;
        std::string reportDirPath = m_config.getValue<std::string>(CCNGW_REPORT_PATH + INCIDENTSA_REPORT,
                                                                   INCIDENTSA_REPORT_DEFAULT_DIR)
                                    + "/";

        fs::local localFs(m_ownTypeInst);
        localFs.createDir(reportDirPath);
        if (Csv::File::generate(reportDirPath + Csv::IncidentA::prefix(m_uvrId),
                                Csv::IncidentA::header,
                                incidents,
                                report_name))
        {
            auto rsp = std::make_shared<primitive::GetIncidentsARsp>(report_name);
            rsp->m_header->to = TypeInst(workers::type::CCNGW, 0);
            workers::node::send(rsp);
        }
    }

    void node::doIncident(const primitive::VerifyCallIndication &ind)
    {
        doIncident(ind.callingNumber,
                   ind.calledNumber,
                   ind.originalCalledNumber,
                   ind.uvrO);
    }

    void node::doIncident(const primitive::VerifyCallRequest &req)
    {
        doIncident(req.verifyCallQuery->callingNumber,
                   req.verifyCallQuery->calledNumber,
                   req.verifyCallQuery->originalCalledNumber,
                   m_uvrId);
    }

    void node::doIncident(const Uvr::OctetStringPtr &callingNumber,
                          const Uvr::OctetStringPtr &calledNumber,
                          const Uvr::OctetStringPtr &originalCalledNumber,
                          const Uvr::UvrId_t &idUvrO)
    {
        // Put incident into storage
        db::incidentA::record recIncident = {
            callingNumber,
            TRecQ763Number::toHashedOctetStringPtr(calledNumber),
            TRecQ763Number::toHashedOctetStringPtr(originalCalledNumber),
            timers::now(),
            Uvr::VerdictTrigger::antifraud,
            Uvr::VerificationVerdict::failure,
            idUvrO};
        db::interfaces::tables::incidentsA::client::put(recIncident);
        increaseCounter(pm::counter::vfs::incidentsA,
                        ::pm::counter::value(1));
    }

} // namespace vfs
