#include "body.hpp"
#include "utils.hpp"

// STL includes.
#include <cstring>

namespace sip
{

    std::string Body::toStr(int &errorCode)
    {
        if (m_body.empty())
        {
            errorCode = BADPARAMETER;
            return "";
        }

        std::string ret;

        if (m_contentType.get() != nullptr)
        {
            ret += m_contentType->toStr(errorCode) + CRLF;

            if (errorCode != SUCCESS)
            {
                return "";
            }
        }

        for (auto it = m_headers.begin(); it != m_headers.end(); it++)
        {
            ret += (*it)->toStr(errorCode) + CRLF;

            if (errorCode != SUCCESS)
            {
                return "";
            }
        }

        if ((m_headers.size() > 0) || (m_contentType.get() != nullptr))
        {
            ret += CRLF;
        }

        ret += m_body;

        errorCode = SUCCESS;
        return ret;
    }

    int Body::doParse(std::string &&buf)
    {
        if (buf.empty())
        {
            return BADPARAMETER;
        }

        m_body = std::move(buf);
        return SUCCESS;
    }

    int Body::doParseMime(const char *startOfBody, size_t length)
    {
        const char *endOfBodyHeader;
        const char *startOfBodyHeader;
        int errorCode;

        if (startOfBody == nullptr)
        {
            return BADPARAMETER;
        }

        startOfBodyHeader = startOfBody;

        errorCode = doParseHeader(startOfBodyHeader, &endOfBodyHeader);

        if (errorCode != SUCCESS)
        {
            return errorCode;
        }

        startOfBodyHeader = endOfBodyHeader;

        if (std::strncmp(startOfBodyHeader, CRLF, 2) == 0)
        {
            startOfBodyHeader = startOfBodyHeader + 2;
        }
        else
        {
            if ((std::strncmp(startOfBodyHeader, "\n", 1) == 0) ||
                (std::strncmp(startOfBodyHeader, "\r", 1) == 0))
            {
                startOfBodyHeader = startOfBodyHeader + 1;
            }
            else
            {
                return SYNTAXERROR;
            }
        }

        endOfBodyHeader = startOfBody + length;

        if (endOfBodyHeader - startOfBodyHeader <= 0)
        {
            return SYNTAXERROR;
        }

        m_body = std::string(startOfBodyHeader, endOfBodyHeader - startOfBodyHeader);

        return SUCCESS;
    }

    int Body::doParseHeader(const char *startOfBodyHeader,
                            const char **next_body)
    {
        const char *start_of_line;
        const char *end_of_line;
        const char *colon_index;
        int errorCode;

        *next_body = nullptr;
        start_of_line = startOfBodyHeader;

        for (;;)
        {
            errorCode = utils::findNextCrLf(start_of_line, &end_of_line);

            if (errorCode == BADPARAMETER)
            {
                return SYNTAXERROR;
            }
            else if (errorCode != 0)
            {
                return errorCode;
            }

            colon_index = std::strchr(start_of_line, ':');

            if (colon_index == nullptr)
            {
                return SYNTAXERROR;
            }

            if (colon_index - start_of_line + 1 < 2)
            {
                return SYNTAXERROR;
            }

            std::string hname = utils::doCopyAndClearLws(start_of_line,
                                                         colon_index - start_of_line);

            if ((end_of_line - 2) - colon_index < 2)
            {
                return SYNTAXERROR;
            }

            std::string hvalue = utils::doCopyAndClearLws(colon_index + 1,
                                                          (end_of_line - 2) - colon_index - 1);

            if (strncasecmp(hname.c_str(), "content-type", 12) == 0)
            {
                errorCode = setContentType(hvalue.c_str());
            }
            else
            {
                errorCode = setHeader(hname.c_str(), hvalue.c_str());
            }

            if (errorCode != SUCCESS)
            {
                return errorCode;
            }

            if (strncmp(end_of_line, CRLF, 2) == 0 ||
                strncmp(end_of_line, "\n", 1) == 0 ||
                strncmp(end_of_line, "\r", 1) == 0)
            {
                *next_body = end_of_line;
                return SUCCESS;
            }

            start_of_line = end_of_line;
        }
    }

    int Body::setContentType(const char *hvalue)
    {
        int errorCode;

        if (hvalue == nullptr)
        {
            return BADPARAMETER;
        }

        m_contentType.reset();
        m_contentType = std::make_shared<header::ContentType>();
        errorCode = m_contentType->doParse(hvalue);

        if (errorCode != SUCCESS)
        {
            m_contentType.reset();
            return errorCode;
        }

        return SUCCESS;
    }

    int Body::setHeader(std::string &&hname,
                        std::string &&hvalue)
    {
        if (hname.empty() || hvalue.empty())
        {
            return BADPARAMETER;
        }

        m_headers.push_back(std::make_shared<header::Generic>(std::move(hname), std::move(hvalue)));
        return SUCCESS;
    }

} // namespace sip
