#include "header_via.hpp"
#include "utils.hpp"

// STL includes.
#include <cstring>

namespace sip::header
{

    Via::Via()
        : Abstract(std::string("Via"))
    {
    }

    std::string Via::valueToStr(int &errorCode)
    {
        if (m_host.empty() || m_version.empty() || m_protocol.empty())
        {
            RETURN_STR_FAIL(BADPARAMETER);
        }

        std::string ret = "SIP/" + m_version + "/" + m_protocol;

        if (std::strchr(m_host.c_str(), ':') != nullptr)
        {
            ret += " [" + m_host + "]";
        }
        else
        {
            ret += " " + m_host;
        }

        if (!m_port.empty())
        {
            ret += ":" + m_port;
        }

        for (auto it = m_params.begin(); it != m_params.end(); it++)
        {
            if (it->m_value.empty())
            {
                ret += ";" + it->m_name;
            }
            else
            {
                ret += ";" + it->m_name + "=" + it->m_value;
            }
        }

        if (!m_comment.empty())
        {
            ret += " (" + m_comment + ")";
        }

        RETURN_STR(SUCCESS, ret);
    }

    int Via::doParse(const char *hvalue)
    {
        const char *version;
        const char *protocol;
        const char *host;
        const char *ipv6host;
        const char *port;
        const char *via_params;
        const char *comment;
        int errorCode;

        if (hvalue == nullptr)
        {
            return BADPARAMETER;
        }

        version = std::strchr(hvalue, '/');
        if (version == nullptr)
        {
            return SYNTAXERROR;
        }

        protocol = std::strchr(version + 1, '/');
        if (protocol == nullptr)
        {
            return SYNTAXERROR;
        }

        if (protocol - version < 2)
        {
            return SYNTAXERROR;
        }

        m_version = utils::doCopyAndClearLws(version + 1,
                                             protocol - version - 1);

        host = std::strchr(protocol + 1, ' ');

        if (host == nullptr)
        {
            return SYNTAXERROR;
        }

        if (host == protocol + 1)
        {
            while (0 == std::strncmp(host, " ", 1))
            {
                host++;

                if (std::strlen(host) <= 1)
                {
                    return SYNTAXERROR;
                }
            }

            host = std::strchr(host + 1, ' ');

            if (host == nullptr)
            {
                return SYNTAXERROR;
            }
        }

        if (host - protocol < 2)
        {
            return SYNTAXERROR;
        }

        m_protocol = utils::doCopyAndClearLws(protocol + 1,
                                              host - protocol - 1);

        comment = std::strchr(host, '(');

        if (comment != nullptr)
        {
            const char *end_comment;

            end_comment = std::strchr(host, ')');

            if (end_comment == nullptr)
            {
                return SYNTAXERROR;
            }

            if (end_comment - comment < 2)
            {
                return SYNTAXERROR;
            }

            m_comment = std::string(comment + 1, end_comment - comment - 1);
            comment--;
        }
        else
        {
            comment = host + std::strlen(host);
        }

        via_params = std::strchr(host, ';');

        if ((via_params != nullptr) && (via_params < comment))
        {
            if (comment - via_params + 1 < 2)
            {
                return SYNTAXERROR;
            }

            std::string tmp(via_params, comment - via_params);
            errorCode = m_params.doParse(tmp.c_str());

            if (errorCode != SUCCESS)
            {
                return errorCode;
            }
        }

        if (via_params == nullptr)
        {
            via_params = comment;
        }

        ipv6host = std::strchr(host, '[');

        if (ipv6host != nullptr && ipv6host < via_params)
        {
            port = std::strchr(ipv6host, ']');

            if (port == nullptr || port > via_params)
            {
                return SYNTAXERROR;
            }

            if (port - ipv6host < 2)
            {
                return SYNTAXERROR;
            }

            m_host = utils::doCopyAndClearLws(ipv6host + 1,
                                              port - ipv6host - 1);

            port = std::strchr(port, ':');
        }
        else
        {
            port = std::strchr(host, ':');
            ipv6host = nullptr;
        }

        if ((port != nullptr) && (port < via_params))
        {
            if (via_params - port < 2)
            {
                return SYNTAXERROR;
            }

            m_port = utils::doCopyAndClearLws(port + 1,
                                              via_params - port - 1);

            if (utils::atoi(m_port) < 0)
            {
                return SYNTAXERROR;
            }
        }
        else
        {
            port = via_params;
        }

        if (ipv6host != nullptr)
        {
            return SUCCESS;
        }

        if (port - host < 2)
        {
            return SYNTAXERROR;
        }

        m_host = utils::doCopyAndClearLws(host + 1,
                                          port - host - 1);

        return SUCCESS;
    }

} // namespace sip::header
