#include "message.hpp"
#include "utils.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"

#include <regex>

#define MIME_MAX_BOUNDARY_LEN 70

namespace sip
{

    int Message::parseRequestStartLine(const char *buf, const char **headers)
    {
        const char *p1;
        const char *p2;

        m_method = "";
        m_statusCode = 0;
        m_reasonPhrase = "";

        *headers = buf;

        // The first token is the method name
        p2 = std::strchr(buf, ' ');

        if (p2 == nullptr)
        {
            return SYNTAXERROR;
        }

        if (*(p2 + 1) == '\0' || *(p2 + 2) == '\0')
        {
            return SYNTAXERROR;
        }

        if (p2 - buf == 0)
        {
            LOGERROR("No space allowed here");
            return SYNTAXERROR;
        }

        m_method = std::string(buf, p2 - buf);

        // The second token is a sip-url or a uri:
        p1 = std::strchr(p2 + 2, ' ');

        if (p1 == nullptr)
        {
            LOGERROR("Uncompliant request-uri");
            return SYNTAXERROR;
        }

        if (p1 - p2 < 2)
        {
            return SYNTAXERROR;
        }

        std::string requestUri(p2 + 1, (p1 - p2 - 1));

        m_uri = std::make_shared<Uri>();
        if (m_uri->doParse(requestUri.c_str()))
        {
            m_uri.reset();
            return SYNTAXERROR;
        }

        // find the the version and the beginning of headers
        {
            const char *hp = p1 + 1; ///< skip space

            if (hp[0] == '\0' || hp[1] == '\0'
                || hp[2] == '\0' || hp[3] == '\0'
                || hp[4] == '\0' || hp[5] == '\0' || hp[6] == '\0')
            {
                LOGERROR("Uncomplete request line");
                return SYNTAXERROR;
            }

            if (((hp[0] != 'S') && (hp[0] != 's'))
                || ((hp[1] != 'I') && (hp[1] != 'i'))
                || ((hp[2] != 'P') && (hp[2] != 'p')) || (hp[3] != '/'))
            {
                LOGERROR("No crlf found/No SIP/2.0 found");
                return SYNTAXERROR;
            }

            hp += 4; // "SIP/"

            while ((hp[0] != g_cr) && (hp[0] != g_lf))
            {
                if (hp[0])
                {
                    if (((hp[0] >= '0') && (hp[0] <= '9'))
                        || (hp[0] == '.'))
                    {
                        hp++;
                    }
                    else
                    {
                        LOGERROR("incorrect sip version string");
                        return SYNTAXERROR;
                    }
                }
                else
                {
                    LOGERROR("No crlf found");
                    return SYNTAXERROR;
                }
            }

            if (hp - p1 < 2)
            {
                return SYNTAXERROR;
            }

            m_version = std::string(p1 + 1, (hp - p1 - 1));

            if (m_version != "SIP/2.0")
            {
                LOGWARN("Wrong version number");
            }

            hp++;

            if (hp[0] && (g_cr == hp[-1]) && (g_lf == hp[0]))
            {
                hp++;
            }

            (*headers) = hp;
        }
        return SUCCESS;
    }

    int Message::parseResponseStartLine(const char *buf, const char **headers)
    {
        const char *statuscode;
        const char *reasonphrase;

        m_method = "";

        *headers = buf;

        statuscode = std::strchr(buf, ' '); // search for first SPACE

        if (statuscode == nullptr)
        {
            return SYNTAXERROR;
        }

        if (statuscode - buf < 7) // must be at least "SIP" "/"  1*DIGIT "." 1*DIGIT
        {
            return SYNTAXERROR;
        }

        m_version = std::string(buf, statuscode - buf);

        reasonphrase = std::strchr(statuscode + 1, ' ');

        if (reasonphrase == nullptr)
        {
            return SYNTAXERROR;
        }

        m_statusCode = std::atoi(statuscode + 1);

        if (m_statusCode == 0)
        {
            return SYNTAXERROR;
        }

        {
            const char *hp = reasonphrase;

            while ((hp[0] != g_cr) && (hp[0] != g_lf))
            {
                if (hp[0])
                {
                    hp++;
                }
                else
                {
                    LOGERROR("No crlf found");
                    return SYNTAXERROR;
                }
            }

            m_reasonPhrase = std::string(reasonphrase + 1, hp - reasonphrase - 1);

            hp++;

            if (hp[0] && (g_cr == hp[-1]) && (g_lf == hp[0]))
            {
                hp++;
            }

            (*headers) = hp;
        }
        return SUCCESS;
    }

    int Message::parseStartLine(const char *buf, const char **headers)
    {
        if (0 == strncmp(buf, (const char *)"SIP/", 4))
        {
            return parseResponseStartLine(buf, headers);
        }
        else
        {
            return parseRequestStartLine(buf, headers);
        }
    }

    int Message::parseHeaders(const char *startOfHeader, const char **body)
    {
        const char *colon_index;
        std::string hname;
        std::string hvalue;
        const char *end_of_header;
        int i;

        for (;;)
        {
            if (startOfHeader[0] == '\0')
            {
                LOGINFO("SIP message does not end with CRLFCRLF");
                return SUCCESS;
            }

            i = utils::findNextCrLf(startOfHeader, &end_of_header);

            if (i != 0 && (i != -2))
            {
                LOGERROR("End of header Not found");
                return i;
            }

            if ((startOfHeader[0] == g_cr) || (startOfHeader[0] == g_lf))
            {
                *body = startOfHeader;
                return SUCCESS;
            }

            colon_index = std::strchr(startOfHeader, ':');

            if (colon_index == nullptr)
            {
                LOGERROR("End of header Not found");
                return SYNTAXERROR;
            }

            if (colon_index - startOfHeader + 1 < 2)
                return SYNTAXERROR;

            if (end_of_header <= colon_index)
            {
                LOGERROR("Malformed message");
                return SYNTAXERROR;
            }

            hname = std::string(startOfHeader, colon_index - startOfHeader);

            {
                const char *end;

                if ((end_of_header[-2] == g_cr) || (end_of_header[-2] == g_lf))
                {
                    end = end_of_header - 2;
                }
                else
                {
                    end = end_of_header - 1;
                }

                if (end - (colon_index + 1) < 1)
                {
                    hvalue = "";
                }
                else
                {
                    hvalue = utils::doCopyAndClearLws(colon_index + 1,
                                                      end - (colon_index + 1));
                }
            }

            i = setMultipleHeader(std::move(hname), std::move(hvalue));

            if (i != 0)
            {
                LOGERROR("End of header \"", hname, "\" with \"", hvalue, "\" Not found");
                return SYNTAXERROR;
            }

            startOfHeader = end_of_header;
        }
    }

    int Message::setHeader(std::string &&hname,
                           std::string &hnameLower,
                           std::string &&hvalue)
    {
        if (hname.empty())
        {
            return SYNTAXERROR;
        }

        sip::header::Id id = sip::header::getHeaderId(hnameLower);
        std::shared_ptr<sip::header::Generic> hdr;
        switch (id)
        {
            case header::Id::from:
            {
                return setFrom(hvalue.c_str());
            }
            case header::Id::to:
            {
                return setTo(hvalue.c_str());
            }
            case header::Id::contentLength:
            {
                return setContentLength(std::move(hvalue));
            }
            case header::Id::contentType:
            {
                return setContentType(hvalue.c_str());
            }
            case header::Id::via:
            {
                return setVia(hvalue.c_str());
            }
            case header::Id::cSeq:
            {
                return setCSeq(hvalue.c_str());
            }
            case header::Id::contact:
            {
                return setContact(hvalue.c_str());
            }
            case header::Id::callId:
            {
                return setCallId(hvalue.c_str());
            }
            case header::Id::diversion:
            {
                return setDiversion(hvalue.c_str());
            }
            case header::Id::allow:
            {
                return setAllow(std::move(hvalue));
            }
            default:
                m_property = 2;
                m_headers.push_back(std::make_shared<header::Generic>(std::move(hname),
                                                                      std::move(hvalue)));
                break;
        }

        return SUCCESS;
    }

    int Message::setMultipleHeader(std::string &&hname,
                                   std::string &&hvalue)
    {
        int i;
        char *ptr;         ///< current location of the search
        const char *comma; ///< This is the separator we are elooking for
        const char *beg;   ///< beg of a header
        const char *end;   ///< end of a header

        std::string hnameLower = hname;
        sip::utils::toLower(hnameLower);

        if (hvalue.empty())
        {
            return setHeader(std::move(hname), hnameLower,
                             std::move(hvalue));
        }

        ptr = &(hvalue[0]);
        comma = std::strchr(ptr, ',');

        if (comma == nullptr || !header::isCommaSeparated(hnameLower))
        {
            return setHeader(std::move(hname), hnameLower,
                             std::move(hvalue));
        }

        beg = ptr;
        bool inquotes = false;
        bool inuri = false;
        std::string _hname(hname);

        // Seach for a comma that is not within quotes or a URI
        for (;; ptr++)
        {
            switch (*ptr)
            {
                case '"':
                {
                    /// Check that the '"' is not escaped
                    const char *p;
                    for (i = 0, p = ptr - 1; p >= beg && *p == '\\'; p--, i++)
                        ;

                    if (i % 2 == 0)
                        inquotes = !inquotes; // the '"' was not escaped

                    break;
                }
                case '<':
                {
                    if (!inquotes && !inuri
                        && ((strncasecmp(ptr + 1, "sip:", 4) == 0
                             || strncasecmp(ptr + 1, "sips:", 5) == 0
                             || strncasecmp(ptr + 1, "http:", 5) == 0
                             || strncasecmp(ptr + 1, "https:", 6) == 0
                             || strncasecmp(ptr + 1, "tel:", 4) == 0)
                            && std::strchr(ptr, '>')))
                    {
                        inuri = true;
                    }

                    break;
                }
                case '>':
                {
                    if (!inquotes && inuri)
                    {
                        inuri = false;
                    }

                    break;
                }
                case '\0':
                    inquotes = false;
                    inuri = false;
                    // Fall through
                case ',':
                {
                    if (!inquotes && !inuri)
                    {
                        if (beg[0] == '\0')
                        {
                            return SUCCESS; // empty header
                        }

                        end = ptr;

                        if ((end - beg + 1) < 2)
                        {
                            beg = end + 1;
                            break; // skip empty header
                        }

                        std::string __hname(_hname);
                        i = setHeader(std::move(__hname), hnameLower,
                                      std::string(beg, end - beg));
                        if (i != 0)
                        {
                            return i;
                        }

                        beg = end + 1;
                    }

                    if (*ptr == '\0')
                    {
                        return SUCCESS;
                    }
                    break;
                }
                default:
                    break;
            }
        }
    }

    int Message::doParse(const char *buf, size_t length)
    {
        int i;
        char *nextHeaderIndex;
        char *tmp;

        std::string bufStr(buf, length);
        tmp = &(bufStr.front());

        /// skip initial \r\n
        tmp += std::strspn(tmp, "\r\n");
        utils::replaceAllLws(tmp);

        /// parse request or status line
        i = parseStartLine(tmp, (const char **)&nextHeaderIndex);

        if (i != 0)
        {
            LOGERROR(" Failed parse start line of message.");
            return i;
        }

        tmp = nextHeaderIndex;

        i = parseHeaders(tmp, (const char **)&nextHeaderIndex);
        if (i != 0)
        {
            LOGERROR(" Failed parse headers");
            return i;
        }

        setContentLength(std::string("0"));

        return SUCCESS;
    }

    int Message::setContentLength(std::string &&hvalue)
    {
        int i;

        if (hvalue.empty())
        {
            return SUCCESS;
        }

        if (m_contentLength.get() != nullptr)
        {
            return SYNTAXERROR;
        }

        m_property = 2;
        m_contentLength = std::make_shared<header::ContentLength>();
        i = m_contentLength->doParse(std::move(hvalue));

        if (i != SUCCESS)
        {
            m_contentLength.reset();
            return i;
        }

        return SUCCESS;
    }

    int Message::setVia(const char *hvalue)
    {

        if (hvalue == nullptr || hvalue[0] == '\0')
        {
            return SUCCESS;
        }

        auto via = std::make_shared<header::Via>();
        int errorCode = via->doParse(hvalue);

        if (errorCode != SUCCESS)
        {
            return errorCode;
        }

        m_property = 2;
        m_via.push_back(via);
        return SUCCESS;
    }

    int Message::setAllow(std::string &&hvalue)
    {
        if (hvalue.empty())
        {
            return SUCCESS;
        }

        auto allow = std::make_shared<header::Allow>();
        int errorCode = allow->doParse(std::move(hvalue));

        if (errorCode != SUCCESS)
        {
            return errorCode;
        }

        m_property = 2;
        m_allow.push_back(allow);
        return SUCCESS;
    }

    int Message::setDiversion(const char *hvalue)
    {

        if (hvalue == nullptr || hvalue[0] == '\0')
        {
            return SUCCESS;
        }

        auto diversion = std::make_shared<header::Diversion>();
        int errorCode = diversion->doParse(hvalue);

        if (errorCode != SUCCESS)
        {
            return errorCode;
        }

        m_property = 2;
        m_diversion.push_back(diversion);
        return SUCCESS;
    }

    int Message::setCSeq(const char *hvalue)
    {
        if (hvalue == nullptr || hvalue[0] == '\0')
        {
            return SUCCESS;
        }

        auto cSeq = std::make_shared<header::CSeq>();
        int errorCode = cSeq->doParse(hvalue);

        if (errorCode != SUCCESS)
        {
            return errorCode;
        }

        m_property = 2;
        m_cSeq = cSeq;
        return SUCCESS;
    }

    int Message::setCallId(const char *hvalue)
    {
        if (hvalue == nullptr || hvalue[0] == '\0')
        {
            return SUCCESS;
        }

        auto callId = std::make_shared<header::CallId>();
        int errorCode = callId->doParse(hvalue);

        if (errorCode != SUCCESS)
        {
            return errorCode;
        }

        m_property = 2;
        m_callId = callId;
        return SUCCESS;
    }

    int Message::setFrom(const char *hvalue)
    {
        if (hvalue == nullptr || hvalue[0] == '\0')
        {
            return SUCCESS;
        }

        if (m_from.get() != nullptr)
        {
            return SYNTAXERROR;
        }

        m_property = 2;
        auto from = std::make_shared<header::From>();
        int errorCode = from->doParse(hvalue);

        if (errorCode != SUCCESS)
        {
            m_from.reset();
            return errorCode;
        }

        m_from = from;

        return SUCCESS;
    }

    int Message::setTo(const char *hvalue)
    {
        if (hvalue == nullptr || hvalue[0] == '\0')
        {
            return SUCCESS;
        }

        if (m_to.get() != nullptr)
        {
            return SYNTAXERROR;
        }

        m_property = 2;
        auto to = std::make_shared<header::To>();
        int errorCode = to->doParse(hvalue);

        if (errorCode != SUCCESS)
        {
            m_to.reset();
            return errorCode;
        }

        m_to = to;

        return SUCCESS;
    }

    int Message::setContact(const char *hvalue)
    {
        if (hvalue == nullptr || hvalue[0] == '\0')
        {
            return SUCCESS;
        }

        auto contact = std::make_shared<header::Contact>();
        int errorCode = contact->doParse(hvalue);

        if (errorCode != SUCCESS)
        {
            return errorCode;
        }

        m_property = 2;
        m_contact.push_back(contact);
        return SUCCESS;
    }

    int Message::setContentType(const char *hvalue)
    {
        if (m_contentType.get() != nullptr)
        {
            return BADPARAMETER;
        }

        if (hvalue == nullptr || hvalue[0] == '\0')
        {
            return SUCCESS;
        }

        m_property = 2;
        auto contentType = std::make_shared<header::ContentType>();
        int errorCode = contentType->doParse(hvalue);

        if (errorCode != 0)
        {
            m_contentType.reset();
            return errorCode;
        }

        m_contentType = contentType;

        return SUCCESS;
    }

    int Message::setStatusCode(int statusCode)
    {
        m_statusCode = statusCode;
        m_property = 2;
        return SUCCESS;
    }
    int Message::setReasonPhrase(const std::string &reasonPhrase)
    {
        m_reasonPhrase = reasonPhrase;
        m_property = 2;
        return SUCCESS;
    }

    int Message::setMethod(const std::string &method)
    {
        m_method = method;
        m_property = 2;
        return SUCCESS;
    }

    std::string Message::defaultProtocolVersion{"SIP/2.0"};

    std::string Message::startLineToStrReq(int &errorCode) const
    {
        if ((m_uri.get() == nullptr) || (m_method.empty()))
        {
            errorCode = BADPARAMETER;
            return "";
        }

        std::string ret = m_uri->toStr(errorCode);

        if (errorCode != SUCCESS)
        {
            return "";
        }

        ret = m_method + ' ' + ret + ' ';

        if (m_version.empty())
        {
            ret += defaultProtocolVersion;
        }
        else
        {
            ret += m_version;
        }

        errorCode = SUCCESS;
        return ret;
    }

    std::string Message::startLineToStrResp(int &errorCode) const
    {
        if ((m_reasonPhrase.empty())
            || (m_statusCode < 100) || (m_statusCode > 699))
        {
            errorCode = BADPARAMETER;
            return "";
        }

        std::string ret;

        if (m_version.empty())
        {
            ret = defaultProtocolVersion;
        }
        else
        {
            ret = m_version;
        }

        ret += ' ' + std::to_string(m_statusCode) + ' ' + m_reasonPhrase;

        errorCode = SUCCESS;
        return ret;
    }

    std::string Message::startLineToStr(int &errorCode) const
    {
        if (!m_method.empty())
        {
            return startLineToStrReq(errorCode);
        }

        if (m_statusCode != 0)
        {
            return startLineToStrResp(errorCode);
        }

        LOGERROR(" ERROR method has no value or status code is 0!");
        errorCode = BADPARAMETER;
        return "";
    }

    std::string Message::encode(int &errorCode)
    {
        int sipfrag = 0;

        if (1 == m_property)
        {
            errorCode = SUCCESS;
            return m_message;
        }
        else
        {
            m_message.clear();
        }

        std::string ret = startLineToStr(errorCode);

        if (errorCode != SUCCESS)
        {
            if (!sipfrag)
            {
                return "";
            }
        }
        else
        {
            ret += CRLF;
        }

        {

            std::list<header::AbstractContainer *> headers;
            headers.push_back(&m_via);
            headers.push_back(&m_recordRoute);
            headers.push_back(&m_route);
            headers.push_back(&m_from);
            headers.push_back(&m_to);
            headers.push_back(&m_callId);
            headers.push_back(&m_diversion);
            headers.push_back(&m_cSeq);
            headers.push_back(&m_contact);
            headers.push_back(&m_authorization);
            headers.push_back(&m_wwwwAuthenticate);
            headers.push_back(&m_proxyAuthenticate);
            headers.push_back(&m_proxyAuthorization);
            headers.push_back(&m_callInfo);
            headers.push_back(&m_contentType);
            headers.push_back(&m_mimeVersion);
            headers.push_back(&m_allow);
            headers.push_back(&m_contentEncoding);
            headers.push_back(&m_alertInfo);
            headers.push_back(&m_errorInfo);
            headers.push_back(&m_accept);
            headers.push_back(&m_acceptEncoding);
            headers.push_back(&m_acceptLanguage);
            headers.push_back(&m_authenticationInfo);
            headers.push_back(&m_proxyAuthentication);

            for (auto it = headers.begin(); it != headers.end(); it++)
            {
                ret += (*it)->toStr(errorCode);

                if (errorCode != SUCCESS)
                {
                    return "";
                }
            }
        }

        for (auto it = m_headers.begin(); it != m_headers.end(); it++)
        {
            ret += (*it)->toStr(errorCode);

            if (errorCode != 0)
            {
                return "";
            }
        }

        if (sipfrag && m_bodies.empty())
        {
            ret += CRLF;

            m_property = 1;
            m_message = ret;

            errorCode = SUCCESS;
            return ret;
        }

        ret += "Content-Length: ";

        if (m_bodies.empty())
        {
            ret += "0" END_HEADERS_STR;

            m_property = 1;
            m_message = ret;

            errorCode = SUCCESS;
            return ret;
        }

        std::string ret1;
        std::string boundary;

        if (m_contentType.get() != nullptr)
        {
            auto contentType = (header::ContentType *)m_contentType.get();
            if (!(contentType->m_type.empty()) && !strcasecmp(contentType->m_type.c_str(), "multipart"))
            {
                GenericParam *ct_param = contentType->m_genParams.match("boundary", errorCode);

                if ((errorCode == SUCCESS) && ct_param && !(ct_param->m_value.empty()))
                {
                    size_t len = ct_param->m_value.length();

                    if ((len > MIME_MAX_BOUNDARY_LEN)
                        || (len == 1 && ct_param->m_value[0] == '"')
                        || (len == 2 && ct_param->m_value[0] == '"' && ct_param->m_value[1] == '"'))
                    {
                        errorCode = SYNTAXERROR;
                        return "";
                    }

                    boundary = CRLF "--";

                    if (ct_param->m_value[0] == '"' && ct_param->m_value[len - 1] == '"')
                    {
                        boundary += std::string(ct_param->m_value.c_str(), len - 2);
                    }
                    else
                    {
                        boundary += ct_param->m_value;
                    }
                }
            }
        }

        for (auto it = m_bodies.begin(); it != m_bodies.end(); it++)
        {
            if (!(boundary.empty()))
            {
                ret1 += boundary + CRLF;
            }

            ret1 += (*it)->toStr(errorCode);

            if (errorCode != SUCCESS)
            {

                return "";
            }
        }

        if (!boundary.empty())
        {
            ret1 += boundary + "--" CRLF;
        }

        ret += std::to_string(ret1.length()) + END_HEADERS_STR + ret1;

        m_property = 1;
        m_message = ret;

        errorCode = SUCCESS;
        return ret;
    }

    Contact Message::getTo(const std::string &prefixSeparator)
    {
        auto to = (::sip::header::To *)m_to.get();

        if (to == nullptr)
        {
            LOGW("Malformed SIP INVITE received: To is missing");
            return Contact();
        }

        Contact ret;
        ret.number = to->m_uri->m_username;
        ret.originalNumber = getOriginalNumber();
        ret.ip = to->m_uri->m_host;
        ret.port = to->m_uri->m_port;

        Uvr::OptionalOperatorId operIdSrc;
        support::parseUriNumber(ret.number, operIdSrc, prefixSeparator);

        return ret;
    }
    Contact Message::getFrom(const std::string &prefixSeparator)
    {
        auto from = (::sip::header::To *)m_from.get();

        if (from == nullptr)
        {
            LOGW("Malformed SIP INVITE received: From is missing");
            return Contact();
        }

        Contact ret;
        ret.number = from->m_uri->m_username;
        ret.name = from->m_displayName;
        ret.ip = from->m_uri->m_host;
        ret.port = from->m_uri->m_port;

        Uvr::OptionalOperatorId operIdSrc;
        support::parseUriNumber(ret.number, operIdSrc, prefixSeparator);

        return ret;
    }

    int Message::getCSeq()
    {
        auto cSeq = (::sip::header::CSeq *)m_cSeq.get();
        if (cSeq == nullptr)
        {
            LOGW("Malformed SIP INVITE received: CSeq is missing");
            return 0;
        }
        return cSeq->getValue();
    }

    std::string Message::getCallId()
    {
        auto callId = (::sip::header::CallId *)m_callId.get();
        if (callId == nullptr)
        {
            LOGW("Malformed SIP INVITE received: CallId is missing");
            return 0;
        }
        return callId->getValue();
    }

    std::string Message::getOriginalNumber()
    {
        std::string originalCalledNumberStr;
        if (!m_diversion.empty())
        {
            auto diversion = (::sip::header::Diversion *)(m_diversion.back()).get();

            if (!diversion->m_value.empty())
            {
                std::regex numberRegex("(?:^|.*\\D)(7\\d{10})(?:\\D.*|$)");
                std::smatch numberMatches;
                if (!std::regex_match(diversion->m_value, numberMatches, numberRegex))
                {
                    LOGT("Diversion header = \"", diversion->m_value,
                         "\" does not contain a "
                         "10-digit number. Original Called number will be empty.");
                }
                else
                {
                    originalCalledNumberStr = numberMatches[1].str();
                    LOGT("Original Called number = ", originalCalledNumberStr);
                }
            }
        }
        return originalCalledNumberStr;
    }

    namespace response
    {
        std::ostream &operator<<(std::ostream &os, const Reason &r)
        {
            std::stringbuf trace;
            std::ostream logger(&trace);
            logger << r.m_code << "(" << r.m_phrase << ")";
            os << trace.str();
            return os;
        }

        bool Reason::operator==(const Reason &other) const
        {
            return (m_code == other.m_code)
                   && (m_phrase == other.m_phrase);
        }

    } // namespace response

} // namespace sip
