#pragma once

#include "body.hpp"
#include "headers.hpp"
#include "header_content_length.hpp"
#include "header_content_type.hpp"
#include "header_from.hpp"
#include "header_via.hpp"
#include "header_cseq.hpp"
#include "header_callid.hpp"
#include "header_diversion.hpp"
#include "header_contact.hpp"
#include "uri.hpp"
#include "header_allow.hpp"

// Boost includes.
#include <boost/asio.hpp>

// STL includes.
#include <list>
#include <memory>
#include <string>

#define END_HEADERS_STR CRLF CRLF

namespace sip
{

    struct Contact
    {
        std::string name;
        std::string prefix;
        std::string number;
        std::string originalNumber;
        std::string ip;
        std::string port;
    };

    class Message
    {
        static std::string defaultProtocolVersion;

    public:
        Message() = default;

        std::string encode(int &errorCode);
        int doParse(const char *tmp, size_t length);

        int handleMessages(boost::asio::streambuf &buffer,
                           size_t &restBodyLength);

        Contact getTo(const std::string &prefixSeparator);
        Contact getFrom(const std::string &prefixSeparator);
        int getCSeq();
        std::string getCallId();

        std::string getOriginalNumber();

        int setStatusCode(int);
        int setReasonPhrase(const std::string &);
        int setMethod(const std::string &);

        int parseHeaders(const char *startOfHeader, const char **body);

    private:
        std::string startLineToStr(int &errorCode) const;
        std::string startLineToStrReq(int &errorCode) const;
        std::string startLineToStrResp(int &errorCode) const;

        int parseStartLine(const char *buf, const char **headers);
        int parseResponseStartLine(const char *buf, const char **headers);
        int parseRequestStartLine(const char *buf, const char **headers);

        int setMultipleHeader(std::string &&hname,
                              std::string &&hvalue);
        int setHeader(std::string &&hname,
                      std::string &hnameLower,
                      std::string &&hvalue);
        int setContentLength(std::string &&hvalue);
        int setVia(const char *hvalue);
        int setAllow(std::string &&hvalue);
        int setCSeq(const char *hvalue);
        int setCallId(const char *hvalue);
        int setFrom(const char *hvalue);
        int setTo(const char *hvalue);
        int setContact(const char *hvalue);
        int setContentType(const char *hvalue);
        int setDiversion(const char *hvalue);

    public:
        std::string m_version;
        UriPtr m_uri;
        std::string m_method;

        int m_statusCode{0};
        std::string m_reasonPhrase;

        header::MultipleLineHeader m_via;
        header::MultipleLineHeader m_diversion;
        header::MultipleLineHeader m_recordRoute;
        header::MultipleLineHeader m_route;
        header::SingleLineHeader m_from;
        header::SingleLineHeader m_to;
        header::SingleLineHeader m_callId;
        header::SingleLineHeader m_cSeq;
        header::MultipleLineHeader m_contact;
        header::MultipleLineHeader m_authorization;
        header::MultipleLineHeader m_wwwwAuthenticate;
        header::MultipleLineHeader m_proxyAuthenticate;
        header::MultipleLineHeader m_proxyAuthorization;
        header::MultipleLineHeader m_callInfo;
        header::SingleLineHeader m_contentType;
        header::SingleLineHeader m_mimeVersion;
        header::MultipleLineHeader m_allow;
        header::MultipleLineHeader m_contentEncoding;
        header::MultipleLineHeader m_alertInfo;
        header::MultipleLineHeader m_errorInfo;
        header::MultipleLineHeader m_accept;
        header::MultipleLineHeader m_acceptEncoding;
        header::MultipleLineHeader m_acceptLanguage;
        header::MultipleLineHeader m_authenticationInfo;
        header::MultipleLineHeader m_proxyAuthentication;

        header::ContentLengthPtr m_contentLength;

        header::GenericPtrList m_headers;
        BodyList m_bodies;

        int m_property{0};
        std::string m_message;
    };

    using MessagePtr = std::shared_ptr<Message>;

    namespace response
    {

        struct Reason
        {
            int m_code;
            std::string m_phrase;

            bool operator==(const Reason &other) const;
        };

        std::ostream &operator<<(std::ostream &os, const Reason &r);

    } // namespace response

} // namespace sip
