#include "transport_tcp.hpp"

#include <iostream>
#include <boost/bind/bind.hpp>

namespace sip
{
    namespace tcp
    {

        namespace transport
        {

            Manager::Manager(boost::asio::io_context &ioc,
                             const Log::Tag &logTag) : m_ioc(ioc),
                                                       m_logTag{logTag},
                                                       m_timerMngr(std::make_shared<timers::manager>(m_ioc))
            {
                m_timerMngr->::timers::interfaces::notifier::client::bind(this);
            }

            Manager::~Manager()
            {
                stop();
            }

            void Manager::start(timers::duration period)
            {
                startMonitorTimer(period);
                startAccept();
            }

            void Manager::stop()
            {
                std::lock_guard<std::recursive_mutex> guard(m_guard);

                trunk::interfaces::controller::server::stop_server();

                m_timerMngr->delAndStopAll();
                stopAccept();

                for (auto const &connIter : m_connections)
                {
                    connIter.second->disconnect();
                }
                m_connections.clear();
            }

            trunk::ConnectionPtr Manager::getConnection(const trunk::Id &id) const
            {
                Guard guard(m_guard);
                auto connIter = m_connections.find(id);
                if (connIter == m_connections.end())
                    return trunk::ConnectionPtr();
                return connIter->second;
            }

            trunk::Id Manager::addConnection(const std::string &address,
                                             const std::string &port)
            {
                Guard guard(m_guard);
                auto connPtr =
                    std::make_shared<trunk::connection>(m_ioc, trunk::type::OUTGOING,
                                                        address, port,
                                                        m_connIdCntr,
                                                        m_logTag);
                m_connIdCntr++;
                connPtr->trunk::interfaces::controller::client1::bind(shared_from_this());
                auto result = m_connections.emplace(connPtr->getId(), connPtr);
                if (!result.second)
                {
                    LOGERROR(m_logTag,
                             " Duplicate connection: ",
                             "old ", *(result.first->second), ", ",
                             "new ", *connPtr);
                }
                LOGINFO(m_logTag, " added connection to ", address, ":", port);
                return connPtr->getId();
            }

            void Manager::addListener(const network::address &ipAddress,
                                      const std::string &port,
                                      const sip::trunk::Role &role)
            {
                boost::system::error_code err;

                AcceptorPtr acceptor = std::make_shared<Acceptor>(m_ioc, role);
                m_acceptors.push_back(acceptor);

                boost::asio::ip::tcp::endpoint endpoint(ipAddress,
                                                        static_cast<boost::asio::ip::port_type>(
                                                            std::stoi(port)));

                acceptor->open(endpoint.protocol(), err);
                if (err)
                {
                    LOGERROR(m_logTag,
                             " Cannot open socket[", ipAddress, ":", port, "]: ",
                             err.message());
                    return;
                }

                acceptor->set_option(Acceptor::reuse_address(true), err);
                if (err)
                {
                    LOGERROR(m_logTag,
                             " Cannot set REUSE_ADDRESS option for socket[", ipAddress, ":", port, "]: ",
                             err.message());
                    return;
                }

                acceptor->bind(endpoint, err);
                if (err)
                {
                    LOGERROR(m_logTag,
                             " Cannot bind to socket[", ipAddress, ":", port, "]: ",
                             err.message());
                    return;
                }

                acceptor->listen(boost::asio::socket_base::max_listen_connections, err);
                if (err)
                {
                    LOGERROR(m_logTag,
                             " Cannot listen to socket[", ipAddress, ":", port, "]: ",
                             err.message());
                    return;
                }

                LOGI(m_logTag, "listener add at ", ipAddress, ":", port,
                     " role=", role2str(role));
            }

            void Manager::send(const trunk::Id &id, Message &msg) const
            {
                auto connPtr = getConnection(id);
                if (connPtr.get() == nullptr)
                {
#if LOG_MACRO_LEVEL >= LOG_LEVEL_INFO
                    int errorCode;
                    LOGINFO(m_logTag,
                            " Cannot find connection with id=", id,
                            " to send ", msg.encode(errorCode), " message");
#endif
                    return;
                }
                int errorCode;
                std::string msgEncoded = msg.encode(errorCode);
                connPtr->send(msgEncoded);
            }

            void Manager::controllBrokenConnection(const trunk::Id &id)
            {
                Guard guard(m_guard);
                auto connIter = m_connections.find(id);
                if (connIter == m_connections.end())
                {
                    LOGINFO(m_logTag,
                            " Cannot find broken connection with ", id);
                    return;
                }
                if (connIter->second->type() == trunk::type::INCOMING)
                {
                    m_connections.erase(connIter);
                    return;
                }

                startConnect(connIter->second);
            }

            void Manager::controllConnectionClosed(const trunk::Id &id)
            {
                Guard guard(m_guard);
                auto connIter = m_connections.find(id);
                if (connIter == m_connections.end())
                {
                    LOGERROR(m_logTag,
                             " Cannot find closed connection with ", id);
                    return;
                }
                if (connIter->second->type() == trunk::type::INCOMING)
                {
                    m_connections.erase(connIter);
                }
            }

            void Manager::startMonitorTimer(timers::duration period)
            {
                std::stringbuf timerName;
                std::ostream timerNameStream(&timerName);
                timerNameStream << m_logTag << " CONNECTIONS_MONITOR";

                m_watchdogTimerId =
                    m_timerMngr->makeTimerAndStart<timers::periodic>(timerName.str(),
                                                                     period);
                LOGDEBUG(m_logTag, " connections watchdog started");
            }

            void Manager::startAllConnections()
            {
                Guard guard(m_guard);
                for (auto const &connIter : m_connections)
                {
                    if (connIter.second->state() == trunk::state::CONNECTION_INIT)
                    {
                        startConnect(connIter.second);
                    }
                }
            }

            void Manager::startConnect(trunk::ConnectionPtr connPtr)
            {
                boost::system::error_code err;
                boost::asio::ip::tcp::resolver resolver(m_ioc);
                auto const results = resolver.resolve(connPtr->peerAddress(), connPtr->peerPort(), err);
                if (err)
                {
                    LOGERROR(m_logTag,
                             " Cannot resolve address[" + connPtr->peerAddress() + ":" + connPtr->peerPort() + "]: ",
                             err.message());
                    return;
                }
                LOGINFO(m_logTag, " start to connect with ", connPtr->peerAddress(), ":", connPtr->peerPort());
                connPtr->reset();
                connPtr->change_state(trunk::state::CONNECTION_STARTED);
                boost::asio::async_connect(*(connPtr->socket()), results,
                                           boost::bind(&Manager::onConnect, this,
                                                       connPtr,
                                                       boost::asio::placeholders::error));
            }

            void Manager::onConnect(trunk::ConnectionPtr connPtr,
                                    const boost::system::error_code &err)
            {
                if (err)
                {
                    if (err != boost::asio::error::operation_aborted)
                    {
                        LOGERROR(m_logTag,
                                 " Cannot connect to server[" + connPtr->peerAddress() + ":" + connPtr->peerPort() + "]: ",
                                 err.message());

                        connPtr->reset();
                    }
                    return;
                }

                connPtr->change_state(trunk::state::CONNECTION_ESTABLISHED);

                LOGINFO(m_logTag, " Outgoing connection up with ", connPtr->peerAddress(), ":", connPtr->peerPort());
                handleSipConnUp(connPtr->getId());

                connPtr->asynReadHeader();
            }

            void Manager::startAccept()
            {
                if (m_acceptors.empty())
                {
                    return;
                }

                m_stopped = false;

                for (auto &acceptor : m_acceptors)
                {
                    auto nextConn =
                        std::make_shared<trunk::connection>(
                            m_ioc, trunk::type::INCOMING,
                            m_connIdCntr,
                            Log::Tag((std::string)m_logTag + "-TCP:SCKT"));
                    m_connIdCntr++;
                    nextConn->trunk::interfaces::controller::client1::bind(shared_from_this());
                    nextConn->reset();
                    nextConn->change_state(trunk::state::CONNECTION_STARTED);
                    nextConn->setRole(acceptor->m_role);
                    acceptor->async_accept(*(nextConn->socket()),
                                           boost::bind(&Manager::onAccept, this,
                                                       nextConn,
                                                       boost::asio::placeholders::error));
                }
            }

            void Manager::stopAccept()
            {
                boost::system::error_code ec;

                for (auto &acceptor : m_acceptors)
                {
                    acceptor->close(ec);
                    if (ec)
                    {
                        LOGERROR(m_logTag, " Failed to close acceptor: ", ec.message());
                    }
                }
                m_acceptors.clear();

                m_stopped = true;
            }

            void Manager::onAccept(trunk::ConnectionPtr connPtr,
                                   const boost::system::error_code &err)
            {
                // Stop listen to new incoming connections
                // if user wants to destroy all connections
                if (m_stopped)
                    return;

                if (!err)
                {
                    connPtr->change_state(trunk::state::CONNECTION_ESTABLISHED);

                    auto result = m_connections.emplace(connPtr->getId(), connPtr);
                    if (!result.second)
                    {
                        LOGERROR(m_logTag,
                                 " Duplicate connection: ",
                                 "old ", *(result.first->second), ", ",
                                 "new ", *connPtr);
                    }

                    handleSipConnUp(connPtr->getId());
                    connPtr->asynReadHeader();
                }
                else if (err != boost::asio::error::operation_aborted)
                {
                    LOGERROR(m_logTag,
                             " Cannot accept connection: ", err.message());
                }
                else
                {
                    return;
                }

                startAccept();
            }

            void Manager::controllMessage(trunk::connection *connPtr,
                                          const MessagePtr &msg)
            {
                if (connPtr->state() < trunk::state::CONNECTION_ESTABLISHED)
                {
                    LOGERROR(m_logTag,
                             " Wrong connection state (expected state=ESTABLISHED): ", *connPtr);
                    return;
                }
                handleMessageTcp(*connPtr, msg);
            }

            void Manager::notify_timeout(const ::timers::id &id,
                                         const ::timers::time_point &start_time)
            {
                if (id == m_watchdogTimerId)
                {
                    LOGT(m_logTag, " watchdog triggered");
                    startAllConnections();
                }
                else
                {
                    handle_timeout(id, start_time);
                }
            }

            void Manager::handle_timeout(const ::timers::id &id,
                                         const ::timers::time_point &start_time)
            {
                /* Some action with an ancestor */
            }

        } // namespace transport

    } // namespace tcp

} // namespace sip
