#pragma once

#include "trunk_tcp.hpp"

// AFAP includes.
#include "components/support/interfaces_template.hpp"
#include "timers/src/timer_periodic.hpp"
#include "timers/src/timer_manager.hpp"

// STL includes.
#include <unordered_map>

namespace sip
{
    namespace tcp
    {
        namespace transport
        {

            using Map = std::unordered_map<trunk::Id, trunk::ConnectionPtr>;

            namespace interfaces
            {

                INTERFACE_S(handler,
                            INTERFACE_SERVER(void, handleMessageTcp, const trunk::connection &, const MessagePtr &);
                            INTERFACE_SERVER(void, handleSipConnUp, const trunk::Id &);)
                INTERFACE_C1(handler,
                             INTERFACE_CLIENT1(void, handleMessageTcp, const trunk::connection &, const MessagePtr &);
                             INTERFACE_CLIENT1(void, handleSipConnUp, const trunk::Id &);)
            } // namespace interfaces

            class Manager : public std::enable_shared_from_this<Manager>,
                            public ::timers::interfaces::notifier::server,
                            public interfaces::handler::client1,
                            public trunk::interfaces::controller::server
            {
                using Mutex = std::recursive_mutex;
                using Guard = std::lock_guard<Mutex>;

            public:
                explicit Manager(boost::asio::io_context &ioc,
                                 const Log::Tag &logTag = Log::Tag(""));
                ~Manager() override;

                void start(timers::duration period);
                virtual void stop();

                trunk::Id addConnection(const std::string &address,
                                        const std::string &port);
                void addListener(const network::address &ipAddress,
                                 const std::string &port,
                                 const sip::trunk::Role &role);
                void send(const trunk::Id &id, Message &msg) const;
                template <typename MsgType, typename... ColumnsTypes>
                void send(const trunk::Id &id, MsgType &msg,
                          ColumnsTypes &...columnsValues) const
                {
                    LOGE(m_logTag, "Remote endpoint should NOT be specified to send TCP data");
                }

            protected:
                void startMonitorTimer(timers::duration period);

                boost::asio::io_context &m_ioc; ///< Context for asynchronous operations
                const Log::Tag m_logTag;

            private:
                trunk::ConnectionPtr getConnection(const trunk::Id &id) const;
                void startAllConnections();

                void startConnect(trunk::ConnectionPtr conn_p);
                void onConnect(trunk::ConnectionPtr conn_p,
                               const boost::system::error_code &err);

                void startAccept();
                void stopAccept();
                void onAccept(trunk::ConnectionPtr conn_p,
                              const boost::system::error_code &err);

                void controllMessage(trunk::connection *connPtr, const MessagePtr &msg) override;
                void controllBrokenConnection(const trunk::Id &id) override;
                void controllConnectionClosed(const trunk::Id &id) override;

                void notify_timeout(const ::timers::id &id,
                                    const ::timers::time_point &start_time) override;
                virtual void handle_timeout(const ::timers::id &id,
                                            const ::timers::time_point &start_time);

                struct Acceptor : public boost::asio::ip::tcp::acceptor
                {
                    Acceptor(boost::asio::io_context &ioc,
                             const sip::trunk::Role &role)
                        : boost::asio::ip::tcp::acceptor(ioc),
                          m_role{role}
                    {
                    }

                    const sip::trunk::Role m_role;
                };

                using AcceptorPtr = std::shared_ptr<Acceptor>;
                using Acceptors = std::vector<AcceptorPtr>;

                std::shared_ptr<timers::manager> m_timerMngr; ///< An entity to handle timers
                Acceptors m_acceptors;                        ///< Acceptor used to listen for incoming connections.
                bool m_stopped{false};                        ///< Indicates stopping to listen to incoming connections
                trunk::Id m_connIdCntr{0};                    ///< monotonic counter to create connection ID
                Map m_connections;                            ///< Connections storage
                timers::id m_watchdogTimerId;                 ///< Watchdog timer ID
                mutable Mutex m_guard;                        ///< Protector of shared data in multiple threads
            };

        } // namespace transport

    } // namespace tcp

} // namespace sip
