#include "transport_udp.hpp"

#include <iostream>
#include <boost/bind/bind.hpp>

namespace sip
{
    namespace udp
    {

        namespace transport
        {

            Manager::Manager(boost::asio::io_context &ioc,
                             const Log::Tag &logTag)
                : m_ioc(ioc),
                  m_logTag(logTag)
            {
            }

            Manager::~Manager()
            {
                stop();
            }

            void Manager::setSocketRcvBufSize(const int val)
            {
                m_socketRcvBufSize = val;
            }

            void Manager::start()
            {
                for (auto &socket : m_sockets)
                {
                    socket->asyncRead();
                }
            }

            void Manager::stop()
            {
                std::lock_guard<std::recursive_mutex> guard(m_guard);

                trunk::interfaces::controller::server::stop_server();

                for (auto &socket : m_sockets)
                {
                    socket->cancel();
                }
                m_sockets.clear();
            }

            trunk::Id Manager::addSocket(const network::address &address,
                                         const std::string &port,
                                         const ::sip::trunk::Role &role)
            {
                std::lock_guard<std::recursive_mutex> guard(m_guard);

                trunk::SocketPtr socket = std::make_shared<trunk::Socket>(
                    m_ioc, address, port, role,
                    Log::Tag((std::string)m_logTag + "-UDP:SCKT"));
                trunk::Id id = m_sockets.size();
                socket->setId(id);
                socket->setRcvBufSize(m_socketRcvBufSize);
                m_sockets.push_back(socket);

                socket->trunk::interfaces::controller::client1::bind(shared_from_this());

                LOGI(m_logTag, "listener add at ", address, ":", port,
                     " role=", role2str(role));

                return id;
            }

            void Manager::controllMessage(trunk::Socket *socket, const MessagePtr &msg)
            {
                interfaces::handler::client1::handleMessageUdp(*socket, msg);
            }

            void asyncSendHandler(
                const boost::system::error_code &error,
                std::size_t bytes_transferred)
            {
            }

            void Manager::send(const trunk::Id &id, Message &msg,
                               const std::string &remIp,
                               const std::string &remPort) const
            {
                send(id, msg, remIp, remPort, asyncSendHandler);
            }
            void Manager::send(const trunk::Id &id, Message &msg,
                               const std::string &remIp,
                               const std::string &remPort,
                               CallbackUdp_t callback) const
            {
                int errorCode;
                std::string msgEncoded = msg.encode(errorCode);
                send(id, msgEncoded, remIp, remPort, callback);
            }
            void Manager::send(const trunk::Id &id, const std::string &msg,
                               const std::string &remIp,
                               const std::string &remPort) const
            {
                send(id, msg, remIp, remPort, asyncSendHandler);
            }
            void Manager::send(const trunk::Id &id,
                               const std::string &msgEncoded,
                               const std::string &remIp,
                               const std::string &remPort,
                               CallbackUdp_t callback) const
            {
                auto socketPtr = m_sockets[id];
                socketPtr->asyncSend(remIp, remPort, msgEncoded, callback);
            }

            std::string Manager::getLocalAddress(const ::sip::trunk::Reference &ref) const
            {
                std::lock_guard<std::recursive_mutex> guard(m_guard);

                auto socketPtr = m_sockets[ref.m_key];
                return socketPtr->getLocalAddress();
            }
            std::string Manager::getLocalPort(const ::sip::trunk::Reference &ref) const
            {
                std::lock_guard<std::recursive_mutex> guard(m_guard);

                auto socketPtr = m_sockets[ref.m_key];
                return socketPtr->getLocalPort();
            }

        } // namespace transport

    } // namespace udp

} // namespace sip
