#include "transports.hpp"

namespace sip
{
    namespace transport
    {

        Manager::Manager(boost::asio::io_context &ioc,
                         const Log::Tag &logTag)
            : m_ioc(ioc),
              m_logTag{logTag}
        {
            m_udpManager = std::make_shared<::sip::udp::transport::Manager>(
                m_ioc,
                Log::Tag((std::string)m_logTag + "-UDP:MNGR"));
            m_tcpManager = std::make_shared<::sip::tcp::transport::Manager>(
                m_ioc,
                Log::Tag((std::string)m_logTag + "-TCP:MNGR"));
        }

        Manager::~Manager()
        {
            stop();
        }

        void Manager::setUdpSocketRcvBufSize(const int val)
        {
            m_udpManager->setSocketRcvBufSize(val);
        }

        void Manager::stop()
        {
            udp::transport::interfaces::handler::server::stop_server();
            tcp::transport::interfaces::handler::server::stop_server();

            m_udpManager->stop();
            m_tcpManager->stop();
        }

        void Manager::start()
        {
            timers::duration duration = std::chrono::milliseconds(100);

            m_udpManager->bind(shared_from_this());
            m_tcpManager->bind(shared_from_this());

            m_udpManager->start();
            m_tcpManager->start(duration);
        }

        void Manager::handleMessageUdp(const udp::trunk::Socket &trunk,
                                       const MessagePtr &msg)
        {
            handleMessage({trunk::Protocol::UDP, trunk.getId(), trunk.getRole()}, msg);
        }
        void Manager::handleMessageTcp(const tcp::trunk::connection &trunk,
                                       const MessagePtr &msg)
        {
            handleMessage({trunk::Protocol::TCP, trunk.getId(), trunk.getRole()}, msg);
        }

        void Manager::handleSipConnUp(const tcp::trunk::Id &id)
        {
            handleTrunkTcpUp({trunk::Protocol::TCP, id});
        }

        void Manager::addListener(trunk::Protocol protocol,
                                  const network::address &address,
                                  const std::string &port,
                                  const sip::trunk::Role &role)
        {
            switch (protocol)
            {
                case trunk::Protocol::UDP:
                    m_udpManager->addSocket(address, port, role);
                    break;
                case trunk::Protocol::TCP:
                    m_tcpManager->addListener(address, port, role);
                    break;
            }
        }

        trunk::Reference Manager::addTrunkTcp(const std::string &address,
                                              const std::string &port)
        {
            return {trunk::Protocol::TCP,
                    m_tcpManager->addConnection(address, port)};
        }

        std::string Manager::getLocalAddress(const ::sip::trunk::Reference &ref) const
        {
            if (ref.m_ulp == ::sip::trunk::Protocol::UDP)
            {
                return m_udpManager->getLocalAddress(ref);
            }
            return std::string();
        }
        std::string Manager::getLocalPort(const ::sip::trunk::Reference &ref) const
        {
            if (ref.m_ulp == ::sip::trunk::Protocol::UDP)
            {
                return m_udpManager->getLocalPort(ref);
            }
            return std::string();
        }

    } // namespace transport

    namespace trunk
    {
        std::string protocol2str(Protocol protocol)
        {
            switch (protocol)
            {
                case Protocol::UDP:
                    return "UDP";
                case Protocol::TCP:
                    return "TCP";
                default:
                    return "UNKWN protocol";
            }
        }

        std::string role2str(Role protocol)
        {
            switch (protocol)
            {
                case Role::INC:
                    return "INC";
                case Role::OUT:
                    return "OUT";
                case Role::ANY:
                    return "ANY";
                default:
                    return "UNKWN protocol";
            }
        }

        std::ostream &operator<<(std::ostream &os, const Reference &ref)
        {
            std::stringbuf trace;
            std::ostream logger(&trace);
            logger << protocol2str(ref.m_ulp) << ":" << ref.m_key;
            logger << "-" << role2str(ref.m_role);
            os << trace.str();
            return os;
        }
    } // namespace trunk

} // namespace sip
