#include "trunk_tcp.hpp"

// Boost includes.
#include <boost/bind/bind.hpp>

namespace sip
{
    namespace tcp
    {
        namespace trunk
        {
            connection::connection(boost::asio::io_context &ioc,
                                   trunk::type type,
                                   int connId,
                                   const Log::Tag &logTag) : m_type(type),
                                                             m_strand(boost::asio::make_strand(ioc.get_executor())),
                                                             m_logTag{logTag}

            {
            }
            connection::connection(boost::asio::io_context &ioc,
                                   trunk::type type,
                                   const std::string &address,
                                   const std::string &port,
                                   int connId,
                                   const Log::Tag &logTag) : m_type(type),
                                                             m_peerAddress(address),
                                                             m_peerPort(port),
                                                             m_strand(boost::asio::make_strand(ioc.get_executor())),
                                                             m_logTag{logTag}
            {
            }

            TcpSocketPtr connection::socket() const
            {
                return m_socket;
            }

            type connection::type() const
            {
                return m_type;
            }

            state connection::state() const
            {
                return m_state;
            }

            const std::string &connection::peerAddress() const
            {
                return m_peerAddress;
            }

            void connection::peerAddress(const std::string &address)
            {
                m_peerAddress = address;
            }

            const std::string &connection::peerPort() const
            {
                return m_peerPort;
            }

            void connection::peerPort(const std::string &port)
            {
                m_peerPort = port;
            }

            void connection::change_state(trunk::state state)
            {
                LOGDEBUG(m_logTag, " Change state ", connState2str(m_state), "->", connState2str(state), " for ", *this);
                m_state = state;
            }

            void connection::send(std::string &msg)
            {
                if (m_state < state::CONNECTION_ESTABLISHED)
                {
                    LOGINFO(m_logTag, " Cannot send message via ", *this, ": wrong state=", connState2str(m_state));
                    return;
                }

                boost::system::error_code err;
                boost::asio::write(*m_socket, boost::asio::buffer(msg), err);
                if (err && err != boost::asio::error::operation_aborted)
                {
                    LOGINFO(m_logTag, " Lost ", *this, ": ", err.message());
                    reset();
                    interfaces::controller::client1::controllBrokenConnection(m_id);
                }
            }

            void connection::disconnect()
            {
                boost::asio::post(m_strand,
                                  std::bind(&connection::disconnect_int, shared_from_this()));
            }

            void connection::disconnect_int()
            {
                LOGDEBUG(m_logTag, " Disconnect ", *this);

                boost::system::error_code err;
                if (m_state > state::CONNECTION_DOWN)
                {
                    change_state(state::CONNECTION_DOWN);
                    m_socket->shutdown(boost::asio::ip::tcp::socket::shutdown_both, err);
                    if (err)
                    {
                        LOGERROR(m_logTag, " Cannot close ", *this, ": ", err.message());
                    }
                }
            }

            void connection::reset()
            {
                change_state(state::CONNECTION_INIT);
                m_socket.reset(new TcpSocket(m_strand));
            }

            std::string make_string(const boost::asio::streambuf &streambuf)
            {
                return {boost::asio::buffers_begin(streambuf.data()),
                        boost::asio::buffers_end(streambuf.data())};
            }

            void connection::asynReadHeader()
            {
                boost::asio::async_read(*m_socket,
                                        m_buffer,
                                        boost::asio::transfer_at_least(1),
                                        boost::bind(&connection::handleReadHeader, shared_from_this(),
                                                    boost::asio::placeholders::error,
                                                    boost::asio::placeholders::bytes_transferred));
            }

            void connection::handleReadHeader(const boost::system::error_code &err,
                                              std::size_t)
            {
                if (!err)
                {
                    size_t bytesConsumed =
                        ::sip::trunk::handleMessages<interfaces::controller::client1, connection>(
                            this,
                            boost::asio::buffer_cast<const char *>(m_buffer.data()),
                            m_buffer.size());

                    m_buffer.consume(bytesConsumed);

                    asynReadHeader();
                }
                else if (err != boost::asio::error::operation_aborted)
                {
                    LOGINFO(m_logTag, " Lost ", *this, ": ", err.message());
                    reset();
                    interfaces::controller::client1::controllBrokenConnection(m_id);
                }
            }

            std::string connType2str(trunk::type type)
            {
                switch (type)
                {
                case type::INCOMING:
                    return "INCOMING";
                case type::OUTGOING:
                    return "OUTGOING";
                default:
                    return "UNKWN(" + std::to_string(static_cast<int>(type)) + ")";
                }
            }

            std::string connState2str(trunk::state state)
            {
                switch (state)
                {
                case state::CONNECTION_INIT:
                    return "INIT";
                case state::CONNECTION_STARTED:
                    return "STARTED";
                case state::CONNECTION_DOWN:
                    return "DOWN";
                case state::CONNECTION_ESTABLISHED:
                    return "ESTABLISHED";
                default:
                    return "UNKWN(" + std::to_string(static_cast<int>(state)) + ")";
                }
            }

            std::ostream &operator<<(std::ostream &os, const connection &c)
            {
                std::stringbuf trace;
                std::ostream logger(&trace);
                logger << "conn[id:" << c.getId() << ":" << c.m_logTag;
                if (c.m_type != type::INCOMING)
                {
                    logger << "->";
                }
                if (c.m_type == type::INCOMING)
                {
                    logger << "<-";
                }
                logger << ";" << connState2str(c.m_state);
                if (c.m_type == type::OUTGOING)
                {
                    logger << ";ep=" << c.m_peerAddress << ":" << c.m_peerPort;
                }
                logger << "]";
                os << trace.str();
                return os;
            }

        } // namespace trunk

    } // namespace tcp

} // namespace sip
