#pragma once

#include "trunks.hpp"

// AFAP includes.
#include "components/support/interfaces_template.hpp"
#include "components/tools/src/Tools.hpp"

#include <boost/asio.hpp>

namespace sip
{
    namespace tcp
    {

        namespace trunk
        {

            class connection;
            using TcpSocket = boost::asio::ip::tcp::socket;
            using TcpSocketPtr = std::shared_ptr<TcpSocket>;
            using ConnectionPtr = std::shared_ptr<connection>;
            using Id = ::sip::trunk::Id;

            namespace interfaces
            {

                INTERFACE_S(controller,
                            INTERFACE_SERVER(void, controllMessage, connection *, const MessagePtr &);
                            INTERFACE_SERVER(void, controllBrokenConnection, const Id &);
                            INTERFACE_SERVER(void, controllConnectionClosed, const Id &);)
                INTERFACE_C1(controller,
                             INTERFACE_CLIENT1(void, controllMessage, connection *, const MessagePtr &);
                             INTERFACE_CLIENT1(void, controllBrokenConnection, const Id &);
                             INTERFACE_CLIENT1(void, controllConnectionClosed, const Id &);)
            } // namespace interfaces

            enum class type
            {
                INCOMING,
                OUTGOING
            };

            enum class state
            {
                CONNECTION_INIT,
                CONNECTION_STARTED,
                CONNECTION_DOWN,
                CONNECTION_ESTABLISHED
            };

            class connection : public std::enable_shared_from_this<connection>,
                               public interfaces::controller::client1,
                               public ::sip::trunk::Generic
            {
            public:
                explicit connection(boost::asio::io_context &ioc,
                                    trunk::type type,
                                    int connId,
                                    const Log::Tag &logTag = Log::Tag(""));
                explicit connection(boost::asio::io_context &ioc,
                                    trunk::type type,
                                    const std::string &address,
                                    const std::string &port,
                                    int connId,
                                    const Log::Tag &logTag = Log::Tag(""));
                /**
                 * @brief Changes connection state
                 *
                 * @param[in] state a new state
                 */
                void change_state(state state);
                void asynReadHeader();

                /// @returns socket associated with the connection.
                TcpSocketPtr socket() const;
                /// @returns connection type
                trunk::type type() const;
                /// @returns connection state
                trunk::state state() const;
                /// @returns remote IP address, empty string is returned in case of OUTGOING connections
                const std::string &peerAddress() const;
                void peerAddress(const std::string &address);
                /// @returns remote port, empty string is returned in case of OUTGOING connections
                const std::string &peerPort() const;
                void peerPort(const std::string &port);

                void send(std::string &msg);
                void disconnect();
                void reset();

                friend std::ostream &operator<<(std::ostream &os, const connection &c);

            private:
                void handleReadHeader(const boost::system::error_code &err,
                                      std::size_t bytesReceived);

                void disconnect_int();

                trunk::type m_type;                                      ///< Connection type
                trunk::state m_state{state::CONNECTION_INIT};            ///< Connection state
                boost::asio::streambuf m_buffer{MAX_SIP_MESSAGE_LENGTH}; ///< Buffer for incoming data.
                std::string m_peerAddress;                               ///< Remote address (applicable for outgoing connections only)
                std::string m_peerPort;                                  ///< Remote port (applicable for outgoing connections only)
                TcpSocketPtr m_socket;                                   ///< TCP socket for connection

                boost::asio::strand<boost::asio::io_context::executor_type> m_strand;

                const Log::Tag m_logTag;
            };

            std::string connType2str(trunk::type type);
            std::string connState2str(trunk::state state);
            std::ostream &operator<<(std::ostream &os, const connection &c);

        } // namespace trunk

    } // namespace tcp

} // namespace sip
