// Own include.
#include "trunk_udp.hpp"

// Boost includes.
#include <boost/bind/bind.hpp>

namespace sip::udp::trunk
{
    Socket::Socket(boost::asio::io_service &ioc,
                   const network::address &ipAddr,
                   const std::string &port,
                   const ::sip::trunk::Role &role,
                   const Log::Tag &logTag)
        : ::sip::trunk::Generic(role),
          m_fromEndpoint(ipAddr, strtol(port.c_str(), nullptr, 10)),
          m_socket(ioc),
          m_logTag(logTag)
    {
        boost::system::error_code err;

        m_socket.open(boost::asio::ip::udp::v4(), err);
        if (err)
        {
            LOGE(m_logTag, "Failed to open socket: ", err.what());
            return;
        }

        m_socket.bind(m_fromEndpoint, err);
        if (err)
        {
            LOGE(m_logTag, "Failed to bind socket: ", err.what());
            return;
        }

        m_localAddress = ipAddr.to_string();
        m_localPort = port;
    }

    Socket::~Socket()
    {
        if (m_socket.is_open())
        {
            m_socket.close();
        }
    }

    void Socket::setRcvBufSize(const int val)
    {
        if (val <= 0)
        {
            return;
        }

        boost::asio::socket_base::receive_buffer_size option(val);
        m_socket.set_option(option);
    }

    void Socket::cancel()
    {
        m_socket.cancel();
    }

    void Socket::asyncSend(const std::string &remIp,
                           const std::string &remPort,
                           const std::string &msg,
                           CallbackUdp_t callback)
    {
        if (!m_socket.is_open())
        {
            LOGE(m_logTag, "Socket must be opened to do sending!");
            return;
        }

        m_socket.async_send_to(boost::asio::buffer(msg),
                               boost::asio::ip::udp::endpoint(boost::asio::ip::address::from_string(remIp),
                                                              strtol(remPort.c_str(), nullptr, 10)),
                               callback);
    }

    void Socket::asyncRead()
    {
        if (!m_socket.is_open())
        {
            LOGE(m_logTag, "Socket must be opened to do receiving!");
            return;
        }

        m_buffer.reset(new char[MAX_SIP_MESSAGE_LENGTH]);

        m_socket.async_receive_from(boost::asio::buffer(m_buffer.get(), MAX_SIP_MESSAGE_LENGTH),
                                    m_fromEndpoint,
                                    boost::bind(&Socket::readDatagram,
                                                shared_from_this(),
                                                boost::asio::placeholders::error,
                                                boost::asio::placeholders::bytes_transferred));
    }

    void Socket::readDatagram(const boost::system::error_code &err, std::size_t udpMsgSize)
    {
        if (err)
        {
            if (err != boost::asio::error::operation_aborted)
            {
                LOGI(m_logTag, "Socket read failure with message: ", err.message());
            }
            return;
        }

        const auto tmpBuffer = std::move(m_buffer);

        asyncRead();

        ::sip::trunk::handleMessages<interfaces::controller::client1, Socket>(
            this,
            tmpBuffer.get(),
            udpMsgSize);
    }

    const std::string &Socket::getLocalAddress() const
    {
        return m_localAddress;
    }

    const std::string &Socket::getLocalPort() const
    {
        return m_localPort;
    }

} // namespace sip::udp::trunk
