#pragma once

#include "message.hpp"
#include "utils.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"

#define CONTENT_LENGTH_HEADER_STR CRLF "content-length:"
#define CONTENT_LENGTH_HEADER_COMPACT_STR CRLF "l:"
#define CONTENT_LENGTH_HEADER_STR2 CRLF "content-length "
#define CONTENT_LENGTH_HEADER_COMPACT_STR2 CRLF "l "

#ifndef MAX_SIP_MESSAGE_LENGTH
#define MAX_SIP_MESSAGE_LENGTH 8000
#endif

namespace sip
{
    namespace trunk
    {

        using Id = uint32_t;

        enum class Protocol
        {
            UDP,
            TCP
        };

        enum class Role
        {
            ANY,
            INC,
            OUT
        };

        struct Generic
        {
            Generic(const Role &role = Role::ANY)
                : m_role(role),
                  m_id{0}
            {
            }
            Role getRole() const
            {
                return m_role;
            }
            void setRole(const Role &role)
            {
                m_role = role;
            }
            void setId(const Id &id)
            {
                m_id = id;
            }
            Id getId() const
            {
                return m_id;
            }

        protected:
            Role m_role;
            Id m_id;
        };

        struct Reference
        {
            Protocol m_ulp;
            Id m_key;
            Role m_role;

            bool operator==(const Reference &other) const
            {
                return (m_ulp == other.m_ulp &&
                        m_key == other.m_key &&
                        m_role == other.m_role);
            }
        };

        std::string protocol2str(Protocol protocol);
        std::string role2str(Role protocol);
        std::ostream &operator<<(std::ostream &os, const Reference &ref);

        template <class S, class T>
        size_t handleMessages(T *object, const char *buf, size_t bufLen)
        {
            size_t bufLenLeft = bufLen;
            size_t consumed = 0;
            char *endHeaders;

            while (bufLenLeft > 0)
            {
                endHeaders = sip::utils::findStr(buf, bufLenLeft, END_HEADERS_STR);
                if (endHeaders == nullptr)
                {
                    break;
                }

                if (buf == endHeaders)
                {
                    consumed += 4;
                    bufLenLeft -= 4;
                    buf += 4;
                    continue;
                }

                *endHeaders = '\0';

                const char *contentLengthHeader = strcasestr(buf, CONTENT_LENGTH_HEADER_STR);

                if (!contentLengthHeader)
                    contentLengthHeader = strcasestr(buf, CONTENT_LENGTH_HEADER_STR2);

                if (!contentLengthHeader)
                    contentLengthHeader = strcasestr(buf, CONTENT_LENGTH_HEADER_COMPACT_STR);

                if (!contentLengthHeader)
                    contentLengthHeader = strcasestr(buf, CONTENT_LENGTH_HEADER_COMPACT_STR2);

                if (contentLengthHeader != nullptr)
                {
                    contentLengthHeader = strchr(contentLengthHeader, ':');
                    contentLengthHeader++;
                }

                if (!contentLengthHeader)
                {
                    LOGWARN(" Message has no content-length");
                }

                int contentLength = contentLengthHeader ? sip::utils::atoi(contentLengthHeader) : 0;

                *endHeaders = END_HEADERS_STR[0];
                endHeaders += sizeof(END_HEADERS_STR) - 1;

                if (contentLength < 0)
                {
                    size_t msgLen = (endHeaders - buf);
                    consumed += msgLen;
                    bufLenLeft -= msgLen;
                    buf += msgLen;
                    continue;
                }

                size_t msgLen = (endHeaders - buf + contentLength);

                if (msgLen > bufLenLeft)
                {
                    return consumed;
                }

                auto msg = std::make_shared<sip::Message>();
                if (msg->doParse(buf, msgLen) == sip::SUCCESS)
                {
                    object->S::controllMessage(object, msg);
                }

                consumed += msgLen;
                bufLenLeft -= msgLen;
                buf += msgLen;
            }

            if (bufLenLeft == 2 && buf[0] == g_cr && buf[1] == g_lf)
            {
                consumed += 2;
            }

            return consumed;
        }

    } // namespace trunk

} // namespace sip
