#include "uri.hpp"
#include "utils.hpp"

// STL includes.
#include <cstring>

namespace sip
{

    int Uri::doParseParams(const std::string &paramsStr)
    {
        std::string pname;
        std::string pvalue;

        const char *comma;
        const char *equal;
        const char *params = paramsStr.c_str();

        if (params[0] != ';')
        {
            return SYNTAXERROR;
        }

        equal = utils::findSeparatorAfter(params + 1, '=', ';');
        comma = std::strchr(params + 1, ';');

        while (comma != nullptr)
        {
            if (equal == nullptr)
            {
                equal = comma;
            }
            else
            {
                if (comma - equal < 2)
                {
                    return SYNTAXERROR;
                }

                pvalue = std::string(equal + 1, comma - equal - 1);
                utils::doUnescape(pvalue);
            }

            if (equal - params < 2)
            {
                return SYNTAXERROR;
            }

            pname = std::string(params + 1, equal - params - 1);
            utils::doUnescape(pname);

            m_params.emplace_back(std::move(pname), std::move(pvalue));

            params = comma;
            equal = utils::findSeparatorAfter(params + 1, '=', ';');
            comma = std::strchr(params + 1, ';');
        }

        comma = params + std::strlen(params);

        if (equal == nullptr)
        {
            equal = comma;
            pvalue = "";
        }
        else
        {
            if (comma - equal < 2)
            {
                return SYNTAXERROR;
            }

            pvalue = std::string(equal + 1, comma - equal - 1);
            utils::doUnescape(pvalue);
        }

        if (equal - params < 2)
        {
            return SYNTAXERROR;
        }

        pname = std::string(params + 1, equal - params - 1);
        utils::doUnescape(pname);

        m_params.emplace_back(std::move(pname), std::move(pvalue));

        return SUCCESS;
    }

    int Uri::doParseHeaders(const char *headers)
    {
        const char *_and;
        const char *equal;

        if (headers[0] != '?')
        {
            return SYNTAXERROR;
        }

        equal = std::strchr(headers, '=');
        _and = std::strchr(headers + 1, '&');

        if (equal == nullptr)
        {
            return SYNTAXERROR;
        }

        do
        {
            std::string hname;
            std::string hvalue;

            hname = std::string(headers + 1, equal - headers - 1);
            utils::doUnescape(hname);

            if (_and != nullptr)
            {
                if (_and - equal < 2)
                {
                    return SYNTAXERROR;
                }

                hvalue = std::string(equal + 1, _and - equal - 1);
                utils::doUnescape(hvalue);
            }
            else
            {
                if (headers + std::strlen(headers) - equal + 1 < 2)
                {
                    return SYNTAXERROR;
                }

                hvalue = std::string(equal + 1, headers + std::strlen(headers) - equal);
                utils::doUnescape(hvalue);
            }

            m_headers.emplace_back(std::move(hname), std::move(hvalue));

            if (_and == nullptr)
            {
                equal = nullptr;
            }
            else
            {
                headers = _and;
                equal = std::strchr(headers, '=');
                _and = std::strchr(headers + 1, '&');

                if (equal == nullptr)
                {
                    return SYNTAXERROR;
                }
            }
        } while (equal != nullptr);

        return SUCCESS;
    }

    int Uri::doParse(const char *buf)
    {
        const char *username;
        const char *password;
        const char *host;
        const char *port;
        const char *params;
        const char *headers;
        const char *tmp;
        int i;

        if (buf == nullptr)
        {
            return BADPARAMETER;
        }

        tmp = std::strchr(buf, ':');

        if (tmp == nullptr)
        {
            return SYNTAXERROR;
        }

        if (tmp - buf < 2)
        {
            return SYNTAXERROR;
        }

        i = 0;

        while (buf + i < tmp)
        {
            if (!utils::isAlpha(buf[i]))
            {
                return SYNTAXERROR;
            }

            i++;
        }

        m_scheme = std::string(buf, tmp - buf);

        if (std::strchr(m_scheme.data(), ' ') != nullptr)
        {
            return SYNTAXERROR;
        }

        if (m_scheme.length() < 3 ||
            (0 != strncasecmp(m_scheme.data(), "sip", 3) &&
             0 != strncasecmp(m_scheme.data(), "sips", 4)))
        {
            size_t j = std::strlen(tmp + 1);

            if (j < 2)
            {
                return SYNTAXERROR;
            }

            m_string = std::string(tmp + 1, j);
            return SUCCESS;
        }

        username = std::strchr(buf, ':');
        if (username == nullptr)
        {
            return SYNTAXERROR;
        }

        host = std::strchr(buf, '@');
        if (host == nullptr)
        {
            host = username;
        }
        else if (username[1] == '@')
        {
            host = username + 1;
        }
        else
        {
            password = utils::findSeparatorAfter(username + 1, ':', '@');

            if (password == nullptr)
            {
                password = host;
            }
            else
            {
                if (host - password < 2)
                {
                    return SYNTAXERROR;
                }

                m_password = std::string(password + 1, host - password - 1);
                utils::doUnescape(m_password);
            }

            if (password - username < 2)
            {
                return SYNTAXERROR;
            }

            m_username = std::string(username + 1, password - username - 1);
            utils::doUnescape(m_username);
        }

        headers = std::strchr(host, '?');

        if (headers == nullptr)
        {
            headers = buf + std::strlen(buf);
        }
        else
        {
            doParseHeaders(headers);
        }

        params = std::strchr(host, ';');

        if (params == nullptr)
        {
            params = headers;
        }
        else
        {
            if (headers - params + 1 < 2)
            {
                return SYNTAXERROR;
            }

            std::string tmpbuf(params, headers - params);
            doParseParams(tmpbuf);
        }

        port = params - 1;

        while (port > host && *port != ']' && *port != ':')
        {
            port--;
        }

        if (*port == ':')
        {
            if (host == port)
            {
                port = params;
            }
            else
            {
                if ((params - port < 2) || (params - port > 8))
                {
                    return SYNTAXERROR;
                }

                m_port = utils::doCopyAndClearLws(port + 1, params - port - 1);
            }
        }
        else
        {
            port = params;
        }

        tmp = port;

        while (tmp > host && *tmp != ']')
        {
            tmp--;
        }

        if (*tmp == ']')
        {
            port = tmp;

            while (host < port && *host != '[')
            {
                host++;
            }

            if (host >= port)
            {
                return SYNTAXERROR;
            }
        }

        if (port - host < 2)
        {
            return SYNTAXERROR;
        }

        m_host = utils::doCopyAndClearLws(host + 1, port - host - 1);

        return SUCCESS;
    }

    std::string Uri::escapeUserinfo(const std::string &string) const
    {
        constexpr char reserved[] = "-_.!~*'()&=+$,;?/\0";
        return sip::utils::escapeNonAsciiAndReserved(string, reserved);
    }
    std::string Uri::escapePassword(const std::string &string) const
    {
        constexpr char reserved[] = "-_.!~*'()&=+$,\0";
        return sip::utils::escapeNonAsciiAndReserved(string, reserved);
    }
    std::string Uri::escapeParam(const std::string &string) const
    {
        constexpr char reserved[] = "-_.!~*'()[]/:&+$\0";
        return sip::utils::escapeNonAsciiAndReserved(string, reserved);
    }
    std::string Uri::escapeHeaderParam(const std::string &string) const
    {
        constexpr char reserved[] = "-_.!~*'()[]/?:+$\0";
        return sip::utils::escapeNonAsciiAndReserved(string, reserved);
    }

    std::string Uri::toStr(int &errorCode)
    {
        if (m_host.empty() && m_string.empty())
        {
            errorCode = BADPARAMETER;
            return "";
        }

        if (m_scheme.empty() && !m_string.empty())
        {
            errorCode = BADPARAMETER;
            return "";
        }

        std::string ret;

        if (m_string.empty() && m_scheme.empty())
        {
            ret = std::string("sip") + ":";
        }
        else
        {
            ret = m_scheme + ":";
        }

        if (!m_string.empty())
        {
            ret += m_string;
            errorCode = SUCCESS;
            return ret;
        }

        if (!m_username.empty())
        {
            ret += escapeUserinfo(m_username);
        }

        if (!m_password.empty() && !m_username.empty())
        {
            ret += ":" + escapePassword(m_password);
        }

        if (!m_username.empty())
        {
            ret += "@";
        }

        if (std::strchr(m_host.c_str(), ':') != nullptr)
        {
            ret += "[" + m_host + "]";
        }
        else
        {
            ret += m_host;
        }

        if (!m_port.empty())
        {
            ret += ":" + m_port;
        }

        for (auto it = m_params.begin(); it != m_params.end(); it++)
        {
            if (strcasecmp(it->m_name.c_str(), "x-obr") == 0 ||
                strcasecmp(it->m_name.c_str(), "x-obp") == 0)
            {
                continue;
            }

            ret += ";" + escapeParam(it->m_name);

            if (!(it->m_value.empty()))
            {
                ret += "=" + escapeParam(it->m_value);
            }
        }

        for (auto it = m_headers.begin(); it != m_headers.end(); it++)
        {
            if (it == m_headers.begin())
            {
                ret += "?";
            }
            else
            {
                ret += "&";
            }
            ret += escapeHeaderParam(it->m_name) + "=" +
                   escapeHeaderParam(it->m_value);
        }

        errorCode = SUCCESS;
        return ret;
    }

} // namespace sip
