#include "utils.hpp"
#include "generic_param.hpp"

// AFAP includes.
#include "components/log/src/Log.hpp"

// STL includes.
#include <algorithm>
#include <climits>
#include <cstring>
#include <iomanip>
#include <sstream>

namespace sip
{
    namespace utils
    {

        void toLower(std::string &str)
        {
            std::transform(str.begin(), str.end(), str.begin(), ::tolower);
        }

        const char *findSeparatorAfter(const char *ch,
                                       int separatorToFind,
                                       int afterSeparator)
        {
            const char *ind;
            const char *tmp{nullptr};

            ind = std::strchr(ch, separatorToFind);

            if (ind == nullptr)
            {
                return nullptr;
            }

            if (afterSeparator != 0)
            {
                tmp = std::strchr(ch, afterSeparator);
            }

            if (tmp != nullptr)
            {
                if (ind < tmp)
                {
                    return ind;
                }
            }
            else
            {
                return ind;
            }

            return nullptr;
        }

        const char *doFindQuote(const char *qstring)
        {
            const char *quote;

            quote = std::strchr(qstring, '"');

            if (quote == qstring)
            {
                return quote;
            }

            if (quote == nullptr)
            {
                return nullptr;
            }

            int i = 1;

            for (;;)
            {
                if (0 == std::strncmp(quote - i, "\\", 1))
                {
                    i++;
                }
                else
                {
                    if (i % 2 == 1) // the '"' was not escaped
                    {
                        return quote;
                    }

                    quote = std::strchr(quote + 1, '"');

                    if (quote == nullptr)
                    {
                        return nullptr;
                    }

                    i = 1;
                }

                if (quote - i == qstring - 1)
                {
                    if (*qstring == '\\')
                    {
                        i++; // an escape char was not counted
                    }

                    if (i % 2 == 0) // the '"' was not escaped
                    {
                        return quote;
                    }
                    else
                    {
                        qstring = quote + 1;
                        quote = std::strchr(quote + 1, '"');

                        if (quote == nullptr)
                        {
                            return nullptr;
                        }

                        i = 1;
                    }
                }
            }
        }

        const std::string_view doUnQuote(const std::string &string)
        {
            if (string.empty())
            {
                return std::string_view();
            }

            int first = 0;
            if (string[first] == '"')
            {
                first++;
            }
            int last = string.length() - 1;
            if (string[last] == '"')
            {
                last--;
            }
            return std::string_view(&string[first], last - first + 1);
        }

        void doUnescape(std::string &string)
        {
            size_t alloc = string.length() + 1;
            unsigned char in;
            int index = 0;
            unsigned int hex;
            const char *ptr;

            ptr = &(string[0]);

            while (--alloc > 0)
            {
                in = *ptr;

                if ('%' == in)
                {
                    /* encoded part */
                    if (alloc > 2 && sscanf(ptr + 1, "%02X", &hex) == 1)
                    {
                        in = (unsigned char)hex;

                        if (ptr[2]
                            && ((ptr[2] >= '0' && ptr[2] <= '9')
                                || (ptr[2] >= 'a' && ptr[2] <= 'f')
                                || (ptr[2] >= 'A' && ptr[2] <= 'F')))
                        {
                            alloc -= 2;
                            ptr += 2;
                        }
                        else
                        {
                            alloc -= 1;
                            ptr += 1;
                        }
                    }
                    else
                    {
                        break;
                    }
                }

                string[index] = in;
                index++;
                ptr++;
            }

            string.resize(index);
        }

        void doClearLws(std::string &word)
        {
            if (word.empty())
            {
                return;
            }

            const char *pbeg = &(word.front()) + std::strspn(&(word.front()), " \r\n\t");
            const char *pend = &(word.back());

            while ((' ' == *pend) || ('\r' == *pend) || ('\n' == *pend) || ('\t' == *pend))
            {
                pend--;

                if (pend < pbeg)
                {
                    word.clear();
                    return;
                }
            }

            std::memmove(&(word[0]), pbeg, pend - pbeg + 1);
            word.resize(pend - pbeg + 1);
        }

        std::string doCopyAndClearLws(const char *src, size_t len)
        {
            if (src == nullptr || len == 0)
            {
                return std::string();
            }

            const char *pbeg = src + strspn(src, " \r\n\t");

            const char *pend = src + len - 1;
            while ((' ' == *pend)
                   || ('\r' == *pend)
                   || ('\n' == *pend)
                   || ('\t' == *pend))
            {
                if (pend < pbeg)
                {
                    return std::string();
                }
                pend--;
            }

            return std::string(pbeg, pend - pbeg + 1);
        }

        std::string escapeNonAsciiAndReserved(const std::string &string, const char *def)
        {
            std::stringbuf buf;
            std::ostream strm(&buf);

            for (size_t i = 0; i < string.length(); i++)
            {
                if (isAlphaNum(string[i]))
                {
                    strm << string[i];
                }
                else
                {
                    int j;
                    for (j = 0; def[j] != '\0' && def[j] != string[i]; j++)
                    {
                        /* Itterate to or not match with any reserved symbols*/
                    }

                    if (def[j] != '\0')
                    {
                        strm << string[i];
                    }
                    else
                    {
                        strm << "%" << std::setw(2) << std::setfill('0') << std::hex << string[i];
                    }
                }
            }

            return buf.str();
        }

        void replaceAllLws(char *sip_message)
        {
            if (sip_message == nullptr)
            {
                return;
            }

            char *tmp = sip_message;

            for (; tmp[0] != '\0'; tmp++)
            {
                if (('\0' == tmp[0])
                    || ('\0' == tmp[1])
                    || ('\0' == tmp[2])
                    || ('\0' == tmp[3]))
                {
                    return;
                }

                if (((g_cr == tmp[0]) && (g_lf == tmp[1]) && (g_cr == tmp[2]) && (g_lf == tmp[3]))
                    || ((g_cr == tmp[0]) && (g_cr == tmp[1]))
                    || ((g_lf == tmp[0]) && (g_lf == tmp[1])))
                {
                    return; /* end of message */
                }

                if (((g_cr == tmp[0])
                     && (g_lf == tmp[1])
                     && ((g_sp == tmp[2]) || (g_hTab == tmp[2])))
                    || ((g_cr == tmp[0])
                        && ((g_sp == tmp[1]) || (g_hTab == tmp[1])))
                    || ((g_lf == tmp[0])
                        && ((g_sp == tmp[1]) || (g_hTab == tmp[1]))))
                {
                    /// replace line end and TAB symbols by SP
                    tmp[0] = g_sp;
                    tmp[1] = g_sp;
                    tmp = tmp + 2;

                    /// replace all following TAB symbols
                    while (g_hTab == tmp[0] || g_sp == tmp[0])
                    {
                        tmp[0] = g_sp;
                        tmp++;
                    }

                    if (tmp[0] == '\0')
                    {
                        return;
                    }
                }
            }
        }

        int strStr(const char *str, const char *buf,
                   const char **index_of_str, const char *endOfBuf)
        {
            size_t slen;

            *index_of_str = nullptr;

            if (str == nullptr || buf == nullptr)
            {
                return sip::BADPARAMETER;
            }

            slen = std::strlen(str);

            while (slen < (size_t)(endOfBuf - buf))
            {
                if (!std::memcmp(str, buf, slen))
                {
                    *index_of_str = buf;
                    return SUCCESS;
                }

                ++buf;
            }

            return sip::SYNTAXERROR;
        }

        int findNextCrLf(const char *startOfHeader, const char **end_of_header)
        {
            const char *tmp = startOfHeader;

            *end_of_header = nullptr;

            while ((g_cr != tmp[0]) && (g_lf != tmp[0]))
            {
                if (tmp[0])
                {
                    tmp++;
                }
                else
                {
                    LOGERROR("Final CRLF is missing");
                    return SYNTAXERROR;
                }
            }

            if ((g_cr == tmp[0]) && (g_lf == tmp[1]))
            {
                tmp++;
            }

            if ((g_sp == tmp[1]) || (g_hTab == tmp[1]))
            {
                LOGERROR("LWS detected.");
                return BADPARAMETER;
            }

            *end_of_header = tmp + 1;
            return SUCCESS;
        }

        char *findStr(const char *bufToSearchWith, size_t bufLenLeft, const char *strToFind)
        {
            const char *search = bufToSearchWith;
            const char *end = bufToSearchWith + bufLenLeft;
            char *p;
            size_t len = std::strlen(strToFind);

            while (search < end)
            {
                p = (char *)std::memchr(search, *strToFind, end - search);

                if (p == nullptr || ((p + len) > end))
                {
                    break;
                }

                if (std::memcmp(p, strToFind, len) == 0)
                {
                    return p;
                }

                search = p + 1;
            }

            return nullptr;
        }

        int atoi(const std::string &number)
        {
            long int i;

            i = strtol(number.c_str(), nullptr, 10);

            if (i == LONG_MIN || i == LONG_MAX)
            {
                return UNDEFINED_ERROR;
            }

            return (int)i;
        }

    } // namespace utils

    namespace support
    {

        static uint s_sipCSeq = 0;

        std::string getNumber(const sip::Contact &contact,
                              const std::string &separator)
        {
            if (contact.prefix.empty())
            {
                return contact.number;
            }

            return (contact.prefix + separator + contact.number);
        }

        void parseUriNumber(Uvr::Number &uriNumber,
                            Uvr::OptionalOperatorId &operId,
                            const std::string &separator)
        {
            operId.reset();
            std::size_t separatorPos = uriNumber.find(separator);
            if (separatorPos != std::string::npos)
            {
                const std::string prefix = uriNumber.substr(0, separatorPos);
                operId = strtoul(prefix.c_str(), nullptr, 10);
                uriNumber.erase(0, separatorPos + separator.length());
            }
        }

        std::string composeInvite(const sip::Contact &from,
                                  const sip::Contact &to,
                                  const std::string &separator)
        {
            ++s_sipCSeq;
            Uvr::CallId_t tCallId = std::to_string(s_sipCSeq);
            int tcSeq;
            return composeInvite(from, to, tCallId, tcSeq, separator);
        }

        std::string composeInvite(const sip::Contact &from,
                                  const sip::Contact &to,
                                  Uvr::CallId_t &callId,
                                  int &cSeq,
                                  const std::string &separator)
        {
            if (callId.empty())
            {
                ++s_sipCSeq;
                callId = std::to_string(s_sipCSeq);
            }
            Uvr::CallId_t tCallId = callId;
            return composeInvite(from, to, (const Uvr::CallId_t &)tCallId, cSeq, separator);
        }

        std::string composeInvite(const sip::Contact &from,
                                  const sip::Contact &to,
                                  const Uvr::CallId_t &callId,
                                  int &cSeq,
                                  const std::string &separator)
        {
            ++s_sipCSeq;
            cSeq = s_sipCSeq;

            std::string sipInvite;
            std::string body;

            body += "v=0\r\n";
            body += "o=user1 53655765 2353687637 IN IP4 " + from.ip + "\r\n";
            body += "s=-\r\n";
            body += "c=IN IP4 " + from.ip + "\r\n";
            body += "t=0 0\r\n";
            body += "m=audio 6000 RTP/AVP 0\r\n";
            body += "a=rtpmap:0 PCMU/8000\r\n";

            sipInvite += "INVITE sip:" + getNumber(to, separator) + "@" + to.ip + ":" + to.port + " SIP/2.0\r\n";
            sipInvite += "Via: SIP/2.0/UDP " + from.ip + ":" + from.port + ";branch=z9hG4bK-1337486-245-0\r\n";
            sipInvite += "From: " + (from.name.empty() ? "" : "\"" + from.name + "\"") + " <sip:" + getNumber(from, separator) + "@" + from.ip + ";user=phone>;tag=1337486SIPpTag00245\r\n";
            sipInvite += "To: " + to.name + " <sip:" + getNumber(to, separator) + "@" + to.ip + ":" + to.port + ";user=phone>\r\n";
            if (!to.originalNumber.empty())
            {
                sipInvite += "Diversion: sip:"
                             "79871112234"
                             "@1.2.3.4;noa=1;reason=user-busy;counter=1;privacy=off;\r\n";
                sipInvite += "Diversion: sip:" + to.originalNumber + "@1.2.3.4;noa=1;reason=user-busy;counter=1;privacy=off;\r\n";
            }
            sipInvite += "Content-Type: application/sdp\r\n";
            sipInvite += "Call-ID: " + callId + "\r\n";
            sipInvite += "CSeq: " + std::to_string(s_sipCSeq) + " INVITE\r\n";
            sipInvite += "Contact: <sip:" + getNumber(from, separator) + "@" + from.ip + ":" + from.port + ">\r\n";
            sipInvite += "Max-Forwards: 70\r\n";
            sipInvite += "Subject: Performance Test\r\n";
            sipInvite += "Content-Length: " + std::to_string(body.length()) + "\r\n";
            sipInvite += "\r\n";
            sipInvite += body;

            return sipInvite;
        }

        std::string composeAck(const sip::Contact &from,
                               const sip::Contact &to,
                               const Uvr::CallId_t &callId,
                               const int &cSeq,
                               const std::string &separator)
        {
            std::string sipAck;
            std::string body;

            sipAck += "ACK sip:" + getNumber(to, separator) + "@" + to.ip + ":" + to.port + " SIP/2.0\r\n";
            sipAck += "Via: SIP/2.0/UDP " + from.ip + ":" + from.port + ";branch=z9hG4bK-1337486-245-0\r\n";
            sipAck += "From: " + from.name + " <sip:" + getNumber(from, separator) + "@" + from.ip + ">;tag=1337486SIPpTag00245\r\n";
            sipAck += "To: " + to.name + " <sip:" + getNumber(to, separator) + "@" + to.ip + ":" + to.port + ">\r\n";
            sipAck += "Max-Forwards: 70\r\n";
            sipAck += "Call-ID: " + callId + "\r\n";
            sipAck += "CSeq: " + std::to_string(cSeq) + " ACK\r\n";
            sipAck += "Content-Length: " + std::to_string(body.length()) + "\r\n";
            sipAck += "\r\n";

            return sipAck;
        }

        std::string composeOptions(const sip::Contact &from,
                                   const sip::Contact &to,
                                   const std::string &separator)
        {
            ++s_sipCSeq;
            Uvr::CallId_t tCallId = std::to_string(s_sipCSeq);
            int tcSeq;
            return composeOptions(from, to, tCallId, tcSeq, separator);
        }

        std::string composeOptions(const sip::Contact &from,
                                   const sip::Contact &to,
                                   const Uvr::CallId_t &callId,
                                   const int &cSeq,
                                   const std::string &separator)
        {
            std::string sipAck;
            std::string body;

            sipAck += "OPTIONS sip:" + getNumber(to, separator) + "@" + to.ip + ":" + to.port + " SIP/2.0\r\n";
            sipAck += "Via: SIP/2.0/UDP " + from.ip + ":" + from.port + ";branch=z9hG4bK-1337486-245-0\r\n";
            sipAck += "From: " + from.name + " <sip:" + getNumber(from, separator) + "@" + from.ip + ">;tag=1337486SIPpTag00245\r\n";
            sipAck += "To: " + to.name + " <sip:" + getNumber(to, separator) + "@" + to.ip + ":" + to.port + ">\r\n";
            sipAck += "Max-Forwards: 70\r\n";
            sipAck += "Call-ID: " + callId + "\r\n";
            sipAck += "CSeq: " + std::to_string(cSeq) + " OPTIONS\r\n";
            sipAck += "Contact: <sip:alice@pc33.atlanta.com>\r\n";
            sipAck += "Accept: application/sdp\r\n";
            sipAck += "Content-Length: " + std::to_string(body.length()) + "\r\n";
            sipAck += "\r\n";

            return sipAck;
        }

        std::string composeCancel(const sip::Contact &from,
                                  const sip::Contact &to,
                                  const std::string &separator)
        {
            ++s_sipCSeq;
            Uvr::CallId_t tCallId = std::to_string(s_sipCSeq);
            int tcSeq;
            return composeCancel(from, to, tCallId, tcSeq, separator);
        }

        std::string composeCancel(const sip::Contact &from,
                                  const sip::Contact &to,
                                  const Uvr::CallId_t &callId,
                                  const int &cSeq,
                                  const std::string &separator)
        {
            std::string sipAck;
            std::string body;

            sipAck += "CANCEL sip:" + getNumber(to, separator) + "@" + to.ip + ":" + to.port + " SIP/2.0\r\n";
            sipAck += "Via: SIP/2.0/UDP " + from.ip + ":" + from.port + ";branch=z9hG4bK-1337486-245-0\r\n";
            sipAck += "From: " + from.name + " <sip:" + getNumber(from, separator) + "@" + from.ip + ">;tag=1337486SIPpTag00245\r\n";
            sipAck += "To: " + to.name + " <sip:" + getNumber(to, separator) + "@" + to.ip + ":" + to.port + ">\r\n";
            sipAck += "Max-Forwards: 70\r\n";
            sipAck += "Call-ID: " + callId + "\r\n";
            sipAck += "CSeq: " + std::to_string(cSeq) + " CANCEL\r\n";
            sipAck += std::string("Reason: Q.850;cause=16;text=\"Normal call clearing\"") + "\r\n";
            sipAck += "Content-Length: " + std::to_string(body.length()) + "\r\n";
            sipAck += "\r\n";

            return sipAck;
        }

    } // namespace support

} // namespace sip
