#include "node.hpp"

// AFAP includes.
#include "components/tools/src/Tools.hpp"
#include "components/db/src/types.hpp"
#include "components/config/src/config_manager.hpp"
#include "components/sip/src/uri.hpp"
#include "modules/storage/src/StorageApi.hpp"
#include "TRecQ763Number.hpp"
#include "UvrAPI.hpp"

#include <regex>
#include <algorithm>

namespace
{
    std::string s_prefixSeparator = sip::config::idSrcSeparatorDefaultValue;

    // Connection parameters variables.
    std::string s_incomingCallsPort = modules::sip::config::incomingCallsPortDefaultValue;
    std::string s_outgoingCallsPort = modules::sip::config::outgoingCallsPortDefaultValue;
    std::string s_interface = modules::sip::config::interfaceDefaultValue;

    sip::response::Reason s_sipResponseTrying{100, "Trying"};
    sip::response::Reason s_sipResponseOutgoingCall;

    sip::response::Reason s_sipResponseOptionsNotStarted;

    std::unordered_map<Uvr::VerificationResponse_t,
                       sip::response::Reason>
        s_mapVerificationVerdictToSip;

    std::string defaultSipPort("5060");

} // namespace

namespace modules
{

    namespace sip
    {

        std::shared_ptr<node> api::node_p;

        static const std::string owner("MODULE:SIP");

        void verifyCallVerdictCallback(const Uvr::VerificationResult &result)
        {
            if (api::node_p)
            {
                api::node_p->handleVerdict(result);
            }
        };

        void api::init(const std::string &configPath,
                       const ::Uvr::UserId_t &apiUser)
        {
            node_p.reset();
            node_p = std::make_shared<node>(configPath, apiUser,
                                            Log::Tag("SIP::APIGW"));
        }

        void api::start(const int &numberOfThreads)
        {
            if (node_p)
            {
                node_p->start(numberOfThreads);
            }
        }

        void api::stop()
        {
            if (node_p)
            {
                node_p->stop();
            }
        }

        void api::wait_stop()
        {
            if (node_p)
            {
                node_p->wait_stop();
                node_p.reset();
            }
        }

        void api::addListener(::sip::trunk::Protocol protocol,
                              const std::string &address,
                              const std::string &port,
                              const ::sip::trunk::Role &role)
        {
            if (node_p)
            {
                node_p->addListener(protocol, address, port, role);
            }
        }

        node::node(const std::string &configPath,
                   const ::Uvr::UserId_t &apiUser,
                   const Log::Tag &logTag)
            : m_logTag(logTag)
        {
            m_userId = Uvr::API::registerVerifyCallVerdictCallback(verifyCallVerdictCallback, apiUser);

            m_transportManager = std::make_shared<::sip::transport::Manager>(m_ioc, logTag);

            configuration::manager configManager;
            if (configManager.loadConfigFile(configPath) == configuration::errors::FAIL)
            {
                LOGW(m_logTag, "Configuration file \"", configPath, "\" could not be loaded.");
                return;
            }

            s_incomingCallsPort = configManager.getValue(config::incomingCallsPort,
                                                         config::incomingCallsPortDefaultValue);
            s_outgoingCallsPort = configManager.getValue(config::outgoingCallsPort,
                                                         config::outgoingCallsPortDefaultValue);
            s_interface = configManager.getValue(config::interface,
                                                 config::interfaceDefaultValue);

            s_mapVerificationVerdictToSip.emplace(Uvr::VerificationResponse::success,
                                                  configManager.getValue(
                                                      ::sip::config::response::verification::success,
                                                      ::sip::config::response::verification::successDefaultValue));
            s_mapVerificationVerdictToSip.emplace(Uvr::VerificationResponse::failure,
                                                  configManager.getValue(
                                                      ::sip::config::response::verification::failure,
                                                      ::sip::config::response::verification::failureDefaultValue));
            s_mapVerificationVerdictToSip.emplace(Uvr::VerificationResponse::timeout,
                                                  configManager.getValue(
                                                      ::sip::config::response::verification::timeout,
                                                      ::sip::config::response::verification::timeoutDefaultValue));
            s_mapVerificationVerdictToSip.emplace(Uvr::VerificationResponse::callingNumberNotRussian,
                                                  configManager.getValue(
                                                      ::sip::config::response::verification::notRussian,
                                                      ::sip::config::response::verification::notRussianDefaultValue));
            s_mapVerificationVerdictToSip.emplace(Uvr::VerificationResponse::callingNumberNotServiced,
                                                  configManager.getValue(
                                                      ::sip::config::response::verification::notServiced,
                                                      ::sip::config::response::verification::notServicedDefaultValue));
            s_mapVerificationVerdictToSip.emplace(Uvr::VerificationResponse::callingNumberNotFound,
                                                  configManager.getValue(
                                                      ::sip::config::response::verification::notFound,
                                                      ::sip::config::response::verification::notFoundDefaultValue));

            s_sipResponseOutgoingCall =
                configManager.getValue(
                    ::sip::config::response::outgoingCall,
                    ::sip::config::response::outgoingCallDefaultValue);

                s_sipResponseOptionsNotStarted =
                    configManager.getValue(
                        ::sip::config::response::options::notStarted,
                        ::sip::config::response::options::notStartedDefaultValue);

            s_prefixSeparator =
                configManager.getValue(
                    ::sip::config::idSrcSeparator,
                    ::sip::config::idSrcSeparatorDefaultValue);

            const auto udpRecBufSize =
                configManager.getValue(
                    ::modules::sip::config::udpSocketReceiveBufferSize,
                    ::modules::sip::config::udpSocketReceiveBufferSizeDefaultValue);

            m_transportManager->setUdpSocketRcvBufSize(udpRecBufSize);
        }

        void node::start(int threadPoolSize)
        {
            m_ioc.start(threadPoolSize);

            addListener(::sip::trunk::Protocol::UDP,
                        s_interface, s_incomingCallsPort, ::sip::trunk::Role::INC);
            addListener(::sip::trunk::Protocol::UDP,
                        s_interface, s_outgoingCallsPort, ::sip::trunk::Role::OUT);

            m_transportManager->bind(shared_from_this());
            m_transportManager->start();
        }

        void node::addListener(::sip::trunk::Protocol protocol,
                               const std::string &listener,
                               const std::string &port,
                               const ::sip::trunk::Role &role)
        {
            boost::system::error_code err;
            network::address ipAddress =
                network::make_address(listener, AF_INET, err);
            if (err)
            {
                return;
            }

            m_transportManager->addListener(protocol, ipAddress, port, role);
        }

        void node::stop()
        {
            m_ioc.stop();

            ::sip::transport::interfaces::handler::server::stop_server();

            m_transportManager->stop();
            LOGINFO(m_logTag, " node::stop done");
        }

        void node::wait_stop()
        {
            m_ioc.wait_stop();
        }

        void node::handleMessage(const ::sip::trunk::Reference &ref, const ::sip::MessagePtr &msg)
        {
            if (msg->m_method == "INVITE")
            {
                auto from = (::sip::header::From *)msg->m_from.get();
                auto to = (::sip::header::To *)msg->m_to.get();
                auto callIdHeader = (::sip::header::CallId *)msg->m_callId.get();

                // Get Calling number.
                if ((from == nullptr) || (from->m_uri == nullptr))
                {
                    LOGI("Malformed 'From:' header!");
                    return;
                }
                Uvr::Number callingNumberStr = from->m_uri->m_username;
                LOGT(m_logTag, "Calling number: ", callingNumberStr);

                // Get Called number
                if ((to == nullptr) || (to->m_uri == nullptr))
                {
                    LOGI("Malformed 'To:' header!");
                    return;
                }
                Uvr::Number calledNumberStr = to->m_uri->m_username;
                LOGT(m_logTag, "Called number: ", calledNumberStr);

                // Get Call ID.
                Uvr::CallId_t callId;
                if (callIdHeader != nullptr)
                {
                    callId = callIdHeader->m_value;
                }

                switch (ref.m_role)
                {
                    case ::sip::trunk::Role::INC:
                    {
                        // Parse source operator ID from calling number prefix.
                        Uvr::OptionalOperatorId operatorIdSrc;
                        ::sip::support::parseUriNumber(callingNumberStr, operatorIdSrc, s_prefixSeparator);
                        if (!operatorIdSrc.has_value())
                        {
                            LOGW(m_logTag, "Calling number = ", callingNumberStr,
                                 " does not contain a prefix with an operator id. undefinedOperatorId = 0 will be used.");
                        }
                        const auto callingNumberOctStr = ::TRecQ763Number::toOctetStringPtr(callingNumberStr);
                        const auto calledNumberOctStr = ::TRecQ763Number::toOctetStringPtr(calledNumberStr);

                        // Get Original Called number.
                        // The bottommost Diversion header is expected to be Original Called number,
                        Uvr::OctetStringPtr originalCalledNumberOctStr;
                        Uvr::Number originalCalledNumberStr = msg->getOriginalNumber();
                        originalCalledNumberOctStr = ::TRecQ763Number::toOctetStringPtr(msg->getOriginalNumber());

                        // Get Generic number.
                        Uvr::OctetStringPtr genericNumberOctStr;
                        const Uvr::Number genericNumberStr = Uvr::Number(::sip::utils::doUnQuote(from->m_displayName));
                        if (!genericNumberStr.empty())
                        {
                            LOGT(m_logTag, "Generic number: ", genericNumberStr);
                            genericNumberOctStr = ::TRecQ763Number::toOctetStringPtr(genericNumberStr);
                        }

                        {
                            ::transaction::errors err;
                            ::transaction::id sipSessionId{::transaction::undef_value};
                            TransactionManager::TransactionGuard guard =
                                m_vfCallsTransMngr.openTransactionEx(
                                    transaction::data(ref, msg,
                                                      callingNumberStr,
                                                      calledNumberStr,
                                                      originalCalledNumberStr,
                                                      genericNumberStr,
                                                      operatorIdSrc.value_or(Uvr::undefinedOperatorId),
                                                      callId),
                                    err, sipSessionId);

                            auto trusted = false;

                            // Verify call.
                            Uvr::SessionId_t sessionId =
                                Uvr::API::verifyCall(m_userId,
                                                     operatorIdSrc.value_or(Uvr::undefinedOperatorId),
                                                     callingNumberOctStr,
                                                     calledNumberOctStr,
                                                     originalCalledNumberOctStr,
                                                     genericNumberOctStr,
                                                     callId,
                                                     trusted);

                            m_vfCallsTransMngr.bind(sessionId, sipSessionId);
                        }

                        sendReply(ref, msg, s_sipResponseTrying);

                        break;
                    }
                    case ::sip::trunk::Role::OUT:
                    {
                        // Parse destination operator ID from called number prefix.
                        Uvr::OptionalOperatorId operatorIdDst;
                        ::sip::support::parseUriNumber(calledNumberStr, operatorIdDst, s_prefixSeparator);
                        const auto callingNumberOctStr = ::TRecQ763Number::toOctetStringPtr(callingNumberStr);
                        const auto calledNumberOctStr = ::TRecQ763Number::toOctetStringPtr(calledNumberStr);

                        // Process the outgoing call.
                        Uvr::API::notifyAboutOutgoingCall(callingNumberOctStr, calledNumberOctStr, callId, operatorIdDst);

                        sendReply(ref, msg, s_sipResponseTrying);

                        sendReply(ref, msg, s_sipResponseOutgoingCall);

                        break;
                    }
                    default:;
                }
            }
            else if (msg->m_method == "OPTIONS")
            {

                Uvr::NodeStatus nodeStatus = Uvr::API::getNodeStatus();
                ::sip::response::Reason optionsReply = s_sipResponseOptionsNotStarted;

                if (nodeStatus == Uvr::NodeStatus::Started)
                {
                    optionsReply = {200, "OK"};
                }

                if (s_sipResponseOptionsNotStarted.m_phrase.empty())
                {
                    std::stringbuf trace;
                    std::ostream logger(&trace);
                    logger << nodeStatus;
                    optionsReply.m_phrase = trace.str();
                }

                ::sip::header::GenericPtrList extraHeaders;
                extraHeaders.emplace_back(std::make_shared<::sip::header::Generic>(
                    "Allow", "INVITE, ACK, OPTIONS, CANCEL"));
                extraHeaders.emplace_back(std::make_shared<::sip::header::Generic>(
                    "Accept", ""));
                extraHeaders.emplace_back(std::make_shared<::sip::header::Generic>(
                    "Accept-Encoding", ""));
                extraHeaders.emplace_back(std::make_shared<::sip::header::Generic>(
                    "Accept-Language", "en"));
                extraHeaders.emplace_back(std::make_shared<::sip::header::Generic>(
                    "Supported", ""));

                sendReply(ref, msg, optionsReply, extraHeaders);
            }
            else if (msg->m_method == "CANCEL")
            {
                ::sip::response::Reason optionsReply{200, "OK"};

                sendReply(ref, msg, optionsReply);
            }
        }

        void node::sendReply(const ::sip::trunk::Reference &ref,
                             const ::sip::MessagePtr &request,
                             const ::sip::response::Reason &reason,
                             const ::sip::header::GenericPtrList &extraHeaders)
        {
            ::sip::Message reply;
            reply.m_from = request->m_from;
            reply.m_to = request->m_to;
            reply.m_via = request->m_via;
            reply.m_cSeq = request->m_cSeq;
            reply.m_callId = request->m_callId;
            reply.m_diversion = request->m_diversion;
            reply.m_headers = extraHeaders;

            if (request->m_method == "OPTIONS")
            {
                auto contact = std::make_shared<::sip::header::Contact>();
                contact->m_uri = request->m_uri;
                std::string name = "transport";
                auto it = find_if(contact->m_uri->m_params.cbegin(),
                                  contact->m_uri->m_params.cend(),
                                  [name](const ::sip::GenericParam &elem)
                                  { return (elem.m_name == name); });
                if (it == contact->m_uri->m_params.cend())
                {
                    std::string value = (ref.m_ulp == ::sip::trunk::Protocol::UDP ? "udp" : "tcp");
                    ::sip::GenericParam param(std::move(name), std::move(value));
                    contact->m_uri->m_params.emplace_back(param);
                }
                reply.m_contact.emplace_back(contact);
            }

            auto from = (::sip::header::From *)request->m_from.get();

            reply.setStatusCode(reason.m_code);
            reply.setReasonPhrase(reason.m_phrase);
            reply.setMethod("");

            switch (ref.m_ulp)
            {
                case ::sip::trunk::Protocol::UDP:
                {
                    std::string *remIp_p;
                    std::string *remPort_p;

                    if (reply.m_via.size() > 0)
                    {
                        /*
                        RFC 3261
                        '...The proxy MUST insert a Via header field value into the copy
                        before the existing Via header field values.'
                        It means the last added Via header is the topmost i.e.
                        the first one with respect to the message head
                        */
                        auto viaIter = reply.m_via.begin();
                        auto via = (::sip::header::Via *)viaIter->get();

                        remIp_p = &(via->m_host);
                        remPort_p = (via->m_port.empty() ? &defaultSipPort : &(via->m_port));
                    }
                    else
                    {
                        remIp_p = &(from->m_uri->m_host);
                        remPort_p = &(from->m_uri->m_port);
                    }

                    m_transportManager->send(ref, reply,
                                             *remIp_p,
                                             *remPort_p);
                    break;
                }
                case ::sip::trunk::Protocol::TCP:
                    m_transportManager->send(ref, reply);
                    break;
            }
        }

        void node::handleVerdict(const Uvr::VerificationResult &result)
        {
            ::transaction::errors res;
            transaction::data data(m_vfCallsTransMngr.closeTransaction(result.sessionId, res));
            if (::transaction::errors::OK == res)
            {
                auto reason = s_mapVerificationVerdictToSip.find(result.verificationVerdict);
                if (reason != s_mapVerificationVerdictToSip.end())
                {
                    sendReply(data.m_ref, data.m_msg, reason->second);
                }
                else
                {
                    LOGERROR(" Unsupported verificationVerdict = ",
                             (int)result.verificationVerdict);
                }
            }
        }

        void node::handleTrunkTcpUp(const ::sip::trunk::Reference &ref)
        {
            LOGINFO("Got SIP conn_up connId=", ref.m_key);
        }

        transaction::data
        node::TransactionManager::closeTransaction(Uvr::SessionId_t sessionId,
                                                   ::transaction::errors &result)
        {
            std::lock_guard<std::recursive_mutex> guard(m_guard);

            auto transactionIter = m_mapUvrToSipSession.find(sessionId);

            transaction::data value;
            if (transactionIter != m_mapUvrToSipSession.end())
            {
                value = BaseTransactionManager::closeTransaction(transactionIter->second, result);
                m_mapUvrToSipSession.erase(transactionIter);
                LOGDEBUG(m_ownInst, " Transaction sessionId=", sessionId, " is closed!");
            }
            else
            {
                LOGERROR(m_ownInst, " Cannot close transaction: sessionId=", sessionId);
                result = ::transaction::errors::FAIL;
            }
            return value;
        }

        void node::TransactionManager::bind(Uvr::SessionId_t sessionId,
                                            ::transaction::id id)
        {
            LOGD(m_ownInst, " bind sessionId=", sessionId, " with SIP transaction::id=", id);
            m_mapUvrToSipSession.emplace(sessionId, id);
        }

    } // namespace sip

} // namespace modules

namespace sip
{
    namespace response
    {

        Reason tag_invoke(boost::json::value_to_tag<Reason>,
                          boost::json::value const &jv)
        {
            boost::json::object const &obj = jv.as_object();
            return {boost::json::value_to<int>(obj.at("code")),
                    boost::json::value_to<std::string>(obj.at("phrase"))};
        }

    } // namespace response

} // namespace sip
