// Own include.
#include "StorageApi.hpp"

// AFAP includes.
#include "UvrAPI.hpp"

namespace
{
    using UvrIdToClientMap = std::unordered_map<Uvr::UvrId_t, Storage::ClientUniquePtr>;

    UvrIdToClientMap s_clients;
    bool s_isStarted = false;

} // namespace

// For tests only. Not part of the public API!
namespace Storage
{
    namespace TestAPI
    {
        Client *getClient(const Uvr::UvrId_t uvrId);

    } // namespace TestAPI

} // namespace Storage

bool Storage::API::registerClient(const Uvr::UvrId_t uvrId,
                                  ClientUniquePtr &&client)
{
    const bool isInserted = s_clients.emplace(uvrId, std::move(client)).second;
    if (!isInserted)
    {
        LOGE("Failed to register Storage client for UVR with ID=", uvrId, ". Client already registered.");
    }

    return true;
}

bool Storage::API::start()
{
    if (s_clients.empty())
    {
        LOGERROR("Start of Storage is impossible: Storage module was not initiated!");
        return false;
    }

    bool isOk = true;
    for (auto &uvrClientPair : s_clients)
    {
        const auto uvrId = uvrClientPair.first;
        const auto &client = uvrClientPair.second;

        if (!client->connect())
        {
            // Information about error with connection to storage
            // should be printed inside connect() function.
            isOk = false;
            continue;
        }

        const auto actualVersion = client->getActualVersion();
        const auto expectedVersion = client->getExpectedVersion();
        if (actualVersion <= 0)
        {
            client->init();
        }
        else if (actualVersion < expectedVersion)
        {
            LOGI("The data in Storage for UVR with ID=", uvrId, " has a lower version than expected.",
                 " Expected: ", expectedVersion, ", Actual: ", actualVersion, ". Upgrading...");
            if (!client->upgrade(actualVersion, expectedVersion))
            {
                LOGE("Failed to upgrade data in Storage for UVR with ID=", uvrId, ".");
                isOk = false;
                continue;
            }
            LOGI("The data in Storage for UVR with ID=", uvrId, " has been successfully upgraded.");
        }
        else if (actualVersion > expectedVersion)
        {
            LOGE("The data in Storage for UVR with ID=", uvrId, " has a higher version than expected.",
                 " Expected: ", expectedVersion, ", Actual: ", actualVersion, ". Downgrading is not supported.");
            isOk = false;
            continue;
        }
    }

    if (!isOk)
    {
        LOGE("Can't start Storage module.");
        return false;
    }

    s_isStarted = true;

    return true;
}

bool Storage::API::isStarted()
{
    return s_isStarted;
}

void Storage::API::stop()
{
    for (const auto &uvrClientPair : s_clients)
    {
        uvrClientPair.second->disconnect();
    }
    s_clients.clear();

    s_isStarted = false;
}

bool Storage::API::storeRecord(const Uvr::UvrId_t uvrId,
                               const Uvr::CallRecord &record)
{
    const auto clientsIter = s_clients.find(uvrId);
    if (clientsIter == s_clients.end())
    {
        LOGE("Can't store record for call with ID=\"", record.callId, "\".",
             " Storage is not initialized for UVR with ID=", uvrId, ".");
        return false;
    }

    clientsIter->second->storeVerdict(record);

    return true;
}

size_t Storage::API::updateRecords(const Uvr::UvrId_t uvrId,
                                   const Uvr::CallRecords &records)
{
    const auto clientsIter = s_clients.find(uvrId);
    if (clientsIter == s_clients.end())
    {
        std::string callsIds = "";
        for (const auto &record : records)
        {
            callsIds += "\"" + record.callId + "\";";
        }

        LOGE("Can't update records for calls with following IDs: ", callsIds,
             "\nStorage is not initialized for UVR with ID=", uvrId, ".");
        return false;
    }

    const auto &client = clientsIter->second;

    size_t nbUpdatedRecords = 0;
    for (const auto &record : records)
    {
        if (client->updateVerdict(record))
        {
            ++nbUpdatedRecords;
        };
    }

    return nbUpdatedRecords;
}

void Storage::API::processCallInformationRequest(
    const Uvr::UvrId_t uvrId,
    const Uvr::CallInformationRequest &request)
{
    const auto clientsIter = s_clients.find(uvrId);
    if (clientsIter == s_clients.end())
    {
        LOGE("Can't process call information request for call with ID=\"", request.callId, "\".",
             " Storage is not initialized for UVR with ID=", uvrId, ".");
        return;
    }

    Uvr::CallInformationResponses responses;
    clientsIter->second->db::interfaces::tables::calls::client::get(responses, request);
    Uvr::API::reportCallInformation(responses);
}

Storage::Client *Storage::TestAPI::getClient(const Uvr::UvrId_t uvrId)
{
    const auto clientsIter = s_clients.find(uvrId);
    if (clientsIter == s_clients.end())
    {
        LOGE("Storage is not initialized for UVR with ID=", uvrId, ".");
        return nullptr;
    }

    return clientsIter->second.get();
}
